/**
 * SLUCE 2 Project
 * http://www.cscs.umich.edu/research/projects/sluce/
 */
package sluceii.agent.decisionmaking;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.TreeMap;

import cern.jet.random.Uniform;

import sluceii.agent.SimpleAgent;
import sluceii.agent.decisionmaking.LinkedComparator.LinkedComparatorType;
import sluceii.params.AppParameters;
import sluceii.land.market.LandListingRecord;
import sluceii.land.market.LandMarketListing;
import sluceii.land.market.decision.LandPriceEstimator;
import sluceii.land.market.decision.LandProfitEstimator;
import sluceii.land.market.decision.LandSearch;
import sluceii.land.market.utility.ConsumptionBundle;
import sluceii.land.market.utility.UtilityStrategy;
import sluceii.landscape.LandCoverDataHelper;
import sluceii.landscape.LandUnit;
import sluceii.landscape.LandUnit.LandAggregateMethod;
import umn.geog.helia.landscape.RasterLayer;

/**
 * This is a simple housing search helper class. Users can extend this class to implement
 * HousingSearch interface.
 * 
 * It uses two type parameters, one is the LandUnit and the other is the agent. They define
 * who search for what.
 * 
 * @author Shipeng Sun
 */
public class SimpleLandSearchHelper<L extends LandUnit, S extends SimpleAgent> implements LandSearch<L>
{
	/**
	 * Spatial Search strategy
	 */
	public enum SpatialSearchStrategyType {random, neighborhood, exponential};
	
	/**
	 * the agent searching for vacancies
	 */
	protected S landSearcher;
	
	/**
	 * 
	 */
	protected LandMarketListing<L> landMarket; 
	
	/**
	 * The utility strategy of the housing searcher
	 */
	protected UtilityStrategy utilStrategy;
	
	/**
	 * The land profit estimator used by the agent (developer)
	 */
	protected LandProfitEstimator profitEstimator;
	
	/**
	 * The type of spatial search 
	 */
	protected SpatialSearchStrategyType spatialSearchStrategy;
	
	/**
	 * 
	 */
	protected Uniform uniformRdmGenerator;
	
	/**
	 * 
	 */
	public SimpleLandSearchHelper(Uniform rndGenerator)
	{
		uniformRdmGenerator = rndGenerator;
	}
	
	public SimpleLandSearchHelper(S houseSearcher, UtilityStrategy utilStrategy, LandMarketListing<L> landMarket)
	{
		this.landSearcher = houseSearcher;
		this.utilStrategy = utilStrategy;
		this.landMarket = landMarket;
	}

	/**
	 * @return the landSearcher
	 */
	public S getHousingSearcher()
	{
		return landSearcher;
	}

	/**
	 * @param landSearcher the landSearcher to set
	 */
	public void setHousingSearcher(S landSearcher)
	{
		this.landSearcher = landSearcher;
	}

	/**
	 * @return the utilStrategy
	 */
	public UtilityStrategy getUtilStrategy()
	{
		return utilStrategy;
	}

	/**
	 * @param utilStrategy the utilStrategy to set
	 */
	public void setUtilStrategy(UtilityStrategy utilStrategy)
	{
		this.utilStrategy = utilStrategy;
	}

	/**
	 * @return the landMarket
	 */
	public LandMarketListing<L> getLandMarket()
	{
		return landMarket;
	}

	/**
	 * @param landMarket the landMarket to set
	 */
	public void setLandMarket(LandMarketListing<L> landMarket)
	{
		this.landMarket = landMarket;
	}
	
	public SpatialSearchStrategyType getSpatialSearchStrategy() 
	{
		return spatialSearchStrategy;
	}

	public void setSpatialSearchStrategy(
			SpatialSearchStrategyType spatialSearchStrategy) {
		this.spatialSearchStrategy = spatialSearchStrategy;
	}

	/**
	 * @return the uniformRdmGenerator
	 */
	public Uniform getUniformRdmGenerator() {
		return uniformRdmGenerator;
	}

	/**
	 * @param uniformRdmGenerator the uniformRdmGenerator to set
	 */
	public void setUniformRdmGenerator(Uniform uniformRdmGenerator) {
		this.uniformRdmGenerator = uniformRdmGenerator;
	}

	/**
	 * Residential households use this method find the best parcel to bid on with a given price estimator
	 * @param totalSearchNumber
	 * @param minUtil
	 * @param resourceQuantity
	 * @param landAppraiser
	 * @return
	 */
	@Override
	public ArrayList<LandListingRecord<L, ?>> searchProperties(int totalSearchNumber,double minUtil,double resourceQuantity,LandPriceEstimator<L> landAppraiser)
	{
		// Search the listing records stored in landMarket
		//ArrayList<LandListingRecord<L, ?>> allList = landMarket.getToSellLandUnitList();
		ArrayList<LandListingRecord<L, ?>> allList = new ArrayList<LandListingRecord<L,?>>();
		allList.addAll(landMarket.getToSellLandUnitList().values());		
		allList.trimToSize();
		if(allList.isEmpty()) return null;
		
		// Search using spatial rules to get a sub-sample
		ArrayList<LandListingRecord<L, ?>> spatialList = spatialSearch(allList,totalSearchNumber, spatialSearchStrategy);
		spatialList.trimToSize();
		if(spatialList.isEmpty()) return null;
		
		// Calculate utilities
		UtilityStrategy utilStrategy = this.utilStrategy;
		ArrayList<Double> utilList = calcUtilities(spatialList,utilStrategy);
		utilList.trimToSize();
		if(utilList.isEmpty()) return null;
		
		ArrayList<Double> highUtilValList = sortArraysWithDictionary(utilList,utilList,true);
		if(highUtilValList.isEmpty()) return null;
		
		ArrayList<LandListingRecord<L,?>> highUtilityList = sortArraysWithDictionary(utilList, spatialList, true);
		highUtilityList.trimToSize();
		if(highUtilityList.isEmpty()) return null;
		
		// Get rid of those whose utilities are lower than the minUtil
		while(highUtilValList.get(highUtilValList.size()-1) < minUtil)
		{
			highUtilValList.remove(highUtilValList.size()-1);
			highUtilityList.remove((highUtilityList.size()-1));
			
			if(highUtilValList.size()==0) break;
		}
		
		highUtilityList.trimToSize();
		if(highUtilityList.isEmpty()) return null;
		
		ArrayList<LandListingRecord<L,?>> rsltList = new ArrayList<LandListingRecord<L,?>>();
		LandListingRecord<L,?> llRecord = null;
		Iterator<LandListingRecord<L,?>> itr = highUtilityList.iterator();
		int cnt = 0;
		double estimatedPrice = 0.0d;
		
		// Don't include those parcels that have already had maximum bid numbers.
		while(itr.hasNext())
		{
			llRecord = itr.next();
			if(landMarket.getOpenBidNumber(llRecord.getLandUnitID()) < llRecord.getMaxBidsAllowed())
			{
				estimatedPrice = landAppraiser.estimateLandUnitPrice(llRecord.getLandUnitObj());
				if(Double.isNaN(estimatedPrice)) estimatedPrice = 0.0d;
				if(Math.max(estimatedPrice, llRecord.getAskPrice()) <= resourceQuantity)
				{
					rsltList.add(llRecord);
					cnt++;
				}
			}
		}
			
		return rsltList;		
	}
	
	/**
	 * Developers use this method to find the best subdivision to bid on 
	 * @param totalSearchNumber
	 * @param minUtil
	 * @param resourceQuantity
	 * @param landAppraiser
	 * @return
	 */
	public ArrayList<LandListingRecord<L, ?>> searchProperties(int targetNum, int totalSearchNumber,double minProfit,double resourceQuantity,LandProfitEstimator landAppraiser)
	{
		// Search the listing records stored in landMarket
		ArrayList<LandListingRecord<L, ?>> allList = new ArrayList<LandListingRecord<L,?>>();
		allList.addAll(landMarket.getToSellLandUnitList().values());		
		allList.trimToSize();
		if(allList.isEmpty()) return null;
		
		// Search using spatial rules to get a sub-sample
		ArrayList<LandListingRecord<L, ?>> spatialList = spatialSearch(allList,totalSearchNumber, spatialSearchStrategy);
		spatialList.trimToSize();
		if(spatialList.isEmpty()) return null;
		
		// Calculate expected profit
		// This must return multiple ArrayList objects, containing estimated profit and costs.
		ArrayList<ArrayList<Double>> profitListVector = calcProfits(spatialList,landAppraiser);
		if(profitListVector == null || profitListVector.isEmpty()) return null;
				
		ArrayList<Double> profitList = profitListVector.get(0);
		profitList.trimToSize();
		if(profitList.isEmpty()) return null;
		
		ArrayList<Double> costList = profitListVector.get(1);
		costList.trimToSize();
		if(costList == null || costList.isEmpty()) return null;
		
		ArrayList<Double> highProfitValList = sortArraysWithDictionary(profitList,profitList,true);
		if(highProfitValList.isEmpty()) return null;
		
		ArrayList<LandListingRecord<L,?>> highProfitLandList = sortArraysWithDictionary(profitList, spatialList, true);
		highProfitLandList.trimToSize();
		if(highProfitLandList.isEmpty()) return null;
		
		// Get rid of those whose expected profits are lower than the minProfit
		while(highProfitValList.get(highProfitValList.size()-1) < minProfit)
		{
			highProfitValList.remove(highProfitValList.size()-1);
			highProfitLandList.remove((highProfitLandList.size()-1));
			
			if(highProfitValList.size()==0) break;
		}
		
		highProfitLandList.trimToSize();
		if(highProfitLandList.isEmpty()) return null;
		
		ArrayList<LandListingRecord<L,?>> rsltList = new ArrayList<LandListingRecord<L,?>>();
		LandListingRecord<L,?> llRecord = null;
		Iterator<LandListingRecord<L,?>> itr = highProfitLandList.iterator();
		int cnt = 0;
		double estimatedPrice = 0.0d;
		
		// Don't include those parcels that have already had maximum bid numbers.
		while(itr.hasNext() && cnt < targetNum)
		{
			llRecord = itr.next();
			if(landMarket.getOpenBidNumber(llRecord.getLandUnitID()) < llRecord.getMaxBidsAllowed())
			{
				//estimatedPrice = landAppraiser.estimateLandUnitPrice(llRecord.getLandUnitObj());
				if(Double.isNaN(estimatedPrice)) estimatedPrice = 0.0d;
				if(Math.max(estimatedPrice, llRecord.getAskPrice()) <= resourceQuantity)
				{
					rsltList.add(llRecord);
					cnt++;
				}
			}
		}
			
		return rsltList;		
	}	
	
	/**
	 * Developers calculate the expected profits from developing the land listed
	 * @param reclList
	 * @param pEstimator
	 * @return A array list containing at least profit and cost array lists
	 */
	private ArrayList<ArrayList<Double>> calcProfits(ArrayList<LandListingRecord<L, ?>> reclList, LandProfitEstimator pEstimator) 
	{
		Iterator<LandListingRecord<L,?>> itRec = reclList.iterator();
		ArrayList<ArrayList<Double>> rsltUtilities = new ArrayList<ArrayList<Double>>();
		LandListingRecord<L,?> curRecord = null;
		L landObj = null;
		double askPrice = 0.0;
		
		ArrayList<Double> profits = new ArrayList<Double>();
		ArrayList<Double> costs   = new ArrayList<Double>();
		
		ArrayList<Number> tmpRslt;
		while(itRec.hasNext())
		{
			curRecord = itRec.next();
			if(curRecord != null)
			{
				askPrice = curRecord.getAskPrice();
				landObj = curRecord.getLandUnitObj();
				
				if(landObj == null)
				{
					throw new NullPointerException("Listing Record has a Null pointer to Land Unit Object.");
				}
				
				tmpRslt = pEstimator.estimateExpectedLandProfit(landObj, askPrice);
				if(tmpRslt == null || tmpRslt.size() < 6)
				{
					throw new NullPointerException("Profit Estimator Returns a Null pointer.");
				}

				profits.add(tmpRslt.get(1).doubleValue());
				costs.add(tmpRslt.get(4).doubleValue());
			}
		}

		rsltUtilities.add(profits);
		rsltUtilities.add(costs);
		
		return rsltUtilities;
	}

	public ArrayList<LandListingRecord<L, ?>> calcExpectedPrice(ArrayList<LandListingRecord<L, ?>> luList, LandPriceEstimator<? extends LandUnit> landAppraiser)
	{
		return null;
	}

	/* (non-Javadoc)
	 * @see sluceii.agent.decisionmaking.HousingSearch#searchProperties(int)
	 */
	@Override
	public ArrayList<LandListingRecord<L, ?>> searchProperties(int targetNum, int sampleNum)
	{
		// Search the listing records stored in landMarket
		ArrayList<LandListingRecord<L, ?>> allList = new ArrayList<LandListingRecord<L, ?>>();
		if(!allList.addAll(landMarket.getToSellLandUnitList().values())) return null;
		
		allList.trimToSize();
		
		// Search using spatial rules to get a sub-sample
		ArrayList<LandListingRecord<L, ?>> spatialList = spatialSearch(allList,sampleNum, SpatialSearchStrategyType.random);
		spatialList.trimToSize();
		
		// Calculate utilities
		UtilityStrategy utilStrategy = this.utilStrategy;
		ArrayList<Double> utilList = calcUtilities(spatialList,utilStrategy);
		utilList.trimToSize();
		
		ArrayList<LandListingRecord<L,?>> highUtilityList = sortArraysWithDictionary(utilList, spatialList, true);
		highUtilityList.trimToSize();
		
		ArrayList<LandListingRecord<L,?>> rsltList = new ArrayList<LandListingRecord<L,?>>();
		LandListingRecord<L,?> llRecord = null;
		Iterator<LandListingRecord<L,?>> itr = highUtilityList.iterator();
		int cnt = 0;
		
		// Don't include those parcels that have already had maximum bid numbers.
		while(itr.hasNext() && cnt < targetNum)
		{
			llRecord = itr.next();
			if(landMarket.getOpenBidNumber(llRecord.getLandUnitID()) < llRecord.getMaxBidsAllowed())
			{
				rsltList.add(llRecord);
				cnt++;
			}
		}
			
		return rsltList;		
	}
	
	/**
	 * Calculate the utilities of a list of parcel/properties/houses
	 * @param spatialList
	 * @param utilStrategy
	 * @return A list of utilities
	 */
	private ArrayList<Double> calcUtilities(ArrayList<LandListingRecord<L,?>> reclList, UtilityStrategy utilStrategy)
	{
		Iterator<LandListingRecord<L,?>> itRec = reclList.iterator();
		ArrayList<Double> rsltUtilities = new ArrayList<Double>();
		LandListingRecord<L,?> curRecord = null;
		LandUnit landObj = null;
		ConsumptionBundle cBundle = null;
		
		while(itRec.hasNext())
		{
			curRecord = itRec.next();
			if(curRecord != null)
			{
				landObj = curRecord.getLandUnitObj();
				//cBundle = LandCoverDataHelper.getConsumptionBundle(landObj, this.landSearcher);
				cBundle = LandCoverDataHelper.getConsumptionBundle(landObj, this.landSearcher);
				
				rsltUtilities.add(utilStrategy.getUtility(cBundle));
			}
		}
		
		return rsltUtilities;
	}

	/**
	 * Right now, only random strategy is implemented.
	 * Parcels that have been bid over the limit and are not affordable for this buyer will be 
	 * bypassed.
	 * 
	 * TODO: 
	 * 
	 * @param recList
	 * @param num
	 * @param searchStrategy
	 * @return
	 */
	protected ArrayList<LandListingRecord<L,?>> spatialSearch(ArrayList<LandListingRecord<L,?>> recList, int num, SpatialSearchStrategyType searchStrategy)
	{
		// All recorded will be sorted according to a probability associated with them.
		// Generate probabilities and associate them with listing records
		double rndNum = 0.0, dist = 0.0, transportCost = 0.0;
		LandListingRecord<L,?> curRecord = null;
		RasterLayer<?> distLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.distanceLayerIndex);
		ArrayList<Double> probArray = new ArrayList<Double>(recList.size());
		probArray.clear();
		
		switch (searchStrategy)
		{
			case random:
			{
				try
				{
					Iterator<LandListingRecord<L,?>> itRec = recList.iterator();
					
					while(itRec.hasNext())
					{
						// As the TreeMap sort elements from lowest to highest. It is easier to interpret when we
						// use 1- random, although the effects are the same.
						rndNum = 1- uniformRdmGenerator.nextDouble();
						curRecord = itRec.next();
						
						if(curRecord.getAskPrice()>100)
						{
							rndNum += 0;
						}
						
						if(landMarket.getCurrentBidsNumber(curRecord.getLandUnitID())>= curRecord.getMaxBidsAllowed())
						{
							// The parcel that has been bid will have little chance being selected.
							rndNum += 1.0;
						}

						dist = curRecord.getLandUnitObj().getValue(distLayer, LandAggregateMethod.average).doubleValue();
						transportCost = dist * AppParameters.unitTransportCost;
						
						if(curRecord.getAskPrice()+ transportCost >= this.landSearcher.getCapitalResource().getAssetQuantity())
						{
							rndNum += 1.0;
						}

						probArray.add(rndNum);
					}		
				}
				catch(Exception e)
				{
					e.printStackTrace();
				}
				break;
			}
			case neighborhood:
			{
				break;
			}
			case exponential:
			{
				break;
			}
			default:
			{
				
			}
		}
		
		// get the elements with highest probabilities and add it the result list
		ArrayList<LandListingRecord<L,?>> rsltList = this.sortArraysWithDictionary(probArray, recList, false, Math.min(num,recList.size()));
		return rsltList;
	}
	
	/**
	 * @param <T>
	 * @param criteriaArray
	 * @param targetArray
	 * @param isDecendingOrder
	 * @return
	 */
	private <T> ArrayList<T> sortArraysWithDictionary(ArrayList<Double> criteriaArray, ArrayList<T> targetArray, boolean isDecendingOrder)
	{
		return sortArraysWithDictionary(criteriaArray, targetArray, isDecendingOrder,criteriaArray.size());
	}
	
	/**
	 * 
	 * @param <T>
	 * @param criteriaArray
	 * @param targetArray
	 * @param isDecendingOrder
	 * @return
	 */
	private <T> ArrayList<T> sortArraysWithDictionary(ArrayList<Double> criteriaArray, ArrayList<T> targetArray, boolean isDecendingOrder, int num)
	{
		if(criteriaArray == null || targetArray == null || targetArray.size() != criteriaArray.size() || num > targetArray.size())
		{
			return null;
		}
		
		TreeMap<Double, ArrayList<T>> sTree = new TreeMap<Double, ArrayList<T>>();
		double key;
		ArrayList<T> val;
		
		for(int i = 0; i <  criteriaArray.size(); i++)
		{
			key = criteriaArray.get(i);
			
			if(isDecendingOrder) key *= -1.0;
			if(sTree.containsKey(key))
			{
				val = sTree.get(key);
				val.add(targetArray.get(i));
			}
			else
			{
				val = new ArrayList<T>();
				val.add(targetArray.get(i));
				sTree.put(key, val);
			}
		}
		
		ArrayList<T> rslt = new ArrayList<T>(criteriaArray.size());
		Collection<ArrayList<T>> vals = sTree.values();
		
		int cnt = 0;
		for(ArrayList<T> valObj : vals)
		{
			rslt.addAll(valObj);
			cnt += valObj.size();
			
			if(cnt >= num) break;
		}
		
		while(rslt.size() > num)
		{
			rslt.remove(rslt.size()-1);
		}
		
		rslt.trimToSize();
		return rslt;
	}
	
	/**
	 * @param <T>
	 * @param criteriaArray
	 * @param targetArray
	 * @param isDecendingOrder
	 * @param num
	 * @return
	 */
	private <T> ArrayList<T> sortArraysWithLinkedSorting(ArrayList<Double> criteriaArray, ArrayList<T> targetArray, boolean isDecendingOrder, int num)
	{
		if(criteriaArray == null || targetArray == null || targetArray.size() != criteriaArray.size() || num > targetArray.size())
		{
			return null;
		}
		
		Integer indexArray[] = new Integer[targetArray.size()];
		for(int i = 0; i < indexArray.length; i++)
		{
			indexArray[i] = i;
		}
		
		// This is always in ascending order
		Arrays.sort(indexArray, new LinkedComparator(LinkedComparatorType.index, criteriaArray));
		
		ArrayList<T> rslt = new ArrayList<T>(num);
		
		for(int i = 0; i < num; i++)
		{
			if(isDecendingOrder)
			{
				rslt.add(targetArray.get(indexArray[indexArray.length - 1- i]));
			}
			else
			{
				rslt.add(targetArray.get(indexArray[i]));
			}
		}
		
		return rslt;
	}
	
	@SuppressWarnings("unused")
	private <T> ArrayList<T> sortArraysWithLinkedSorting(ArrayList<Double> criteriaArray, ArrayList<T> targetArray, boolean isDecendingOrder)
	{
		return sortArraysWithLinkedSorting(criteriaArray, targetArray, isDecendingOrder, targetArray.size());
	}
}
