/**
 * SLUCE 2 Project
 * http://www.cscs.umich.edu/research/projects/sluce/
 */
package sluceii.agent;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Hashtable;

import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.random.RandomHelper;
import repast.simphony.valueLayer.GridValueLayer;
import sluceii.agent.decisionmaking.DeveloperDecisionHelper;
import sluceii.agent.decisionmaking.SimpleLandSearchHelper;
import sluceii.agent.decisionmaking.SimpleLandSearchHelper.SpatialSearchStrategyType;
import sluceii.app.lmm.ContextCreator;
import sluceii.landscape.*;
import sluceii.landscape.Parcel.ParcelDevelopStatus;
import sluceii.land.development.DeveloperSpecialtyParam;
import sluceii.land.development.EstimatedLandProfitInfo;
import sluceii.land.development.LandProfitInfoRecord;
import sluceii.land.development.SubdivisionType;
import sluceii.land.market.LandBidRecord;
import sluceii.land.market.LandListingRecord;
import sluceii.land.market.LandMarketListing;
import sluceii.land.market.SimpleMarketLand;
import sluceii.land.market.decision.LandProfitEstimator;
import sluceii.land.market.exchange.LandBuyingAction;
import sluceii.land.market.exchange.LandBuyingProcess;
import sluceii.land.market.exchange.LandSellingAction;
import sluceii.land.market.exchange.LandSellingProcess;
import sluceii.land.market.utility.UtilityStrategy;
import sluceii.params.AppParameters;
import sluceii.params.Constants;
import sluceii.params.ModelParameters;
import umn.geog.helia.landscape.LandscapeLayer;
import umn.geog.helia.landscape.LandscapeLayerGroup;
import umn.geog.helia.landscape.RasterLayer;
import umn.geog.helia.landscape.data.IntegerMatrix;

/**
 * Land developers buy farm land from rural land owners and divide it into lots
 * and sell to residential households. 
 * 
 * @author Shipeng Sun
 *
 */
public class Developer extends SimpleAgent implements LandBuyingProcess<Subdivision>, LandSellingProcess<Parcel>
{
	/**
	 * A Static reference to the parcel land market listing
	 * Developers are sellers on this market. Buyers are residential households.
	 */
	public static LandMarketListing<Parcel> parcelMarket;
	
	/**
	 * A Static reference to the land market listing
	 * Developers are buyers on this market. Rural land owners are sellers.
	 */
	public static LandMarketListing<Subdivision> subdivisionMarket;
	
	/**
	 * The strategy used by this agent to calculate his utility.
	 * Not used yet. Ideally, this should be used to calculated expected utilities under risks.
	 */
	private UtilityStrategy utilStrategy;
	
	/**
	 * Buying action object. It helps implement some buying process.
	 * This object only handles details with subdivision market. Other things must be processed by the agent.
	 */
	private LandBuyingAction<Subdivision,Developer> buyAction;
	
	
	/**
	 * Selling action object. It helps implement some selling process.
	 * Things other than parcel market must be taken care of by the agent.
	 */
	private LandSellingAction<Parcel,Developer> sellAction;
	
	/**
	 * An object used to estimated the expected profit from developing a farm/subdivision.
	 */
	private LandProfitEstimator landProfitEstimator;
	
	/**
	 * Parameters for all subdivision types. 
	 * This allows possibility of one developer doing multiple types.
	 */
	private ArrayList<DeveloperSpecialtyParam> developmentSpecialtyParams;
	
	/**
	 * The developer is likely to bid on this one.
	 */
	private Subdivision targetedLandUnit;
	
	/**
	 * Using this subdivision type to the targeted land unit
	 */
	private int targetedDevelopType;
	
	/**
	 * 
	 */
	private LandProfitInfoRecord targtedLandProfitRecord;

	/**
	 * if this is true, the ask price of sellers will be updated to the highest bid price to give
	 * buyers instant rejection.
	 */
	private boolean autoUpdateAskPrice;
	
	/**
	 * Estimated land profit by this developer for all farms/subdivisions on the market.
	 * A trade-off between memory and CPU time.
	 */
	protected EstimatedLandProfitInfo<Developer, LandUnit> landProfitEstimation;
	
	/**
	 * A comprehensive decision-making helper used to do many types of calculations.
	 */
	protected DeveloperDecisionHelper decisionHelper;
	
	/**
	 * Default constructor
	 */
	public Developer()
	{
		developmentSpecialtyParams = new ArrayList<DeveloperSpecialtyParam>();
		
		sellAction = new LandSellingAction<Parcel,Developer>();
		buyAction = new LandBuyingAction<Subdivision,Developer>();
		//landProfitEstimator = new LandProfitEstimator<Subdivision>();
		
		landProfitEstimation = new EstimatedLandProfitInfo<Developer, LandUnit>(this);
		autoUpdateAskPrice = true;
		
		decisionHelper = new DeveloperDecisionHelper();
		initDecisionHelper();
		
		targtedLandProfitRecord = null;
		targetedLandUnit = null;
	}

	/* (non-Javadoc)
	 * @see sluceii.agent.SimpleAgent#getAgentTypeCode()
	 */
	@Override
	public int getAgentTypeCode() 
	{
		return AgentTypeCode.landDeveloper.ordinal();
	}

	/**
	 * Initialize the decision helper
	 */
	private void initDecisionHelper() 
	{
		decisionHelper.setDeveloperAgent(this);
		decisionHelper.setResidentialLandTransportCoefficient(AppParameters.residentialLandTransportCostCoefficient);
		decisionHelper.setBorrowingInterest(0.06);
		decisionHelper.setSavingInterest(0.03);
		decisionHelper.setMaxStockNum(12);
	}

	/**
	 * @return the developmentSpecialtyParams
	 */
	public ArrayList<DeveloperSpecialtyParam> getDevelopmentSpecialtyParams() 
	{
		return developmentSpecialtyParams;
	}

	/**
	 * @param developmentSpecialtyParams the developmentSpecialtyParams to set
	 */
	public void setDevelopmentSpecialtyParams(ArrayList<DeveloperSpecialtyParam> developmentSpecialtyParams) 
	{
		this.developmentSpecialtyParams = developmentSpecialtyParams;
	}

	@Override
	public void acceptBidOffer(LandBidRecord<Parcel, ?> landBidRecord)
	{
		sellAction.acceptBidOffer(landBidRecord);
		ModelParameters.activity = true;
	}

	@Override
	public void clearTransaction(LandBidRecord<Parcel, ?> landBidRecord)
	{
		// As a seller to clear transaction
		sellAction.clearTransaction(landBidRecord);
		
		Parcel landObj = landBidRecord.getLandUnitObj();
		ContextCreator.contextAssist.writeStdOutMessage(" Developer " + this.agentID + " sold parcel " + landObj.getLandUnitID() + " at (" + landObj.getXCenterIndex() +"," + landObj.getYCenterIndex() + ")", 2, true);
	}

	@Override
	public void checkBidOffers()
	{
		sellAction.checkBidOffers();
	}

	@Override
	public void putLandOnMarket(Parcel parcelObj, double askPrice, double wta)
	{
		this.sellAction.putLandOnMarket(parcelObj, askPrice, wta, AppParameters.maxBidNums,RunEnvironment.getInstance().getCurrentSchedule().getTickCount());
	}
	
	@Override
	public void putLandOnMarket(Parcel landObj, double askPrice, double wta, ArrayList<Number> extraLandAttributes) 
	{
		this.sellAction.putLandOnMarket(landObj, askPrice, wta, AppParameters.maxBidNums, RunEnvironment.getInstance().getCurrentSchedule().getTickCount(), extraLandAttributes);	
	}

	@Override
	public void rejectBidOffer(LandBidRecord<Parcel, ?> landBidRecord)
	{
		sellAction.rejectBidOffer(landBidRecord);
	}


	@Override
	public void bidAccepted(LandBidRecord<Subdivision, ?> landBidRecord)
	{
		buyAction.bidAccepted(landBidRecord);
		
		Subdivision sObj = landBidRecord.getLandUnitObj();
		// Update the location of the agent
		SimpleAgent.agentGrid.moveTo(this, sObj.getXCenterIndex(), sObj.getYCenterIndex());
		super.setPosition(sObj.getXCenterIndex(), sObj.getYCenterIndex());
		ModelParameters.activity = true;
		
		ContextCreator.contextAssist.writeStdOutMessage(" Developer " + this.agentID + " purchased subidivision " + sObj.getLandUnitID() + " at (" + sObj.getXCenterIndex() +"," + sObj.getYCenterIndex() + ")", 2, true);
	}


	@Override
	public boolean bidOnLand(LandListingRecord<Subdivision, ?> listedLandObj, double utility, double bidPrice, double wtp)
	{
		if(autoUpdateAskPrice ) listedLandObj.setAskPrice(bidPrice);
		return this.buyAction.bidOnLand(listedLandObj, utility, bidPrice, wtp, RunEnvironment.getInstance().getCurrentSchedule().getTickCount());
	}

	@Override
	public boolean bidOnLand(LandListingRecord<Subdivision, ?> listedParcelObj, double utility, double bidPrice, double wtp, ArrayList<Number> extraLandAttri) 
	{
		// TODO Auto-generated method stub
		return false;
	}
	
	@Override
	public void bidRejected(LandBidRecord<Subdivision, ?> landBidRecord)
	{
		buyAction.bidRejected(landBidRecord);
		
	}

	@Override
	@ScheduledMethod(start = 1, interval = 1, shuffle = false, priority = ScheduleParameters.RANDOM_PRIORITY, duration = -1.0)
	public void step()
	{
		ContextCreator.contextAssist.writeStdOutMessage(" Developer Agent " + this.agentID + " step method...", 4, true);
		
		super.step();
		
		if(this.propertyList.size() > 1)
		{
			int stop = 1;
			stop += 0;
		}
		
		sellAction.setLandMarket(Developer.parcelMarket);
		sellAction.setSellerAgent(this);
		
		buyAction.setLandMarket(Developer.subdivisionMarket);
		buyAction.setBuyerAgent(this);
		
		this.buySubdivision();
		
		this.developSubdivision();
		
		this.buildStructure();
		
		this.sellParcel();
		
		this.checkBidOffers();
		
		ContextCreator.contextAssist.writeStdOutMessage(" Developer Agent " + this.agentID + " step method done", 4, true);
	}

	/**
	 * Evaluate and buy a farm/subdivision from the subdivison market
	 */
	private void buySubdivision() 
	{
		DeveloperDecisionHelper ddH = decisionHelper;
		
		this.landProfitEstimator = ddH;
		
		if(!ddH.isToBuy()) return;
		
		// Try to buy a subdivision from rural land owners / farmers
		boolean bidSuccess = false;
		int numTries = 0;
		int numSuccess = 0;
		Subdivision landObj = null;
		double landOwnerAskPrice;
		LandProfitInfoRecord piRec = null, piRecAlternative = null;
		double bidPrice, wtp;
		do
		{
			// This can also be implemented in search market listing. Do NOT return those parcels 
			// that have reached maximum allowed bids.
			ArrayList<LandListingRecord<Subdivision,?>> subdivList = searchMarketListing(2,AppParameters.numSearchSubdivision,0.0,this.getCapitalResource().getAssetQuantity(),this.landProfitEstimator);
			if(subdivList == null || subdivList.size() <= 0) return;
			
			if(subdivList.size()== 1)
			{
				for(LandListingRecord<Subdivision,?> candidateSubdiv : subdivList)
				{
					landObj = candidateSubdiv.getLandUnitObj();
					piRec = this.landProfitEstimation.getRecord(landObj);
					if(piRec == null || piRec.landProfit <= 0.0) continue; // not profitable
					
					landOwnerAskPrice = piRec.landCost;
					bidPrice = (float)(landOwnerAskPrice + piRec.landProfit*0.15);
					wtp = (float)(landOwnerAskPrice + piRec.landProfit*0.5);
					if(wtp <= candidateSubdiv.getWillingToAccept())
						continue; // Not possible to pay below WTA
					
					bidSuccess = bidOnLand(candidateSubdiv, piRec.landProfit,bidPrice, wtp);
					
					if(bidSuccess)
					{
						this.targetedLandUnit = landObj;
						this.targetedDevelopType = piRec.developTypeIndex;
						this.targtedLandProfitRecord = piRec;
						numSuccess++; // Only bid for one single parcel for one time
						ContextCreator.contextAssist.writeStdOutMessage(" Developer " + this.agentID + " bid on subdivision " + landObj.getLandUnitID() + " with $" + bidPrice, 2, true);
					}
					
					if(numSuccess >= AppParameters.maxBidCapacity)
					{
						break;
					}
				}
			}
			else if(subdivList.size()== 2)
			{
				LandListingRecord<Subdivision,?> candidateSubdiv = subdivList.get(1);
				landObj = candidateSubdiv.getLandUnitObj();
				piRecAlternative = this.landProfitEstimation.getRecord(landObj);
				
				candidateSubdiv = subdivList.get(0);
				landObj = candidateSubdiv.getLandUnitObj();
				piRec = this.landProfitEstimation.getRecord(landObj);
				
				if(piRec == null || piRecAlternative == null)
				{
					throw new NullPointerException("Null ProfitInfoRecord.");
				}
				
				landOwnerAskPrice = piRec.landCost;
				bidPrice = (float)(landOwnerAskPrice + piRec.landProfit*0.05);
				wtp = (float)Math.max(bidPrice,(landOwnerAskPrice + (piRec.landProfit - piRecAlternative.landProfit)*0.95));
				
				bidSuccess = bidOnLand(candidateSubdiv, piRec.landProfit,bidPrice, wtp);
				
				if(bidSuccess)
				{
					this.targetedLandUnit = landObj;
					this.targetedDevelopType = piRec.developTypeIndex;
					this.targtedLandProfitRecord = piRec;
					numSuccess++; // Only bid for one single parcel for one time
					ContextCreator.contextAssist.writeStdOutMessage(" Developer " + this.agentID + " bid on subdivision " + landObj.getLandUnitID() + " with $" + bidPrice, 2, true);
				}
				
				if(numSuccess >= AppParameters.maxBidCapacity)
				{
					break;
				}
				
			}

		} while (!bidSuccess && numTries++ < AppParameters.maxBiddingTries);
		
	}

	private ArrayList<LandListingRecord<Subdivision, ?>> searchMarketListing(int targetedNumber, int totalSearchNumber,double minProfit,double resourceQuantity,LandProfitEstimator landAppraiser) 
	{
		// How many parcels/houses the buyer want to evaluate?
		SimpleLandSearchHelper<Subdivision,Developer> searchHelper = new SimpleLandSearchHelper<Subdivision,Developer>(RandomHelper.getUniform());
		searchHelper.setHousingSearcher(this);
		searchHelper.setLandMarket(Developer.subdivisionMarket);
		searchHelper.setUtilStrategy(utilStrategy);
		searchHelper.setSpatialSearchStrategy(SpatialSearchStrategyType.random);
		
		ArrayList<LandListingRecord<Subdivision,?>> parcelList = searchHelper.searchProperties(targetedNumber, totalSearchNumber,minProfit,resourceQuantity,landAppraiser);
		
		return parcelList;
	}

	/**
	 * Develop the subdivision bought.
	 * Using predefined subdivision type, divide the subdivision/farm to multiple parcel lots 
	 */
	private void developSubdivision() 
	{
		if(targetedLandUnit == null || this.getPropertyList().size() == 0) return;
		
		Subdivision landObj = targetedLandUnit;
		
		// This means the transaction has not been accomplished yet.
		if(!getPropertyList().contains(landObj)) return;
		
		DeveloperSpecialtyParam devType = developmentSpecialtyParams.get(targetedDevelopType);
		SubdivisionType landType = SubdivisionType.subdivisionTypeCollections.get(targetedDevelopType);
		
		ArrayList<Integer> parcelIDs = landObj.getLowerLeverlIDList();
		if(parcelIDs == null || parcelIDs.size() != 1)
		{
			throw new ClassCastException("Parcel Object expected.");
		}
		
		Parcel pObj = landObj.getLowerLevelUnits().get(parcelIDs.get(0));
		
		//int parcelSize = (int) landType.getMeanSize();
		//ArrayList<Parcel> newParcels = subdivideParcel(pObj, parcelSize);
		String templateName = landType.getSubdivisionTypeName();
		int ss = 0;
		if(templateName == "Rural Lot")
			ss  = ss + 0;
		ArrayList<Parcel> newParcels = subdivideParcel(pObj, SubdivisionTemplateHelper.getTemplate(templateName + " PID"),landType.getSubdivisionTypeID(),true);
		if(newParcels == null || newParcels.size() <= 0) return;
		
		this.removeProperty(landObj);
		for(Parcel paObj : newParcels)
		{
			this.addProperty(paObj);
			paObj.setParcelStatus(ParcelDevelopStatus.developed);
			LandCoverDataHelper.addToParcelList(paObj);
		}
		
		this.getCapitalResource().subtractAsset(devType.developCost);
		
		targetedLandUnit = null;
	}
	
	/**
	 * Divide a parcel into smaller ones. This parcel is exactly the same as a farm/subdivision.
	 * In other words, the farm/subdivision only has one single sub-level land unit object.
	 * 
	 * @param pObj
	 * @param parcelSize
	 * @return
	 */
	@SuppressWarnings({ "unchecked", "unused" })
	private ArrayList<Parcel> subdivideParcel(Parcel pObj, int parcelSize) 
	{
		int parentParcelSize = pObj.getNumOfCells();
		ArrayList<Parcel> rsltParcels = new ArrayList<Parcel>();
		ArrayList<Integer> cellIDList = pObj.getCellIDList();
		int cellID;
		
		LandProfitInfoRecord piRec = this.landProfitEstimation.getRecord(this.targetedLandUnit);
		
		// Change the current bid parcel from rural to urban use
		RasterLayer<IntegerMatrix> parcelIDLayer = (RasterLayer<IntegerMatrix>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.parcelIDLayerIndex);
		RasterLayer<?> luTypeLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.landUseTypeLayerIndex);
		RasterLayer<?> tpcLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.finalPriceLayerIndex);
		GridValueLayer vlLUType = (GridValueLayer)SimpleAgent.context.getValueLayer(Constants.LANDUSETYPE);
		
		if(parcelSize == 1)
		{
			for(Integer cellID1 : cellIDList)
			{
				Parcel newParcelObj = new Parcel(true);
				newParcelObj.addCellID(cellID1);
				rsltParcels.add(newParcelObj);
				
				// Update the parcel ID layer
				newParcelObj.putValue(parcelIDLayer, newParcelObj.getLandUnitID());
				newParcelObj.putValue(luTypeLayer, parcelSize);
				newParcelObj.putValue(tpcLayer, piRec.landCost);
				newParcelObj.putValue(vlLUType, luTypeLayer);
				
				newParcelObj.updateCoordinates(parcelIDLayer);
			}
		}
		else if(parentParcelSize == 9)// 3 by 3 parcel
		{
			Collections.sort(cellIDList);
			
			if(parcelSize == 2)
			{
				// Four new parcels
				// # 1, 2, 3, 4
				for(int iP = 0; iP < 4; iP++)
				{
					Parcel newParcelObj = new Parcel(true);
					rsltParcels.add(newParcelObj);
					
					if(iP == 3) // #4
					{
						cellID = cellIDList.get(6);
						newParcelObj.addCellID(cellID);
						cellID = cellIDList.get(7);
						newParcelObj.addCellID(cellID);
					}
					else // # 1,2,3,4
					{
						cellID = cellIDList.get(iP);
						newParcelObj.addCellID(cellID);
						cellID = cellIDList.get(iP+3);
						newParcelObj.addCellID(cellID);
					}
					
					// Update the parcel ID layer
					newParcelObj.putValue(parcelIDLayer, newParcelObj.getLandUnitID());
					newParcelObj.putValue(luTypeLayer, parcelSize);
					newParcelObj.putValue(tpcLayer, piRec.landCost);
					newParcelObj.putValue(vlLUType, luTypeLayer);
					
					newParcelObj.updateCoordinates(parcelIDLayer);
				}
			}
			else if(parcelSize == 3)
			{
				// Three new parcels
				for(int iP = 0; iP < 3; iP++)
				{
					Parcel newParcelObj = new Parcel(true);
					rsltParcels.add(newParcelObj);
					
					cellID = cellIDList.get(iP);
					newParcelObj.addCellID(cellID);
					cellID = cellIDList.get(iP+3);
					newParcelObj.addCellID(cellID);
					cellID = cellIDList.get(iP+6);
					newParcelObj.addCellID(cellID);
					
					// Update the parcel ID layer
					newParcelObj.putValue(parcelIDLayer, newParcelObj.getLandUnitID());
					newParcelObj.putValue(luTypeLayer, parcelSize);
					newParcelObj.putValue(tpcLayer, piRec.landCost);
					newParcelObj.putValue(vlLUType, luTypeLayer);
					newParcelObj.updateCoordinates(parcelIDLayer);
				}				
			}
			else if(parcelSize == 4)
			{
				// two new parcels
				for(int iP = 0; iP < 2; iP++)
				{
					Parcel newParcelObj = new Parcel(true);
					rsltParcels.add(newParcelObj);
					
					if(iP == 0)
					{
						cellID = cellIDList.get(iP);
						newParcelObj.addCellID(cellID);
						cellID = cellIDList.get(iP+3);
						newParcelObj.addCellID(cellID);
						cellID = cellIDList.get(iP+6);
						newParcelObj.addCellID(cellID);
						cellID = cellIDList.get(iP+7);
						newParcelObj.addCellID(cellID);
					}
					else
					{
						cellID = cellIDList.get(iP);
						newParcelObj.addCellID(cellID);
						cellID = cellIDList.get(iP+1);
						newParcelObj.addCellID(cellID);
						cellID = cellIDList.get(iP+4);
						newParcelObj.addCellID(cellID);
						cellID = cellIDList.get(iP+7);
						newParcelObj.addCellID(cellID);
					}
	
					// Update the parcel ID layer
					newParcelObj.putValue(parcelIDLayer, newParcelObj.getLandUnitID());
					newParcelObj.putValue(luTypeLayer, parcelSize);
					newParcelObj.putValue(tpcLayer, piRec.landCost);
					newParcelObj.putValue(vlLUType, luTypeLayer);
					newParcelObj.updateCoordinates(parcelIDLayer);
				}
			}
		}
		else if(parentParcelSize == 16)
		{
			return null;
		}
		
		return rsltParcels;
	}

	/**
	 * Divide the given parcel/subdivision using a template
	 * @param pObj
	 * @param templateObj
	 * @return A series of new parcel objects
	 */
	private ArrayList<Parcel> subdivideParcel(Parcel pObj, RasterLayer<?> templateObj, int subdivisionTypeID)
	{
		if(templateObj == null) throw new NullPointerException("Null Template Object");
		
		ArrayList<Parcel> rsltParcels = new ArrayList<Parcel>();
		ArrayList<Integer> cellIDList = pObj.getCellIDList();
		int cellID;
		
		LandProfitInfoRecord piRec = this.landProfitEstimation.getRecord(this.targetedLandUnit);
		
		// Change the current bid parcel from rural to urban use
		@SuppressWarnings("unchecked")
		RasterLayer<IntegerMatrix> parcelIDLayer = (RasterLayer<IntegerMatrix>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.parcelIDLayerIndex);
		RasterLayer<?> luTypeLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.landUseTypeLayerIndex);
		RasterLayer<?> lcTypeLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.landCoverTypeLayerIndex);
		RasterLayer<?> sdTypeLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.subdivisionTypeLyaerIndex);
		RasterLayer<?> tpcLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.finalPriceLayerIndex);
		GridValueLayer vlLUType = (GridValueLayer)SimpleAgent.context.getValueLayer(Constants.LANDUSETYPE);
		
		// First, get the number of parcel numbers in the template
		int tW = templateObj.getCols(); // templateWidth
		int tH = templateObj.getRows();	// templateHeight	
		int templateParcelID; // the parcel ID in the template
		ArrayList<Integer> templateParcelIDs = new ArrayList<Integer>(); // the list of parcel IDs in the template

		for(int i = 0; i < tW; i++)
		{
			for(int j = 0; j < tH; j++)
			{
				templateParcelID = templateObj.getValueAt(i, j).intValue();
				if(templateParcelIDs.contains(templateParcelID))
				{
					continue;
				}
				else if(templateParcelID > 0)
				{
					templateParcelIDs.add(templateParcelID);
				}
			}
		}
		
		int templateParcelsNum = templateParcelIDs.size();
		if(templateParcelsNum <= 0) return null;
		
		ArrayList<Integer> startPos;
		int x, y;
		ArrayList<Integer> cellInTemplate = new ArrayList<Integer>();

		while(cellIDList != null && cellIDList.size() > 0)
		{
			Integer cellIDInt = cellIDList.get(0);
			cellInTemplate.clear();
			
			// Determine if the cellIDInt and subsequent cells can hold a template
			startPos = LandCoverDataHelper.cellID2XY(cellIDInt.intValue());

			for(int i = 0; i < tW; i++)
			{
				for(int j = 0; j < tH; j++)
				{
					if(templateObj.getValueAt(i, j).intValue() > 0)
					{
						x = startPos.get(0).intValue() + i;
						y = startPos.get(1).intValue() + j;
						cellID = LandCoverDataHelper.XY2CellID(x, y);
						cellInTemplate.add(cellID);
					}
				}
			}
			
			// the possible template can be contained in the parcel then it is good.
			if(cellIDList.containsAll(cellInTemplate))
			{
				// Create parcels using the template
				Hashtable<Integer, Parcel> newParcels = new Hashtable<Integer,Parcel>();
				Parcel newParcelObj = null;
				for(int i = 0; i < tW; i++)
				{
					for(int j = 0; j < tH; j++)
					{
						if(templateObj.getValueAt(i,j).intValue() <= 0) continue;

						x = startPos.get(0).intValue() + i;
						y = startPos.get(1).intValue() + j;
						cellID = LandCoverDataHelper.XY2CellID(x, y);
						templateParcelID = templateObj.getValueAt(i, j).intValue();
						
						if(newParcels.containsKey(templateParcelID))
						{
							newParcelObj = newParcels.get(templateParcelID);
							newParcelObj.addCellID(cellID);
						}
						else
						{
							newParcelObj = new Parcel(true);
							rsltParcels.add(newParcelObj);
							newParcelObj.addCellID(cellID);
							
							newParcels.put(templateParcelID, newParcelObj);
						}
					}
				}
				
				Object[] parcels = newParcels.values().toArray();
				for(int k = 0; k < parcels.length; k++)
				{
					newParcelObj = (Parcel) parcels[k];
					newParcelObj.putValue(parcelIDLayer, newParcelObj.getLandUnitID());
					newParcelObj.putValue(luTypeLayer, 1);
					newParcelObj.putValue(sdTypeLayer,subdivisionTypeID);
					//newParcelObj.putValue(tpcLayer, piRec.landCost);
					newParcelObj.putValue(vlLUType, sdTypeLayer);
					newParcelObj.updateCoordinates(parcelIDLayer);
				}

				cellIDList.removeAll(cellInTemplate);
			}
			else
			{
				cellIDList.remove(cellIDInt);
			}
		}
		
		return rsltParcels;
	}
	
	/**
	 * Divide the given parcel/subdivision using a template
	 * @param pObj
	 * @param templateObj
	 * @return A series of new parcel objects
	 */
	private ArrayList<Parcel> subdivideParcel(Parcel pObj, RasterLayer<?> templateObj, int subdivisionTypeID, boolean multiOrientations)
	{
		if(!multiOrientations) return subdivideParcel(pObj, templateObj,subdivisionTypeID);
		if(templateObj == null) throw new NullPointerException("Null Template Object");
		
		ArrayList<Parcel> rsltParcels = new ArrayList<Parcel>();
		ArrayList<Integer> cellIDList = pObj.getCellIDList();
		int cellID;
		
		LandProfitInfoRecord piRec = this.landProfitEstimation.getRecord(this.targetedLandUnit);
		
		// Change the current bid parcel from rural to urban use
		@SuppressWarnings("unchecked")
		RasterLayer<IntegerMatrix> parcelIDLayer = (RasterLayer<IntegerMatrix>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.parcelIDLayerIndex);
		RasterLayer<?> luTypeLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.landUseTypeLayerIndex);
		RasterLayer<?> lcTypeLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.landCoverTypeLayerIndex);
		RasterLayer<?> sdTypeLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.subdivisionTypeLyaerIndex);
		RasterLayer<?> tpcLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.finalPriceLayerIndex);
		GridValueLayer vlLUType = (GridValueLayer)SimpleAgent.context.getValueLayer(Constants.LANDUSETYPE);
		
		// First, get the number of parcel numbers in the template
		int tW = templateObj.getCols(); // templateWidth
		int tH = templateObj.getRows();	// templateHeight	
		int templateParcelID; // the parcel ID in the template
		ArrayList<Integer> templateParcelIDs = new ArrayList<Integer>(); // the list of parcel IDs in the template

		for(int i = 0; i < tW; i++)
		{
			for(int j = 0; j < tH; j++)
			{
				templateParcelID = templateObj.getValueAt(i, j).intValue();
				if(templateParcelIDs.contains(templateParcelID))
				{
					continue;
				}
				else if(templateParcelID > 0)
				{
					templateParcelIDs.add(templateParcelID);
				}
			}
		}
		
		int templateParcelsNum = templateParcelIDs.size();
		if(templateParcelsNum <= 0) return null;
		
		ArrayList<Integer> startPos;
		int x, y, numOrientations = 4;
		ArrayList<Integer> cellInTemplate = null;
		ArrayList<ArrayList<Integer>> cellInTemplateArray = new ArrayList<ArrayList<Integer>>();
		for(int k = 0; k < numOrientations; k++)
		{
			ArrayList<Integer> cellInTemplate000 = new ArrayList<Integer>();
			cellInTemplateArray.add(cellInTemplate000);
			
			ArrayList<Integer> cellInTemplate180 = new ArrayList<Integer>();
			cellInTemplateArray.add(cellInTemplate180);
		}
		
		double orientationVal = 0;
		int loopStartID = -1;
		
		while(cellIDList != null && cellIDList.size() > 0)
		{
			Integer cellIDInt = cellIDList.get(0);
			if(cellIDInt == loopStartID) break;
			
			cellInTemplate = null;
			
			for(int k = 0; k < numOrientations*2; k++)
			{
				cellInTemplateArray.get(k).clear();
			}
			
			// Determine if the cellIDInt and subsequent cells can hold a template
			startPos = LandCoverDataHelper.cellID2XY(cellIDInt.intValue());
			if(startPos.get(0).intValue() == 0) cellIDInt += 0;

			for(int i = 0; i < tW; i++)
			{
				for(int j = 0; j < tH; j++)
				{
					if(templateObj.getValueAt(i,j).intValue() > 0)
					{
						for(int k = 0; k < numOrientations; k++)
						{
							double angle = k*(2*Math.PI/numOrientations);

							// x' = x*cosine(alpha) - y*sine(alpha); y' = y*cosine(alpha) + x*sine(alpha)
							x = (int) Math.round(startPos.get(0).intValue() + i*Math.cos(angle) - j*Math.sin(angle));
							y = (int) Math.round(startPos.get(1).intValue() + j*Math.cos(angle) + i*Math.sin(angle));
							cellID = LandCoverDataHelper.XY2CellID(x, y);
							cellInTemplateArray.get(2*k + 0).add(cellID);
							
							// mirror by Y axis
							x = (int) Math.round(startPos.get(0).intValue() + i*Math.cos(angle) - j*Math.sin(angle));
							y = (int) Math.round(startPos.get(1).intValue() - j*Math.cos(angle) - i*Math.sin(angle));
							cellID = LandCoverDataHelper.XY2CellID(x, y);
							cellInTemplateArray.get(2*k + 1).add(cellID);
						}
					}
				}
			}
			
			boolean mirror = false;
			
			for(int k = 0; k < numOrientations*2; k++)
			{
				if(cellIDList.containsAll(cellInTemplateArray.get(k)))
				{
					cellInTemplate = cellInTemplateArray.get(k);
					orientationVal = (int)(k/2)*(2*Math.PI/numOrientations);
					if((k/2.0 - Math.floor(k/2.0)) > 0.01) mirror = true;
					break;
				}
			}

			// the possible template can be contained in the parcel then it is good.
			if(cellInTemplate != null)
			{
				// Create parcels using the template
				Hashtable<Integer, Parcel> newParcels = new Hashtable<Integer,Parcel>();
				Parcel newParcelObj = null;
				for(int i = 0; i < tW; i++)
				{
					for(int j = 0; j < tH; j++)
					{
						if(templateObj.getValueAt(i,j).intValue() <= 0) continue;
						
						// x' = x*cosine(alpha) - y*sine(alpha); y' = y*cosine(alpha) + x*sine(alpha)
						x = (int) Math.round(startPos.get(0).intValue() + i*Math.cos(orientationVal) - j*Math.sin(orientationVal));
						
						if(mirror)
							y = (int) Math.round(startPos.get(1).intValue() - j*Math.cos(orientationVal) - i*Math.sin(orientationVal));
						else
							y = (int) Math.round(startPos.get(1).intValue() + j*Math.cos(orientationVal) + i*Math.sin(orientationVal));
						
						cellID = LandCoverDataHelper.XY2CellID(x, y);
						templateParcelID = templateObj.getValueAt(i, j).intValue();
						
						if(newParcels.containsKey(templateParcelID))
						{
							newParcelObj = newParcels.get(templateParcelID);
							newParcelObj.addCellID(cellID);
						}
						else
						{
							newParcelObj = new Parcel(true);
							rsltParcels.add(newParcelObj);
							newParcelObj.addCellID(cellID);
							
							newParcels.put(templateParcelID, newParcelObj);
						}
					}
				}
				
				Object[] parcels = newParcels.values().toArray();
				for(int k = 0; k < parcels.length; k++)
				{
					newParcelObj = (Parcel) parcels[k];
					newParcelObj.putValue(parcelIDLayer, newParcelObj.getLandUnitID());
					newParcelObj.putValue(luTypeLayer, 1);
					newParcelObj.putValue(sdTypeLayer,subdivisionTypeID);
					//newParcelObj.putValue(tpcLayer, piRec.landCost);
					newParcelObj.putValue(vlLUType, sdTypeLayer);
					newParcelObj.updateCoordinates(parcelIDLayer);
				}

				cellIDList.removeAll(cellInTemplate);
				loopStartID = -1;
			}
			else
			{
				if(cellIDList.remove(cellIDInt))// should not simply remove this one, others might need it.
				{
					cellIDList.add(cellIDInt); // add it to the tail
					if(loopStartID == -1) 
						loopStartID = cellIDInt;
				}
			}
		}
		
		return rsltParcels;
	}
	
	/**
	 * Divide the given parcel/subdivision using a group of templates.
	 * This is essentially similar to the famous TETRIS game. Using blocks to fill a space with as little as open space.
	 * TODO: try all possible templates for each location.
	 * @param pObj
	 * @param templateObj
	 * @return A series of new parcel objects
	 */
	@SuppressWarnings({"unused", "unchecked"})
	private ArrayList<Parcel> subdivideParcel(Parcel pObj, LandscapeLayerGroup templateObjs, double rrientationStepAngle)
	{
		if(templateObjs == null) throw new NullPointerException("Null Template Object");
		
		ArrayList<Parcel> rsltParcels = new ArrayList<Parcel>();
		ArrayList<Integer> cellIDList = pObj.getCellIDList();
		int cellID;
		
		LandProfitInfoRecord piRec = this.landProfitEstimation.getRecord(this.targetedLandUnit);
		
		// Change the current bid parcel from rural to urban use
		RasterLayer<IntegerMatrix> parcelIDLayer = (RasterLayer<IntegerMatrix>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.parcelIDLayerIndex);
		RasterLayer<?> luTypeLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.landUseTypeLayerIndex);
		RasterLayer<?> tpcLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.finalPriceLayerIndex);
		GridValueLayer vlLUType = (GridValueLayer)SimpleAgent.context.getValueLayer(Constants.LANDUSETYPE);
		
		LandscapeLayer[] templateLayers = templateObjs.getFastLayersArray();
		if(templateLayers.length <= 0) return null;
		
		RasterLayer<IntegerMatrix> templateObj = (RasterLayer<IntegerMatrix>) templateLayers[0];
		
		// First, get the number of parcel numbers in the template
		int tW = templateObj.getCols(); // templateWidth
		int tH = templateObj.getRows();	// templateHeight	
		int templateParcelID; // the parcel ID in the template
		ArrayList<Integer> templateParcelIDs = new ArrayList<Integer>(); // the list of parcel IDs in the template

		for(int i = 0; i < tW; i++)
		{
			for(int j = 0; j < tH; j++)
			{
				templateParcelID = templateObj.getValueAt(i, j).intValue();
				if(templateParcelIDs.contains(templateParcelID))
				{
					continue;
				}
				else if(templateParcelID > 0)
				{
					templateParcelIDs.add(templateParcelID);
				}
			}
		}
		
		int templateParcelsNum = templateParcelIDs.size();
		if(templateParcelsNum <= 0) return null;
		
		ArrayList<Integer> startPos;
		int x, y, numOrientations = 4;
		ArrayList<Integer> cellInTemplate = null;
		ArrayList<ArrayList<Integer>> cellInTemplateArray = new ArrayList<ArrayList<Integer>>();
		for(int k = 0; k < numOrientations; k++)
		{
			ArrayList<Integer> cellInTemplate000 = new ArrayList<Integer>();
			cellInTemplateArray.add(cellInTemplate000);
		}
		
		double orientationVal = 0;
		
		while(cellIDList != null && cellIDList.size() > 0)
		{
			Integer cellIDInt = cellIDList.get(0);
			
			cellInTemplate = null;
			
			for(int k = 0; k < numOrientations; k++)
			{
				cellInTemplateArray.get(k).clear();
			}
			
			// Determine if the cellIDInt and subsequent cells can hold a template
			startPos = LandCoverDataHelper.cellID2XY(cellIDInt.intValue());

			for(int i = 0; i < tW; i++)
			{
				for(int j = 0; j < tH; j++)
				{
					if(templateObj.getValueAt(i,j).intValue() > 0)
					{
						for(int k = 0; k < numOrientations; k++)
						{
							double angle = k*(2*Math.PI/numOrientations);

							// x' = x*cosine(alpha) - y*sine(alpha); y' = y*cosine(alpha) + x*sine(alpha)
							x = (int) (startPos.get(0).intValue() + i*Math.cos(angle) - j*Math.sin(angle));
							y = (int) (startPos.get(1).intValue() + j*Math.cos(angle) + i*Math.sin(angle));
							cellID = LandCoverDataHelper.XY2CellID(x, y);
							cellInTemplateArray.get(k).add(cellID);
						}
					}
				}
			}
			
			for(int k = 0; k < numOrientations; k++)
			{
				if(cellIDList.containsAll(cellInTemplateArray.get(k)))
				{
					cellInTemplate = cellInTemplateArray.get(k);
					orientationVal = k*(2*Math.PI/numOrientations);
					break;
				}
			}

			// the possible template can be contained in the parcel then it is good.
			if(cellInTemplate != null)
			{
				// Create parcels using the template
				Hashtable<Integer, Parcel> newParcels = new Hashtable<Integer,Parcel>();
				Parcel newParcelObj = null;
				for(int i = 0; i < tW; i++)
				{
					for(int j = 0; j < tH; j++)
					{
						if(templateObj.getValueAt(i,j).intValue() <= 0) continue;

						// x' = x*cosine(alpha) - y*sine(alpha); y' = y*cosine(alpha) + x*sine(alpha)
						x = (int) (startPos.get(0).intValue() + i*Math.cos(orientationVal) - j*Math.sin(orientationVal));
						y = (int) (startPos.get(1).intValue() + j*Math.cos(orientationVal) + i*Math.sin(orientationVal));
						cellID = LandCoverDataHelper.XY2CellID(x, y);
						templateParcelID = templateObj.getValueAt(i, j).intValue();
						
						if(newParcels.containsKey(templateParcelID))
						{
							newParcelObj = newParcels.get(templateParcelID);
							newParcelObj.addCellID(cellID);
						}
						else
						{
							newParcelObj = new Parcel(true);
							rsltParcels.add(newParcelObj);
							newParcelObj.addCellID(cellID);
							
							newParcels.put(templateParcelID, newParcelObj);
						}
					}
				}
				
				Object[] parcels = newParcels.values().toArray();
				for(int k = 0; k < parcels.length; k++)
				{
					newParcelObj = (Parcel) parcels[k];
					newParcelObj.putValue(parcelIDLayer, newParcelObj.getLandUnitID());
					newParcelObj.putValue(luTypeLayer, newParcelObj.getNumOfCells());
					newParcelObj.putValue(tpcLayer, piRec.landCost);
					newParcelObj.putValue(vlLUType, luTypeLayer);
					newParcelObj.updateCoordinates(parcelIDLayer);
				}

				cellIDList.removeAll(cellInTemplate);
			}
			else
			{
				cellIDList.remove(cellIDInt);
			}
			
		}
		
		return rsltParcels;
	}
	
	/**
	 * Build housing structure on the parcels just developed.
	 */
	private void buildStructure() 
	{
		DeveloperSpecialtyParam devType = developmentSpecialtyParams.get(targetedDevelopType);

		int optimalNum = Integer.MAX_VALUE;
		int num = 0;
		
		for(SimpleMarketLand luObj : this.getPropertyList())
		{
			if(luObj instanceof Parcel)
			{
				if(((Parcel)luObj).getParcelStatus() == ParcelDevelopStatus.developed)
				{
					((Parcel)luObj).setParcelStatus(ParcelDevelopStatus.built);
					this.getCapitalResource().subtractAsset(devType.buildCost);
					num++;
				}
			}
			
			if(num >= optimalNum) break;
		}
	}

	/**
	 * Build housing structure on the parcels just developed.
	 */
	@SuppressWarnings("unused")
	private void buildStructure(int desiredNum) 
	{
		DeveloperSpecialtyParam devType = developmentSpecialtyParams.get(targetedDevelopType);

		int num = 0;
		
		for(SimpleMarketLand luObj : this.getPropertyList())
		{
			if(luObj instanceof Parcel)
			{
				if(((Parcel)luObj).getParcelStatus() == ParcelDevelopStatus.developed)
				{
					((Parcel)luObj).setParcelStatus(ParcelDevelopStatus.built);
					this.getCapitalResource().subtractAsset(devType.buildCost);
					num++;
				}
			}
			
			if(num >= desiredNum) break;
		}
	}
		
	/**
	 * Sell the developed parcel
	 */
	private void sellParcel() 
	{
		DeveloperDecisionHelper ddH = this.decisionHelper;
		LandProfitInfoRecord piRec = this.targtedLandProfitRecord;

		if(piRec == null)
		{
			return;
		}

		Parcel pObj = null;
		
		for(SimpleMarketLand luObj : this.getPropertyList())
		{
			if(luObj instanceof Parcel)
			{
				pObj = (Parcel)luObj;
				
				if(pObj.getParcelStatus()!= ParcelDevelopStatus.built) continue;
				if(!ddH.isToSell(pObj)) continue;
				
				// Put the parcel on the parcel/housing market
				this.putLandOnMarket(pObj,piRec.salePrice, piRec.salePrice);
				pObj.setParcelStatus(ParcelDevelopStatus.inbidding);
			}
		}
	}	
	
	/**
	 * Add a land profit estimation record to the data storage.
	 * 
	 * @param luObj
	 * @param piRec
	 */
	public void addLandProfitEstimationRecord(LandUnit luObj, LandProfitInfoRecord piRec)
	{
		landProfitEstimation.addRecord(luObj, piRec);
	}


}
