/**
 * SLUCE 2 Project
 * http://www.cscs.umich.edu/research/projects/sluce/
 */
package sluceii.agent.econestimator;

import java.util.ArrayList;

import org.apache.commons.math.stat.regression.OLSMultipleLinearRegression;

import sluceii.agent.SimpleAgent;
import sluceii.land.market.SimpleMarketLand;
import sluceii.land.market.decision.LandPriceEstimator;
import sluceii.landscape.LandUnit;
import sluceii.landscape.LandUnit.LandAggregateMethod;
import sluceii.landscape.Parcel;
import umn.geog.helia.landscape.CompoundCell;
import umn.geog.helia.landscape.LandscapeLayerGroup;
import umn.geog.helia.landscape.RasterLayer;

/**
 * @author Shipeng
 *
 *	This class uses simple linear regression to estimate housing price.
 *  Its functionality will be mainly implemented using the Apache Commons Math Library 
 *  at http://commons.apache.org/math/
 *  
 *  More java options at http://math.nist.gov/javanumerics/
 */
public class LinearRegressionPriceEstimator extends LandPriceEstimator<Parcel>
{

	/**
	 * The layer group with Y and X variables.
	 * The control layer will indicate where to get the values.
	 */
	protected LandscapeLayerGroup varLayers;
	
	/**
	 * The control layer
	 */
	protected RasterLayer<?> controlLayer;
	
	/**
	 * Estimated coefficients
	 */
	protected double[] bHat;
	
	/**
	 * the OLS regression object
	 */
	protected OLSMultipleLinearRegression OLSObj;
	
	/**
	 * Does the regression equation have a constant term?
	 */
	protected boolean hasConstantTerm;
	
	/**
	 * A reference to the transaction price layer
	 */
	protected RasterLayer<?> transactionPriceLayer;
	
	/**
	 * 
	 */
	protected RasterLayer<?> landUnitIDLayer;
	
	/**
	 * 
	 */
	public LinearRegressionPriceEstimator()
	{
		// TODO Auto-generated constructor stub
		OLSObj = new OLSMultipleLinearRegression();
	}
	
	/**
	 * @return the transactionPriceLayer
	 */
	public RasterLayer<?> getTransactionPriceLayer()
	{
		return transactionPriceLayer;
	}

	/**
	 * @param transactionPriceLayer the transactionPriceLayer to set
	 */
	public void setTransactionPriceLayer(RasterLayer<?> transactionPriceLayer)
	{
		this.transactionPriceLayer = transactionPriceLayer;
	}
	
	/**
	 * @return the landUnitIDLayer
	 */
	public RasterLayer<?> getLandUnitIDLayer()
	{
		return landUnitIDLayer;
	}

	/**
	 * @param landUnitIDLayer the landUnitIDLayer to set
	 */
	public void setLandUnitIDLayer(RasterLayer<?> landUnitIDLayer)
	{
		this.landUnitIDLayer = landUnitIDLayer;
	}
	/**
	 * Load X and Y data from the varLayers controlled by controlLayer.
	 * Only values in the control layer that are greater than the minVal and smaller than
	 * the maxVal will be included in the variables X and Y.
	 * 
	 * minVal < ... < maxVal
	 * 
	 * @param minVal
	 * @param maxVal
	 * @param addConstant
	 * @return success or failure
	 */
	public boolean LoadDataFraomLayers(double minVal, double maxVal, boolean hasConstant)
	{
		if(this.controlLayer == null || this.varLayers == null || this.varLayers.getDirectLayersNum() < 2)
		{
			return false;
		}
		
		this.hasConstantTerm = hasConstant;
		
		ArrayList<Double> allVars = new ArrayList<Double>();
		double curControlVal;
		CompoundCell curVals;
		int nObs = 0;
		
		for(int i = 0; i < controlLayer.getCols(); i++)
		{
			for(int j=0; j < controlLayer.getRows(); j++)
			{
				curControlVal = controlLayer.getValueAt(i, j).doubleValue();
				
				if(curControlVal > minVal && curControlVal < maxVal)
				{
					
					curVals = varLayers.getLayersValue(i, j);

					// Y
					allVars.add(curVals.get(0).doubleValue());
					
					// Constant terms
					if(hasConstantTerm)
					{
						allVars.add(1.0);
					}
					
					// X
					for(int k = 1; k < curVals.size(); k++)
					{
						allVars.add(curVals.get(k).doubleValue());
					}
					
					nObs++;
				}
				
			}
		}
		
		double[] dat = new double[allVars.size()];
		for(int m = 0; m < allVars.size(); m++) dat[m] = allVars.get(m);
		OLSObj.newSampleData(dat, nObs, varLayers.getDirectLayersNum()-1+(hasConstantTerm?1:0));
		
		return true;
	}
	
	public void LoadSampleData(double[] y, double[][] x)
	{
		OLSObj.newSampleData(y, x);
	}
	
	public void LoadSampleData(double[] data, int nobs, int nvars)
	{
		OLSObj.newSampleData(data, nobs, nvars);
	}
	
	public void EstimateCoefficients()
	{
		bHat = OLSObj.estimateRegressionParameters();
	}
	/**
	 * @param agent
	 */
	public LinearRegressionPriceEstimator(SimpleAgent agent)
	{
		super(agent);
		// TODO Auto-generated constructor stub
	}

	/* (non-Javadoc)
	 * @see sluceii.agent.decisionmaking.LandPriceEstimator#estimatedLandUnitPrice(sluceii.landscape.LandUnit)
	 */
	@Override
	public double estimateLandUnitPrice(SimpleMarketLand luObj)
	{
		// TODO Auto-generated method stub
		if(bHat == null) this.EstimateCoefficients();
		double rsltVal = Double.NaN;
		
		CompoundCell ptVals = ((LandUnit) luObj).getValue(this.varLayers, LandAggregateMethod.average);
		
		if(hasConstantTerm)
		{
			if(ptVals.size() == bHat.length)
			{
				rsltVal = 0.0;
				ptVals.put(0, 1.0);
				for(int i = 0 ; i < ptVals.size(); i++)
				{
					rsltVal += bHat[i]*ptVals.get(i).doubleValue();
				}
			}
		}
		else
		{
			if(ptVals.size() == (bHat.length+1))
			{
				rsltVal = 0.0;
				ptVals.put(0, 1.0);
				for(int i = 0 ; i < bHat.length; i++)
				{
					rsltVal += bHat[i]*ptVals.get(i+1).doubleValue();
				}
			}
		}
		
		return rsltVal;
	}

}
