/**
 * SLUCE 2 Project
 * http://www.cscs.umich.edu/research/projects/sluce/
 */
package sluceii.agent.econestimator;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;

import sluceii.agent.SimpleAgent;
import sluceii.land.market.LandBidRecord;
import sluceii.land.market.LandBidRecord.BidRecordStatus;
import sluceii.land.market.SimpleMarketLand;
import sluceii.land.market.decision.LandPriceEstimator;
import umn.geog.helia.landscape.RasterLayer;
import sluceii.landscape.LandUnit;
import sluceii.landscape.Parcel;

/**
 * @author Shipeng Sun
 *
 */
public class NeighborhoodAveragePriceEstimator extends LandPriceEstimator<Parcel>
{
	/**
	 * The size of the neighborhood
	 */
	protected int neighborhoodRadius;
	
	
	/**
	 * In the case of distance decay of weight, the coefficient of decay.
	 * The weight for a Land Unit with a distance of d is exp(-d*decayCoefficient).
	 */
	protected double decayCoefficient;
	
	/**
	 * The hash table of a land unit object list
	 */
	Hashtable<Integer, ? extends LandUnit> luObjList;
	
	/**
	 * A reference to the transaction price layer
	 */
	protected RasterLayer<?> transactionPriceLayer;
	
	/**
	 * 
	 */
	protected RasterLayer<?> landUnitIDLayer;
	

	/**
	 * 
	 */
	public NeighborhoodAveragePriceEstimator()
	{
		decayCoefficient = 0.0d;
		neighborhoodRadius = 1;
	}

	/**
	 * @param agent
	 */
	public NeighborhoodAveragePriceEstimator(SimpleAgent agent)
	{
		super(agent);
	}

	/**
	 * @return the neighborhoodRadius
	 */
	public int getNeighborhoodRadius()
	{
		return neighborhoodRadius;
	}

	/**
	 * @param neighborhoodRadius the neighborhoodRadius to set
	 */
	public void setNeighborhoodRadius(int neighborhoodRadius)
	{
		this.neighborhoodRadius = neighborhoodRadius;
	}

	/**
	 * @return the decayCoefficient
	 */
	public double getDecayCoefficient()
	{
		return decayCoefficient;
	}

	/**
	 * @param decayCoefficient the decayCoefficient to set
	 */
	public void setDecayCoefficient(double decayCoefficient)
	{
		this.decayCoefficient = decayCoefficient;
	}

	/**
	 * @return the luObjList
	 */
	public Hashtable<Integer, ? extends LandUnit> getLuObjList()
	{
		return luObjList;
	}

	/**
	 * @param luObjList the luObjList to set
	 */
	public void setLuObjList(Hashtable<Integer, ? extends LandUnit> luObjList)
	{
		this.luObjList = luObjList;
	}

	/**
	 * @return the transactionPriceLayer
	 */
	public RasterLayer<?> getTransactionPriceLayer()
	{
		return transactionPriceLayer;
	}

	/**
	 * @param transactionPriceLayer the transactionPriceLayer to set
	 */
	public void setTransactionPriceLayer(RasterLayer<?> transactionPriceLayer)
	{
		this.transactionPriceLayer = transactionPriceLayer;
	}
	
	/**
	 * @return the landUnitIDLayer
	 */
	public RasterLayer<?> getLandUnitIDLayer()
	{
		return landUnitIDLayer;
	}

	/**
	 * @param landUnitIDLayer the landUnitIDLayer to set
	 */
	public void setLandUnitIDLayer(RasterLayer<?> landUnitIDLayer)
	{
		this.landUnitIDLayer = landUnitIDLayer;
	}	
	
	/* (non-Javadoc)
	 * @see sluceii.agent.decisionmaking.LandPriceEstimator#estimatedLandUnitPrice(sluceii.landscape.LandUnit)
	 */
	@Override
	public double estimateLandUnitPrice(SimpleMarketLand luObj)
	{
		switch(priceSource)
		{
			case finalTransaction:
			{
				// Use the final transaction price layer to calculate the estimated price
				int xCenter = luObj.getXCenterIndex();
				int yCenter = luObj.getYCenterIndex();
				
				Integer id = -1;
				Double price = 0.0d;
				
				HashMap<Integer,Double> luIDValues = new HashMap<Integer,Double>();
				luIDValues.clear();
				
				// Go through the neighborhood
				for(int x = xCenter-neighborhoodRadius; x < xCenter+this.neighborhoodRadius; x++)
				{
					for(int y = yCenter-this.neighborhoodRadius; y < yCenter+this.neighborhoodRadius; y++)
					{
						if(x >= 0 && x < landUnitIDLayer.getCols() && y >=0 && y < landUnitIDLayer.getRows())
						{
							if((x-xCenter)*(x-xCenter)+(y-yCenter)*(y-yCenter) <= neighborhoodRadius*neighborhoodRadius*2)
							{
								// Push the price and land unit id into a dictionary to avoid repeated accounting
								id = landUnitIDLayer.getValueAt(x, y).intValue();
								price = transactionPriceLayer.getValueAt(x, y).doubleValue();
								
								// This indicates that the parcel is not 
								if(id < 0 || price <=0) continue;
								
								luIDValues.put(landUnitIDLayer.getValueAt(x, y).intValue(), transactionPriceLayer.getValueAt(x, y).doubleValue());
							}
						}
					}
				}
				
				double totalWght = 0.0d;
				double totalWghtedPrices = 0.0d;
				
				Iterator<Integer> it = luIDValues.keySet().iterator();
				LandUnit neighborLUObj = null;
				double dist = 0.0d;
				
				while(luObjList != null && it.hasNext())
				{
					id = it.next();
					price = luIDValues.get(id);
					neighborLUObj = luObjList.get(id);
					
					if(neighborLUObj != null)
					{
						dist = ((LandUnit) luObj).distanceToInCell(neighborLUObj);
						totalWghtedPrices += price*Math.exp(-decayCoefficient*dist);
						totalWght += Math.exp(-decayCoefficient*dist);
					}
				}
				
				// In this case, no transaction is found in the neighborhood.
				if(totalWghtedPrices == 0.0d) break;
				
				return totalWghtedPrices/totalWght;
				
			}
			case bidRecords: // Use bidding records to get the transaction price and history
			{
				// Use the final transaction price layer to calculate the estimated price
				int xCenter = luObj.getXCenterIndex();
				int yCenter = luObj.getYCenterIndex();
				
				HashMap<Integer,Double> luIDValues = new HashMap<Integer,Double>();
				luIDValues.clear();
				
				double luPrice = 0.0d;
				ArrayList<LandBidRecord<Parcel,?>> bidRecords;
				LandBidRecord<Parcel,?> curRecord;
				
				// Go through the neighborhood
				for(int x = xCenter-neighborhoodRadius; x < xCenter+this.neighborhoodRadius; x++)
				{
					for(int y = yCenter-this.neighborhoodRadius; y < yCenter+this.neighborhoodRadius; y++)
					{
						if(landUnitIDLayer.isInBoundary(x,y))
						{
							if((x-xCenter)*(x-xCenter)+(y-yCenter)*(y-yCenter) <= neighborhoodRadius*neighborhoodRadius*2)
							{
								// Push the price and land unit id into a dictionary to avoid repeated accounting
								
								bidRecords = landMarket.getBidRecordsByLandUnitID(landUnitIDLayer.getValueAt(x, y).intValue());
								if(bidRecords == null) continue;
								
								for(int k = bidRecords.size()-1; k >=0; k--)
								{
									curRecord = bidRecords.get(k);
									if(curRecord.getRecordStatus() == BidRecordStatus.completed)
									{
										luPrice = curRecord.getBidPrice();
									}
								}
								
								luIDValues.put(landUnitIDLayer.getValueAt(x, y).intValue(), luPrice);
							}
						}
					}
				}
				
				double totalWght = 0.0d;
				double totalWghtedPrices = 0.0d;
				
				Iterator<Integer> it = luIDValues.keySet().iterator();
				Integer id = -1;
				Double price = 0.0d;
				LandUnit neighborLUObj = null;
				double dist = 0.0d;
				
				while(it.hasNext())
				{
					id = it.next();
					price = luIDValues.get(id);
					neighborLUObj = luObjList.get(id);
					
					if(luObjList != null)
					{
						dist = ((LandUnit) luObj).distanceToInCell(neighborLUObj);
						totalWghtedPrices += price*Math.exp(-decayCoefficient*dist);
						totalWght += Math.exp(-decayCoefficient*dist);
					}
				}
				
				// In this case, no transaction is found in the neighborhood.
				if(totalWghtedPrices == 0.0d) break;
				
				return totalWght/totalWghtedPrices;
			}
			default:
			{
				
			}
		}
		
		// Didn't find anything
		return Double.NaN;
	}

}
