/**
 * SLUCE 2 Project
 * http://www.cscs.umich.edu/research/projects/sluce/
 */
package sluceii.agent;

import java.util.ArrayList;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.random.RandomHelper;
import repast.simphony.valueLayer.GridValueLayer;
import sluceii.agent.decisionmaking.ResidentialHouseholdDecisionHelper;
import sluceii.agent.decisionmaking.SimpleLandSearchHelper;
import sluceii.agent.decisionmaking.SimpleLandSearchHelper.SpatialSearchStrategyType;
import sluceii.agent.econestimator.SimplePriceEstimator;
import sluceii.app.lmm.ContextCreator;
import sluceii.params.AppParameters;
import sluceii.params.Constants;
import sluceii.params.ModelParameters;
import sluceii.land.management.LandManagementStrategy;
import sluceii.land.market.LandBidRecord;
import sluceii.land.market.LandListingRecord;
import sluceii.land.market.LandMarketListing;
import sluceii.land.market.SimpleMarketLand;
import sluceii.land.market.decision.LandPriceEstimator;
import sluceii.land.market.decision.LandPriceEstimator.PriceSourceType;
import sluceii.land.market.exchange.LandBuyingAction;
import sluceii.land.market.exchange.LandBuyingProcess;
import sluceii.land.market.exchange.LandSellingAction;
import sluceii.land.market.exchange.LandSellingProcess;
import sluceii.land.market.utility.CobbDouglas;
import sluceii.land.market.utility.UtilityStrategy;
import sluceii.landscape.LandCoverDataHelper;
import sluceii.landscape.LandUnit;
import sluceii.landscape.Parcel;
import umn.geog.helia.landscape.RasterLayer;

/**
 * This class implements Residential Household.
 * This is one of the most complicated agents in SLUCE2. Residential households 
 * buy and sell land/houses. They also manage their land and consequently impact
 * the carbon measures.
 * 
 * @author Shipeng Sun
 *
 */
public class ResidentialHousehold extends SimpleAgent implements LandBuyingProcess<Parcel>, LandSellingProcess<Parcel>
{
	/**
	 * The utilities when the household acquired the properties.
	 * These values can be used to determine if an agent want to search for new dwellings.
	 */
	protected ArrayList<Double> initPropertyUtilitis;
	
	/**
	 * If the utility drops more than this number, the agent will begin to consider relocation.
	 */
	protected double utilDropPtgRelocationTrigger;
	
	/**
	 * The index of the primary residence of the property list.
	 * Homestead residence/house. Decision to sell this will be relocation.
	 * Other properties would not cause relocation.
	 */
	protected int primaryResidenceIndex;

	/**
	 * How many and which parcels/properties this agent has bid on.
	 * Currently, buyers can only bid on one single land/parcel/property.
	 */
	protected ArrayList<LandBidRecord<Parcel,?>> bidRecordList;
	
	/**
	 * A Static reference to the land market listing
	 */
	public static LandMarketListing<Parcel> landMarket;
	
	/**
	 * The pressure to buy a new property. This value can be used to control how
	 * many agents get into the market at each step.
	 */
	private float pressureToBuy;
	
	/**
	 * The pressure to sell a property. This value can be used to control how
	 * many agents leave the market at each step. 
	 */
	private float pressureToSell;
	
	/**
	 * The strategy used by this agent to calculate his utility
	 * 
	 */
	private UtilityStrategy utilStrategy;
	
	/**
	 * Buying action object. It helps implement some buying process.
	 * This object only handles details with land market. Other things must be processed by the agent.
	 */
	private LandBuyingAction<Parcel,ResidentialHousehold> buyAction;
	
	/**
	 * Selling action object. It helps implement some selling process.
	 * Things other than land market must be taken care of by the agent.
	 */
	private LandSellingAction<Parcel,ResidentialHousehold> sellAction;
	
	/**
	 * A land buying/selling/relocation decision helper for residential household
	 */
	protected ResidentialHouseholdDecisionHelper decisionHelper;
	
	/**
	 * Default constructor
	 */
	public ResidentialHousehold()
	{
		// Initialize the property list
		super();
		this.initPropertyUtilitis = new ArrayList<Double>();
		this.bidRecordList = new ArrayList<LandBidRecord<Parcel,?>>();
		
		pressureToSell = 0.0f;
		pressureToBuy = 1.0f;
		
		buyAction = new LandBuyingAction<Parcel,ResidentialHousehold>();
		buyAction.setBuyerAgent(this);
		buyAction.setLandMarket(ResidentialHousehold.landMarket);
		
		sellAction = new LandSellingAction<Parcel,ResidentialHousehold>();
		sellAction.setSellerAgent(this);
		sellAction.setLandMarket(ResidentialHousehold.landMarket);
		
		this.utilStrategy = new CobbDouglas();
		this.primaryResidenceIndex = 0;
		
		this.decisionHelper = new ResidentialHouseholdDecisionHelper();
		this.decisionHelper.setHouseholdAgent(this);
		
		// Initialize a price estimator for the agent. It is used to form
		// price expectation.
		//NeighborhoodAveragePriceEstimator priceEstimator = new NeighborhoodAveragePriceEstimator();
		SimplePriceEstimator priceEstimator = new SimplePriceEstimator();

		priceEstimator.setAgent(this);
		priceEstimator.setPriceSource(PriceSourceType.finalTransaction);
		priceEstimator.setLandUnitIDLayer(referenceLayer);
		priceEstimator.setLandMarket(landMarket);
		
//		priceEstimator.setDecayCoefficient(0.5);
//		priceEstimator.setNeighborhoodRadius(3);
//		priceEstimator.setLuObjList(LandCoverDataHelper.parcelList);		
		
		RasterLayer<?> tpLayer = (RasterLayer<?>) LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.finalPriceLayerIndex);
		priceEstimator.setTransactionPriceLayer(tpLayer);
		
		RasterLayer<?> luIDLayer = (RasterLayer<?>) LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.parcelIDLayerIndex);
		priceEstimator.setLandUnitIDLayer(luIDLayer);
		
		priceEstimator.setLandMarket(landMarket);
		
		this.decisionHelper.setLandAppraiser(priceEstimator);
		
		// Home owner will not re-sale even if the utility drops to zero.
		utilDropPtgRelocationTrigger = 0.20;
	}

	/**
	 * @return the initPropertyUtilitis
	 */
	public ArrayList<Double> getInitPropertyUtilitis()
	{
		return initPropertyUtilitis;
	}

	/**
	 * @param initPropertyUtilitis the initPropertyUtilitis to set
	 */
	public void setInitPropertyUtilitis(ArrayList<Double> initPropertyUtilitis)
	{
		this.initPropertyUtilitis = initPropertyUtilitis;
	}

	/**
	 * @return the landMarket
	 */
	public static LandMarketListing<Parcel> getLandMarket()
	{
		return landMarket;
	}

	public static ArrayList<String> getExtraBidFieldNames()
	{
		ArrayList<String> fields = new ArrayList<String>();
		fields.add("dist");
		fields.add("openSpace");
		fields.add("tranCost");
		fields.add("houseBudget");
		fields.add("nonHousingU");
		return fields;
	}
	
	/**
	 * @param landMarket the landMarket to set
	 */
	public static void setLandMarket(LandMarketListing<Parcel> landMarket)
	{
		ResidentialHousehold.landMarket = landMarket;
	}

	/**
	 * @return the pressureToBuy
	 */
	public float getPressureToBuy()
	{
		return pressureToBuy;
	}

	/**
	 * @param pressureToBuy the pressureToBuy to set
	 */
	public void setPressureToBuy(float pressureToBuy)
	{
		this.pressureToBuy = pressureToBuy;
	}

	/**
	 * @return the pressureToSell
	 */
	public float getPressureToSell()
	{
		return pressureToSell;
	}

	/**
	 * @param pressureToSell the pressureToSell to set
	 */
	public void setPressureToSell(float pressureToSell)
	{
		this.pressureToSell = pressureToSell;
	}

	/**
	 * @return the utilStrategy
	 */
	public UtilityStrategy getUtilStrategy()
	{
		return utilStrategy;
	}

	/**
	 * @param utilStrategy the utilStrategy to set
	 */
	public void setUtilStrategy(UtilityStrategy utilStrategy)
	{
		this.utilStrategy = utilStrategy;
	}

	/**
	 * @return the primaryResidenceIndex
	 */
	public int getPrimaryResidenceIndex()
	{
		return primaryResidenceIndex;
	}

	/**
	 * @param primaryResidenceIndex the primaryResidenceIndex to set
	 */
	public void setPrimaryResidenceIndex(int primaryResidenceIndex)
	{
		this.primaryResidenceIndex = primaryResidenceIndex;
	}

	/**
	 * @return the utilDropPtgRelocationTrigger
	 */
	public double getUtilDropPtgRelocationTrigger()
	{
		return utilDropPtgRelocationTrigger;
	}

	/**
	 * @param utilDropPtgRelocationTrigger the utilDropPtgRelocationTrigger to set
	 */
	public void setUtilDropPtgRelocationTrigger(double utilDropPtgRelocationTrigger)
	{
		this.utilDropPtgRelocationTrigger = utilDropPtgRelocationTrigger;
	}

	/* (non-Javadoc)
	 * @see sluceii.agent.SimpleAgent#getAgentTypeCode()
	 */
	@Override
	public int getAgentTypeCode() 
	{
		return AgentTypeCode.residentialHousehold.ordinal();
	}

	@Override
	@ScheduledMethod(start = 1, interval = 1, shuffle = false, priority = ScheduleParameters.LAST_PRIORITY, duration = -1.0)
	public void step()
	{
		//ContextCreator.contextAssist.writeStdOutMessage(" Household Agent " + this.agentID + " step method...", 2, true);

		super.step();

		sellAction.setLandMarket(ResidentialHousehold.landMarket);
		buyAction.setLandMarket(ResidentialHousehold.landMarket);

		this.sellLand();
		this.buyLand();
		
		this.updateLandManagementStrategies();
		
		ContextCreator.contextAssist.writeStdOutMessage(" Household Agent " + this.agentID + " step method done", 3, true);
	}
	
	/**
	 * A household agent updates his/her land management strategies according to their neighbors
	 */
	public void updateLandManagementStrategies() 
	{
		if(propertyList == null || propertyList.size() <= 0) return;
		if(primaryResidenceIndex >= propertyList.size()) return;
		
		SimpleMarketLand luObj = propertyList.get(primaryResidenceIndex);
		if(luObj == null) return;
		
		if(luObj instanceof Parcel)
		{
			Parcel pObj = (Parcel)luObj;
			RasterLayer<?> pIDLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.parcelIDLayerIndex);
			java.util.HashSet<Integer> ngbrIDs = LandCoverDataHelper.findNeighbors(pObj, pIDLayer);

			int curIndex = LandManagementStrategy.strategyCollections.indexOf(this.getLandManagementStrategies());
			if(curIndex >= 0) 
				pObj.putValue((GridValueLayer)SimpleAgent.context.getValueLayer(Constants.MANAGEMENTTYPE), curIndex + 1);
			else
				curIndex += 0;
				
			ArrayList<Integer> counts = new ArrayList<Integer>(4);
			for(int i = 0; i < LandManagementStrategy.strategyCollections.size(); i++)
			{
				counts.add(0);
			}
			int sum = 0;
			
			for(Integer nID : ngbrIDs)
			{
				if(SimpleAgent.landOwnership.containsKey(nID))
				{
					SimpleAgent ngbrAgent = SimpleAgent.landOwnership.get(nID);
					if(ngbrAgent instanceof ResidentialHousehold)
					{
						int index = LandManagementStrategy.strategyCollections.indexOf(ngbrAgent.getLandManagementStrategies());
						if(index == -1) continue;
						counts.set(index,counts.get(index)+1);
						sum += 1;
					}
				}
			}
			
			if(sum <= 0) return;
			
			// Choose the one with most popular strategy, but over 0.5
			int maxCount = 0, maxI = -1;
			for(int i = (counts.size()-1); i >= 0 ; i--)
			{
				if(counts.get(i) > maxCount)
				{
					maxCount = counts.get(i);
					maxI = i;
				}
			}
			
			if(sum >0 && maxCount > sum*0.5 && maxI >= 0)
			{
				this.setLandManagementStrategies(LandManagementStrategy.strategyCollections.get(maxI));
				pObj.putValue((GridValueLayer)SimpleAgent.context.getValueLayer(Constants.MANAGEMENTTYPE), maxI + 1);
			}
		}
	}

	/**
	 * Sell Land parcel/houses
	 */
	private synchronized void sellLand()
	{
		// There are multiple routes here. 
		//ResidentialHouseholdDecisionHelper decisionHelper = new ResidentialHouseholdDecisionHelper();
		//decisionHelper.setHouseholdAgent(this);
		double askPrice, wta;
		
		// Java for collection loop
		// TODO: lower the ask price after each unsuccessful selling
		// For sellers, lower the price; for buyers, increase their bids.
		// In the current model, the bidding price is primarily based on income and 
		// utility. It does not have much space for this.
		synchronized (propertyList) 
		{
			for(int luIndex = 0; luIndex < this.propertyList.size(); luIndex++)
			{
				LandUnit luObj = (LandUnit) this.propertyList.get(luIndex);
				
				LandListingRecord<Parcel,?> landListed = landMarket.getListingRecordByLandUnitID(luObj.getLandUnitID());
	
				if(landListed == null)
				{
					// Check if the seller want to put it on the market
					if(decisionHelper.isToSell(luObj))
					{
						this.pressureToBuy = 0.0f;
						// Put this land unit/parcel on the market
						askPrice = decisionHelper.calcAskPrice(luObj);
						wta = decisionHelper.calcWTAPrice(luObj);
						putLandOnMarket((Parcel)luObj,askPrice, wta);
					}
				}
				else
				{
					this.checkBidOffers();
				}
			}
		}		
	}
	
	/**
	 * Buy Land.
	 * 
	 * The method is synchronized because the property list is a shared resource and could be changed
	 * by multiple methods. sellLand() is synchronized because of the same reason.
	 */
	private synchronized void buyLand()
	{
		if(this.propertyList.size() == 0 && decisionHelper.isToBuy())
		{
			boolean bidSuccess = false;
			int numTries = 0;
			int numSuccess = 0;
			ArrayList<Number> utilityAndPrice; 
			ArrayList<Number> extraLandAttri = new ArrayList<Number>();
			do
			{
				// This can also be implemented in search market listing. Do NOT return those parcels 
				// that have reached maximum allowed bids.
				ArrayList<LandListingRecord<Parcel,?>> parcelList = this.searchMarketListing(1,AppParameters.numSearchParcel);
				if(parcelList != null && parcelList.size()>0)
				{
					for(LandListingRecord<Parcel,?> candidateParcel : parcelList)
					{
						utilityAndPrice = decisionHelper.calcUtilityAndPricesForPurchase(candidateParcel.getLandUnitObj());
						if(utilityAndPrice == null) continue;
						
						extraLandAttri.clear();
						extraLandAttri.addAll(utilityAndPrice.subList(3, utilityAndPrice.size()));
						
						bidSuccess = bidOnLand(candidateParcel, utilityAndPrice.get(0).floatValue(),utilityAndPrice.get(1).floatValue(), utilityAndPrice.get(2).floatValue(), extraLandAttri);
						
						if(bidSuccess)
						{
							numSuccess++; // Only bid for one single parcel for one time
						}
						
						if(numSuccess >= AppParameters.maxBidCapacity)
						{
							break;
						}
					}
				}

			} while (!bidSuccess && numTries++ < AppParameters.maxBiddingTries);
		}
		else
		{
			// Increase the year that this household in current status
		}		
	}
	
	/**
	 * This method implements the search process for a potential home buyer.
	 * 
	 * TODO: How to search the vacancy list. 
	 * Do we assume complete information or incomplete information?
	 * 
	 * @param number the number returned 
	 * @return A List of candidate parcels for bidding
	 */
	protected ArrayList<LandListingRecord<Parcel,?>> searchMarketListing(int targetNumber, int sampleNum )
	{
		// How many parcels/houses the buyer want to evaluate?
		SimpleLandSearchHelper<Parcel,ResidentialHousehold> searchHelper = new SimpleLandSearchHelper<Parcel,ResidentialHousehold>(RandomHelper.getUniform());
		searchHelper.setHousingSearcher(this);
		searchHelper.setLandMarket(ResidentialHousehold.landMarket);
		searchHelper.setUtilStrategy(utilStrategy);
		ArrayList<LandListingRecord<Parcel,?>> parcelList = searchHelper.searchProperties(targetNumber, sampleNum);
		
		return parcelList;
	}

	/**
	 * This method implements the search process for a potential home buyer.
	 * 
	 * TODO: How to search the vacancy list. 
	 * Do we assume complete information or incomplete information?
	 * 
	 * @param totalSearchNumber the total number of land units to be searched
	 * @param minUtil 
	 * @param resourceQuantity
	 * @param landAppraiser
	 * @return A List of candidate parcels for bidding
	 */
	public ArrayList<LandListingRecord<Parcel,?>> searchMarketListing(int totalSearchNumber,double minUtil,double resourceQuantity,LandPriceEstimator<Parcel> landAppraiser)
	{
		// How many parcels/houses the buyer want to evaluate?
		SimpleLandSearchHelper<Parcel,ResidentialHousehold> searchHelper = new SimpleLandSearchHelper<Parcel,ResidentialHousehold>(RandomHelper.getUniform());
		searchHelper.setHousingSearcher(this);
		searchHelper.setLandMarket(landMarket);
		searchHelper.setUtilStrategy(utilStrategy);
		searchHelper.setSpatialSearchStrategy(SpatialSearchStrategyType.random);
		
		ArrayList<LandListingRecord<Parcel,?>> parcelList = searchHelper.searchProperties(totalSearchNumber,minUtil,resourceQuantity,landAppraiser);
		
		return parcelList;
	}
	
	
	@Override
	public synchronized boolean bidOnLand(LandListingRecord<Parcel,?> listedParcelObj,double utility, double bidPrice, double wtp)
	{
		return buyAction.bidOnLand(listedParcelObj, utility, bidPrice, wtp, RunEnvironment.getInstance().getCurrentSchedule().getTickCount());
	}
	
	@Override	
	public synchronized boolean bidOnLand(LandListingRecord<Parcel,?> listedParcelObj,double utility, double bidPrice, double wtp, ArrayList<Number> extraLandAttri)
	{
		return buyAction.bidOnLand(listedParcelObj, utility, bidPrice, wtp, RunEnvironment.getInstance().getCurrentSchedule().getTickCount(), extraLandAttri);
	}	
	/**
	 * @param e
	 * @return
	 */
	public boolean addBidRecord(LandBidRecord<Parcel, ?> e)
	{
		return this.bidRecordList.add(e);
	}

	@Override
	public void bidRejected(LandBidRecord<Parcel,?> landBidRecord)
	{
		// As a buyer
		buyAction.bidRejected(landBidRecord);
		this.pressureToBuy = 1.0f;
	}

	@Override
	public void bidAccepted(LandBidRecord<Parcel,?> landBidRecord)
	{
		// as a buyer buyer
		this.pressureToBuy = 0.0f;
		
		buyAction.bidAccepted(landBidRecord);
		initPropertyUtilitis.add(landBidRecord.getBuyerUtility());
		finalizeBidAction(landBidRecord);
	}

	
	/**
	 * Finalize the transaction as a buyer
	 * @param landBidRecord
	 */
	public void finalizeBidAction(LandBidRecord<Parcel,?> landBidRecord)
	{
		Parcel pObj = landBidRecord.getListingRecord().getLandUnitObj();
		
		// Change the current bid parcel from rural to urban use
		//RasterLayer<?> luTypeLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.landUseTypeLayerIndex);
		//pObj.putValue(luTypeLayer, 1);
		
		RasterLayer<?> tpcLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.finalPriceLayerIndex);
		pObj.putValue(tpcLayer, landBidRecord.getBidPrice());
		
		//GridValueLayer vlLUType = (GridValueLayer)SimpleAgent.context.getValueLayer(Constants.LANDUSETYPE);
		//pObj.putValue(vlLUType, tpcLayer);
				
		// Update the location of the agent
		agentGrid.moveTo(this, pObj.getXCenterIndex(),pObj.getYCenterIndex());
		this.setPosition(pObj.getXCenterIndex(), pObj.getYCenterIndex());
		
		this.pressureToBuy = 0.0f;

		ModelParameters.activity = true;	
		ModelParameters.totalTansactionNum++;
		ContextCreator.contextAssist.writeStdOutMessage(" Household Agent " + this.agentID + " purchased parcel " + pObj.getLandUnitID() + " and moved to (" + pObj.getXCenterIndex() + "," + pObj.getYCenterIndex() + ")", 2, true);
	}

	@Override
	public void putLandOnMarket(Parcel parcelObj, double askPrice, double wta)
	{
		sellAction.putLandOnMarket(parcelObj, askPrice, wta, AppParameters.maxBidNums,RunEnvironment.getInstance().getCurrentSchedule().getTickCount());
		//landMarket.addParcelOnMarket(parcelObj, this, askPrice, wta, AppParameters.maxBidNums);
		this.pressureToSell = 0.0f;
	}
	
	/**
	 * See if the properties are listed too long. If the time is over a predefined time,
	 * dump the properties to land broker.
	 */
	public void checkListedTime()
	{
		LandUnit curLandObj = null;
		LandListingRecord<Parcel, ?> listingRecObj = null;
		double dumpPeriod = 10;

		for(int i = 0; i < getPropertyList().size(); i++)
		{
			curLandObj = (LandUnit) getPropertyList().get(i);
			listingRecObj = landMarket.getListingRecordByLandUnitID(curLandObj.getLandUnitID());

			// See if need to lower the asking price and wta price
			double listedTime = listingRecObj.getListedTime();
			double curTime = RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
			
			// For every wait period, the seller will lower the price.
			if( (curTime - listedTime) > dumpPeriod)
			{
				((LandBroker)landMarket.getRandomBroker()).acceptPrivateBid(listingRecObj, listingRecObj.getWillingToAccept());
			}
		}		
	}

	@Override
	public void checkBidOffers()
	{
		sellAction.checkBidOffers(AppParameters.waitPeriodBeforePriceDrop,true,AppParameters.priceDropRateHH,AppParameters.priceDropRateHH, RunEnvironment.getInstance().getCurrentSchedule().getTickCount());
	}	
	
	@Override
	public void acceptBidOffer(LandBidRecord<Parcel,?> landBidRecord)
	{
		sellAction.acceptBidOffer(landBidRecord);
	}
	
	@Override
	public void rejectBidOffer(LandBidRecord<Parcel,?> landBidRecord)
	{
		sellAction.rejectBidOffer(landBidRecord);
	}

	/* (non-Javadoc)
	 * @see sluceii.land.exchange.LandSellingProcess#clearTransaction(sluceii.land.market.LandBidRecord)
	 */
	@Override
	public void clearTransaction(LandBidRecord<Parcel,?> landBidRecord)
	{
		// Clear the transaction as a seller
		// Must remove the initial utility first before getting rid of the property.
		this.removeInitUtility(landBidRecord.getLandUnitObj());
		sellAction.clearTransaction(landBidRecord);
		this.pressureToBuy = 1.0f;
		
		// Now s/he is homeless, stay at the center of the landscape
		agentGrid.moveTo(this, AppParameters.worldXCenter,AppParameters.worldYCenter);
		this.setPosition(AppParameters.worldXCenter,AppParameters.worldYCenter);
		Parcel pObj = landBidRecord.getLandUnitObj();
		ContextCreator.contextAssist.writeStdOutMessage(" Household Agent " + this.agentID + " sold parcel " + pObj.getLandUnitID() + " and moved to (" + this.getX() + "," + this.getY() + ")", 2, true);
	}
	
	/**
	 * @param landObj
	 * @return the initial utility of the specified land unit.
	 */
	public double getInitUtility(SimpleMarketLand landObj)
	{
		if(this.initPropertyUtilitis == null || this.initPropertyUtilitis.size() == 0)
		{
			return 0;
		}
		else
		{
			int i = this.propertyList.indexOf(landObj);
			if(i < 0) return 0;
			
			return this.initPropertyUtilitis.get(i);
		}
	}
	
	/**
	 * @param landObj
	 * @return the initial utility of the specified land unit.
	 */
	public boolean removeInitUtility(LandUnit landObj)
	{
		if(this.initPropertyUtilitis == null || this.initPropertyUtilitis.size() == 0)
		{
			return false;
		}
		else
		{
			int i = this.propertyList.indexOf(landObj);
			if(i < 0) return false;
			
			initPropertyUtilitis.remove(i);
			return true;
		}
	}

	@Override
	public void putLandOnMarket(Parcel landObj, double askPrice, double wta,ArrayList<Number> extraLandAttributes) 
	{
	}
}
