package sluceii.agent;

import java.util.ArrayList;
import java.util.Hashtable;

import repast.simphony.annotate.AgentAnnot;
import repast.simphony.context.Context;
import repast.simphony.space.grid.Grid;
import sluceii.land.management.LandManagementStrategy;
import sluceii.land.market.SimpleMarketAgent;
import sluceii.land.market.SimpleMarketLand;
import sluceii.land.market.resource.AssetResource;
import sluceii.land.market.resource.AssetResource.AssetType;
import sluceii.landscape.LandUnit;
import umn.geog.helia.landscape.RasterLayer;

/**
 * @author Derek Robinson
 * Modified by Shipeng Sun
 *
 */
@AgentAnnot(displayName = "SimpleAgent")
public class SimpleAgent implements SimpleMarketAgent
{
	public enum AgentTypeCode {generic, ruralLandOwner, landDeveloper, residentialHousehold, landBroker};
	
	/**
	 * A reference to the context
	 */
	public static Context<SimpleAgent> context;
	
	/**
	 * A reference to the agent grid, through which agents can be found.
	 */
	public static Grid<SimpleAgent> agentGrid;
	
	/**
	 * Keep a dictionary of land unit ID and agent to enable query of land owners
	 */
	public static Hashtable<Integer, SimpleAgent> landOwnership = new Hashtable<Integer, SimpleAgent>();

	/**
	 * A unique identifier of the agent
	 */
	protected int agentID;
	
	/**
	 * the indices of X, Y in the grid
	 */
	protected int xIndex, yIndex;
	
	/**
	 * The actual coordinates of X, Y
	 */
	protected float xPos, yPos;
	
	/**
	 * The property/parcel list of this owner
	 */
	protected ArrayList<SimpleMarketLand> propertyList;
	
	/**
	 * this is a static reference layer used to convert real coordinates to/from grid coordinates
	 */
	public static RasterLayer<?> referenceLayer;
	
	/**
	 * In this version, a simple capital asset is used to decide if an agent affords a parcel/property.
	 */
	protected AssetResource capitalResource;
	
	/**
	 * The beta in C-D utility function
	 */
	protected double preferenceForProximity;
	
	/**
	 * The Land Management Strategies used by this SimpleAgent. <br>
	 * The first one must be irrigation right now.
	 */
	protected ArrayList<LandManagementStrategy> landManagementStrategies;
	
	/**
	 * Default constructor
	 */
	public SimpleAgent()
	{
		this.capitalResource = new AssetResource(AssetType.capital);
		this.capitalResource.setAssetQuantity(0);
		
		this.propertyList = new ArrayList<SimpleMarketLand>();
		this.landManagementStrategies = new ArrayList<LandManagementStrategy>();
	}
		
	/**
	 * @return the agentID
	 */
	public int getAgentID()
	{
		return agentID;
	}

	/**
	 * @return
	 */
	public int getX()
	{
		return this.xIndex;
	}
	
	/**
	 * @return
	 */
	public int getY()
	{
		return this.yIndex;
	}
	
	/**
	 * @param agentID the agentID to set
	 */
	public void setAgentID(int agentID)
	{
		this.agentID = agentID;
	}

	/**
	 * Set the position of this agent in the landscape/grid
	 * When possible, the row/column index will be updated accordingly.
	 * @param x
	 * @param y
	 */
	public void setPosition(float x, float y)
	{
		this.xPos = x;
		this.yPos = y;
		
		if(SimpleAgent.referenceLayer != null)
		{
			this.xIndex = referenceLayer.getCellCol(x);
			this.yIndex = referenceLayer.getCellRow(y);
		}
	}
	
	/**
	 * Set the index of row and column in the virtual grid.
	 * When possible, the real coordinates will be updated accordingly.
	 * @param xIndex
	 * @param yIndex
	 */
	public void setPosition(int xIndex, int yIndex)
	{
		this.xIndex = xIndex;
		this.yIndex = yIndex;
		
		if(SimpleAgent.referenceLayer != null)
		{
			this.xPos = (float)SimpleAgent.referenceLayer.getCellCenterX(xIndex);
			this.yPos = (float)SimpleAgent.referenceLayer.getCellCenterY(yIndex);
		}
	}
	
	/**
	 * @return the propertyList
	 */
	public ArrayList<SimpleMarketLand> getPropertyList()
	{
		return propertyList;
	}

	/**
	 * @return
	 */
	public int getPropertyNum()
	{
		if(propertyList == null) return 0;
		return propertyList.size();
	}
	
	/**
	 * @param propertyList the propertyList to set
	 */
	public void setPropertyList(ArrayList<SimpleMarketLand> propertyList)
	{
		this.propertyList = propertyList;
	}

	/**
	 Schedule the step method for agents. The method is scheduled starting at 
	 tick one with an interval of 1 tick. Specifically, the step starts at 1,
	 and recurs at 2,3,4,...etc
	 
	 This shuffle property of this method is turned on, so agents can have random orders to 
	 execute this method. Otherwise, some agents always take actions earlier than others.
	 */
	public void step()
	{
		// Override by subclasses
	}

	public int isAgent()
	{
		return 0;
	}

	public int isCell()
	{
		return 0;
	}
	
	/**
	 * @return the captialResource
	 */
	public AssetResource getCapitalResource()
	{
		return capitalResource;
	}
	
	/**
	 * @return
	 */
	public double getCapitalResourceQuantity()
	{
		return capitalResource.getAssetQuantity();
	}

	/**
	 * @param captialResource the captialResource to set
	 */
	public void setCapitalResource(AssetResource capitalResource)
	{
		this.capitalResource = capitalResource;
	}

	/**
	 * @return the preferenceForProximity
	 */
	public double getPreferenceForProximity()
	{
		return preferenceForProximity;
	}

	/**
	 * @param preferenceForProximity the preferenceForProximity to set
	 */
	public void setPreferenceForProximity(double preferenceForProximity)
	{
		this.preferenceForProximity = preferenceForProximity;
	}

	/**
	 * Update the agent's resources after buying or selling a parcel/house/property
	 */
	protected void updateResources()
	{
		
	}
	
	/**
	 * @param landProperty
	 * @return
	 */
	public synchronized boolean addProperty(SimpleMarketLand landProperty)
	{
		landOwnership.put(landProperty.getLandUnitID(), this);
		return propertyList.add((LandUnit)landProperty);
	}
	
	/**
	 * @param landProperty
	 * @return
	 */
	public synchronized boolean removeProperty(SimpleMarketLand landProperty)
	{
		// No need to remove, the put method of Hashtable will overwrite this record.
		//landOwnership.remove(landProperty.getLandUnitID());
		return this.propertyList.remove(landProperty);
	}

	public ArrayList<LandManagementStrategy> getLandManagementStrategies()
	{
		return landManagementStrategies;
	}

	/**
	 * @param landManagementStrategies the landManagementStrategies to set
	 */
	public void setLandManagementStrategies(
			ArrayList<LandManagementStrategy> landManagementStrategies) {
		this.landManagementStrategies = landManagementStrategies;
	}

	@Override
	public ArrayList<Number> getAgentAttributeValueList() 
	{
		ArrayList<Number> attriList = new ArrayList<Number>();
		attriList.add(this.getAgentID());
		attriList.add(this.getAgentTypeCode());
		attriList.add(this.getCapitalResourceQuantity());
		attriList.add(this.getPropertyNum());
		attriList.add(this.getPreferenceForProximity());
		attriList.add(this.getX());
		attriList.add(this.getY());
		return attriList;
	}

	public static ArrayList<String> getAgentAttributeNamesList() 
	{
		ArrayList<String> attriList = new ArrayList<String>();
		attriList.add("agentID");
		attriList.add("agentType");
		attriList.add("capitalResourceQuantity");
		attriList.add("propertyNum");
		attriList.add("preferenceForProximity");
		attriList.add("X");
		attriList.add("Y");
		return attriList;
	}

	@Override
	public int getAgentTypeCode() 
	{
		// Rural Land Owner 
		return AgentTypeCode.generic.ordinal();
	}
}