/**
 * SLUCE 2 Project
 * http://www.cscs.umich.edu/research/projects/sluce/
 */
package sluceii.land.market.exchange;

import java.util.ArrayList;

import sluceii.land.market.LandBidRecord;
import sluceii.land.market.LandListingRecord;
import sluceii.land.market.LandMarketListing;
import sluceii.land.market.LandBidRecord.BidRecordStatus;
import sluceii.land.market.SimpleMarketAgent;
import sluceii.land.market.SimpleMarketLand;

/**
 *
 *	This class implements some buying actions that are related to the land market.<br>
 *	Those related to the buyer agents are not handled here. Each buyer agent should take care
 * 	of her own resources and properties.
 * 
 * @author Shipeng Sun
 * @version 1.0
 */
public class LandBuyingAction<T extends SimpleMarketLand, S extends SimpleMarketAgent & LandBuyingProcess<T>>
{
	/**
	 * A reference to the land market listing
	 */
	protected LandMarketListing<T> landMarket;
	
	/**
	 * The buyer agent
	 */
	private S buyerAgent;

	/**
	 * Default constructor
	 */
	public LandBuyingAction()
	{
	}
	
	/**
	 * Constructor with buyer agent
	 */
	public LandBuyingAction(S agent)
	{
		buyerAgent = agent;
	}

	/**
	 * Constructor
	 * @param landMarket
	 * @param buyerAgent
	 */
	public LandBuyingAction(LandMarketListing<T> landMarket, S buyerAgent)
	{
		super();
		this.landMarket = landMarket;
		this.buyerAgent = buyerAgent;
	}

	/**
	 * @return the landMarket
	 */
	public LandMarketListing<T> getLandMarket()
	{
		return landMarket;
	}

	/**
	 * @param landMarket the landMarket to set
	 */
	public void setLandMarket(LandMarketListing<T> landMarket)
	{
		this.landMarket = landMarket;
	}

	/**
	 * @return the buyerAgent
	 */
	public S getBuyerAgent()
	{
		return buyerAgent;
	}

	/**
	 * @param buyerAgent the buyerAgent to set
	 */
	public void setBuyerAgent(S buyerAgent)
	{
		this.buyerAgent = buyerAgent;
	}
	
	/**
	 * Bid on the specified parcel object.
	 * Depending on the limit on the number of bids on a parcel, this may or may not succeed.
	 * @param listedLandObj The land listing record object
	 * @param utility buyer agent's utility for the land
	 * @param bidPrice bidding price
	 * @param wtp willingness to pay
	 * @return false if the seller does not accept bid anymore; otherwise, true.
	 */
	public boolean bidOnLand(LandListingRecord<T, ? extends SimpleMarketAgent> listedLandObj, double utility, double bidPrice, double wtp, double bidTime)
	{
		// The buyer cannot bid on the one that has already got maximum number of bids on it.
		if(landMarket.getOpenBidNumber(listedLandObj.getLandUnitID())>= listedLandObj.getMaxBidsAllowed())
		{
			return false;
		}
		
		// If the bidding price is lower than the WTA, it is rejected immediately.
		if(bidPrice < listedLandObj.getWillingToAccept()) return false;
		
		LandBidRecord<T,S> bidRecord = new LandBidRecord<T,S>();
		bidRecord.setListingRecObj(listedLandObj);
		
		bidRecord.setBuyerUtility(utility);
		bidRecord.setBidPrice(bidPrice);
		bidRecord.setWillingToPay(wtp);
		bidRecord.setRecordStatus(BidRecordStatus.open);
		bidRecord.setBuyer(this.buyerAgent);
		bidRecord.setBidTime(bidTime);
		
		landMarket.addBidRecord(bidRecord);
		
		return true;
	}
	
	/**
	 * Bid on the specified parcel object.
	 * Depending on the limit on the number of bids on a parcel, this may or may not succeed.
	 * @param listedLandObj The land listing record object
	 * @param utility buyer agent's utility for the land
	 * @param bidPrice bidding price
	 * @param wtp willingness to pay
	 * @return false if the seller does not accept bid anymore; otherwise, true.
	 */
	public boolean bidOnLand(LandListingRecord<T, ? extends SimpleMarketAgent> listedLandObj, double utility, double bidPrice, double wtp, double bidTime, ArrayList<Number> extraLandAttributes)
	{
		// The buyer cannot bid on the one that has already got maximum number of bids on it.
		if(landMarket.getOpenBidNumber(listedLandObj.getLandUnitID())>= listedLandObj.getMaxBidsAllowed())
		{
			return false;
		}
		
		// If the bidding price is lower than the WTA, it is rejected immediately.
		if(bidPrice < listedLandObj.getWillingToAccept()) return false;
		
		LandBidRecord<T,S> bidRecord = new LandBidRecord<T,S>();
		bidRecord.setListingRecObj(listedLandObj);
		
		bidRecord.setBuyerUtility(utility);
		bidRecord.setBidPrice(bidPrice);
		bidRecord.setWillingToPay(wtp);
		bidRecord.setRecordStatus(BidRecordStatus.open);
		bidRecord.setBuyer(this.buyerAgent);
		bidRecord.setBidTime(bidTime);
		
		landMarket.addBidRecord(bidRecord);
		
		if(extraLandAttributes != null)
		{
			bidRecord.addLandBidAttribute(extraLandAttributes);
		}
		
		return true;
	}
	
	/**
	 * The seller has accepted this buyer/household's bid
	 * @param landBidRecord
	 */
	public void bidAccepted(LandBidRecord<T,?> landBidRecord)
	{
		// Add this property/land parcel to this agent's property list
		buyerAgent.addProperty(landBidRecord.getLandUnitObj());
		landBidRecord.setRecordStatus(BidRecordStatus.accepted);
		
		// Notify seller to clean up.
		((LandSellingProcess<T>)landBidRecord.getListingRecord().getSeller()).clearTransaction(landBidRecord);
		
		// Cancel other bids, necessary?
		//bidRecordList.remove(landBidRecord);
		buyerAgent.getCapitalResource().subtractAsset(landBidRecord.getBidPrice());
	}
	
	/**
	 * The seller rejected the buyer's offer.
	 * @param landBidRecord
	 */
	public void bidRejected(LandBidRecord<T,?> landBidRecord)
	{
		landBidRecord.setRecordStatus(BidRecordStatus.rejected);
	}
}
