/**
 * SLUCE 2 Project
 * http://www.cscs.umich.edu/research/projects/sluce/
 */
package sluceii.land.market.exchange;

import java.util.ArrayList;

import sluceii.land.market.LandBidRecord;
import sluceii.land.market.LandListingRecord;
import sluceii.land.market.LandMarketListing;
import sluceii.land.market.LandBidRecord.BidRecordStatus;
import sluceii.land.market.LandListingRecord.ListingRecordStatus;
import sluceii.land.market.SimpleMarketAgent;
import sluceii.land.market.SimpleMarketLand;
import sluceii.landscape.LandUnit;
import sluceii.landscape.Parcel;

/**
 * @author Shipeng Sun
 * @version 1.0
 * 
 * This class implements the selling process that involves the land market.
 * Two types in the class definition are the type of land units and the seller type.
 * The seller must implement the Land Selling Process interface. 
 *
 */
public class LandSellingAction<T extends LandUnit, S extends SimpleMarketAgent & LandSellingProcess<T>>
{
	
	/**
	 * A reference to the land market listing service object 
	 */
	protected LandMarketListing<T> landMarket;
	
	
	/**
	 * The seller who is try to sell his/her properties
	 */
	protected S sellerAgent;
	
	/**
	 * Default constructor
	 */
	public LandSellingAction()
	{
	}

	/**
	 * Constructor
	 * @param landMarket
	 * @param sellerAgent
	 */
	public LandSellingAction(LandMarketListing<T> landMarket,S sellerAgent)
	{
		super();
		this.landMarket = landMarket;
		this.sellerAgent = sellerAgent;
	}

	/**
	 * @return the landMarket
	 */
	public LandMarketListing<T> getLandMarket()
	{
		return landMarket;
	}

	/**
	 * @param landMarket the landMarket to set
	 */
	public void setLandMarket(LandMarketListing<T> landMarket)
	{
		this.landMarket = landMarket;
	}

	/**
	 * @return the sellerAgent
	 */
	public SimpleMarketAgent getSellerAgent()
	{
		return sellerAgent;
	}

	/**
	 * @param sellerAgent the sellerAgent to set
	 */
	public void setSellerAgent(S sellerAgent)
	{
		this.sellerAgent = sellerAgent;
	}

	/**
	 * @param landCover
	 */
	public void putLandOnMarket(T landUnitObj, double askPrice, double wta, int maxAllowedBidNum, double listedTime)
	{
		landMarket.addParcelOnMarket(landUnitObj, sellerAgent, askPrice, wta, maxAllowedBidNum, listedTime, null);
	}
	
	/**
	 * @param landObj
	 * @param askPrice
	 * @param wta
	 * @param maxBidNums
	 * @param extraLandAttributes
	 */
	public void putLandOnMarket(T landObj, double askPrice, double wta, int maxAllowedBidNum, double listedTime, ArrayList<Number> extraLandAttributes) 
	{
		landMarket.addParcelOnMarket(landObj, sellerAgent, askPrice, wta, maxAllowedBidNum, listedTime, extraLandAttributes);
	}
	
	/**
	 * Check and possibly accept offer from the bidder
	 */
	public void checkBidOffers()
	{
		SimpleMarketLand curLandObj = null;
		LandBidRecord<T, ?> bidRecObj = null;
		
		double maxBidPrice = -10000.0f;
		int maxBidIndex = -1;
		double wta = 0;
		ArrayList<LandBidRecord<T, ?>> bidRecordList = null;
		
		for(int i = 0; i < sellerAgent.getPropertyList().size(); i++)
		{
			curLandObj = sellerAgent.getPropertyList().get(i);
			bidRecordList = landMarket.getBidRecordsByLandUnitID(curLandObj.getLandUnitID());
			if(bidRecordList == null) continue;
			maxBidIndex = -1;
			maxBidPrice = -10000.0f;
			
			for(int j = 0; j < bidRecordList.size(); j++)
			{
				bidRecObj = bidRecordList.get(j);
				wta = bidRecObj.getWillingToAccept();
				
				if(bidRecObj.getLandUnitID() == curLandObj.getLandUnitID() && bidRecObj.getListingRecord().getSeller() == sellerAgent)
				{
					// Check out the highest bid price
					if(bidRecObj.getRecordStatus() == BidRecordStatus.open && bidRecObj.getBidPrice() > maxBidPrice)
					{
						maxBidPrice = bidRecObj.getBidPrice();
						maxBidIndex = j;
					}
				}
			}
			
			// if the maximum bid price is not less than the willingness to accept price, deal
			if(maxBidPrice >= wta)
			{
				this.acceptBidOffer(bidRecordList.get(maxBidIndex));
				
				// Change the status of all other records
				for(int j = 0; j < bidRecordList.size(); j++)
				{
					bidRecObj = bidRecordList.get(j);
					if(bidRecObj.getListingRecord().getSeller() != this.sellerAgent) continue;
					
					if(j == maxBidIndex)
					{
						bidRecObj.setRecordStatus(BidRecordStatus.accepted);
					}
					else
					{
						bidRecObj.setRecordStatus(BidRecordStatus.rejected);
						rejectBidOffer(bidRecObj);
					}
				}

			}
		}		
	}
	
	/**
	 * Check and possibly accept offer from the bidder.
	 * This function also drops price after the wait period.
	 */
	public void checkBidOffers(int waitPeriod, boolean isRate, double askPriceDrop, double wtaDrop, double curTime)
	{
		SimpleMarketLand curLandObj = null;
		LandBidRecord<T, ?> bidRecObj = null;
		
		double maxBidPrice = -10000.0f;
		int maxBidIndex = -1;
		double wta = 0;
		ArrayList<LandBidRecord<T, ?>> bidRecordList = null;
		ArrayList<SimpleMarketLand> sellerPropertyList = sellerAgent.getPropertyList();
		boolean lowerWTA;
		
		synchronized(sellerPropertyList)
		{
			for(int i = 0; i < sellerPropertyList.size(); i++)
			{
				curLandObj = sellerPropertyList.get(i);
				bidRecordList = landMarket.getBidRecordsByLandUnitID(curLandObj.getLandUnitID());
				
				synchronized(bidRecordList)
				{
					for(int j = 0; j < bidRecordList.size(); j++)
					{
						bidRecObj = bidRecordList.get(j);
						if(bidRecObj.getListingRecord().getSeller() != this.sellerAgent) continue;
						
						wta = bidRecObj.getWillingToAccept();
						
						if(bidRecObj.getLandUnitID() == curLandObj.getLandUnitID())
						{
							// Check out the highest bid price
							if(bidRecObj.getRecordStatus() == BidRecordStatus.open && bidRecObj.getBidPrice() > maxBidPrice)
							{
								maxBidPrice = bidRecObj.getBidPrice();
								maxBidIndex = j;
							}
						}
					}

					lowerWTA = false;
				
					// if the maximum bid price is not less than the willingness to accept price, deal
					if(maxBidPrice == -10000.0f) // if maxBidPrice == -10000.0f, then no bidding at all.
					{
						lowerWTA = true;
					}
					else if(maxBidPrice >= wta) 
					{
						// Change the status of all other records
						for(int j = 0; j < bidRecordList.size(); j++)
						{
							bidRecObj = bidRecordList.get(j);
							if(bidRecObj.getListingRecord().getSeller() != this.sellerAgent) continue;
							
							if(j == maxBidIndex)
							{
								bidRecObj.setRecordStatus(BidRecordStatus.accepted);
								acceptBidOffer(bidRecordList.get(maxBidIndex));
							}
							else
							{
								bidRecObj.setRecordStatus(BidRecordStatus.rejected);
								rejectBidOffer(bidRecObj);
							}
						}
		
					}
					else
					{
						for(int j = 0; j < bidRecordList.size(); j++)
						{
							bidRecObj = bidRecordList.get(j);
							if(bidRecObj.getListingRecord().getSeller() != this.sellerAgent) continue;
							if(bidRecObj.getRecordStatus() == BidRecordStatus.open)
							{
								bidRecObj.setRecordStatus(BidRecordStatus.rejected);
								rejectBidOffer(bidRecObj);
							}
						}
						
						lowerWTA = true;
					}
					
					if(lowerWTA)
					{
						// See if need to lower the asking price and wta price
						LandListingRecord<T, ?> lRec = landMarket.getListingRecordByLandUnitID(curLandObj.getLandUnitID());
						
						if(lRec != null)
						{
							double listedTime = lRec.getListedTime();
							double updatedAskPrice, updatedWTA;
							
							// For every wait period, the seller will lower the price.
							if( (curTime - listedTime)% waitPeriod == 0)
							{
								updatedAskPrice = (isRate)?(lRec.getAskPrice()*(1-askPriceDrop)):(lRec.getAskPrice()-askPriceDrop);
								updatedWTA = (isRate)?(lRec.getWillingToAccept()*(1-wtaDrop)):(lRec.getWillingToAccept()-wtaDrop);
								lRec.setAskPrice(updatedAskPrice);
								lRec.setWillingToAccept(updatedWTA);
								lRec.setListedTime(curTime);
							}
						}
						else
						{
							int temp = 0;
							temp += 0;
						}
					}
				}
			}
		}
	}	
	
	/**
	 * Remove the parcel from the listing and this agent have to give up the ownership.
	 * The land owner might keep all the bidding records for future price formation.
	 * This owner needs to notify s/he has accepted the offer from one buyer.
	 * @param landBidRecord
	 */
	public void acceptBidOffer(LandBidRecord<T,?> landBidRecord)
	{
		LandBuyingProcess<T> buyerInterface = (LandBuyingProcess<T>)landBidRecord.getBuyer();
		if(buyerInterface != null)
			buyerInterface.bidAccepted(landBidRecord);
		else
			throw new NullPointerException("The seller cannot find the buyer object.");
		
		// This has to be done here because, after this, many things will change.
		if(landMarket.isToStoreTransactionRecords())
		{
			ArrayList<Number> transactionAttriList = new ArrayList<Number>();
			if(transactionAttriList != null)
			{
				transactionAttriList.addAll(sellerAgent.getAgentAttributeValueList());
				transactionAttriList.addAll(landBidRecord.getBuyer().getAgentAttributeValueList());
				transactionAttriList.addAll(landBidRecord.getAllLandAttributes());

				landMarket.addTransactionDataRecord(transactionAttriList);
			}
		}
	}

	/**
	 * Reject the bidder
	 * @param landBidRecord
	 */
	public void rejectBidOffer(LandBidRecord<T,?> landBidRecord)
	{
		LandBuyingProcess<T> buyerInterface = (LandBuyingProcess<T>)landBidRecord.getBuyer();
		if(buyerInterface != null)
			buyerInterface.bidRejected(landBidRecord);
		else
			throw new NullPointerException("The seller cannot find the buyer object.");
	}
	
	/**
	 * The buyer has accepted the transaction.
	 * update the market listing and property list 
	 */
	public void clearTransaction(LandBidRecord<T,?> landBidRecord)
	{
		// Remove this parcel from the market listing
		landMarket.removeListing(landBidRecord.getListingRecord());
		landBidRecord.getListingRecord().setRecordStaus(ListingRecordStatus.done);
		landBidRecord.setRecordStatus(BidRecordStatus.completed);
		
		// Remove this parcel from the owner's property list
		sellerAgent.removeProperty(landBidRecord.getLandUnitObj());
		
		// Gain some cash
		sellerAgent.getCapitalResource().addAsset(landBidRecord.getBidPrice());
	}

	/**
	 * @param landObj
	 * @param askPrice
	 * @param wta
	 * @param maxBidNums
	 * @param extraLandAttributes
	 */
	public void putLandOnMarket(Parcel landObj, double askPrice, double wta, int maxBidNums, ArrayList<Number> extraLandAttributes) 
	{
	}
	
}
