/**
 * SLUCE 2 Project http://www.cscs.umich.edu/research/projects/sluce/
 */
package sluceii.land.market;

import java.util.ArrayList;
import java.util.Collection;

import sluceii.land.market.exchange.LandBuyingProcess;
import sluceii.land.market.SimpleMarketLand;

/**
 * This is the land bid record. It has the parcel/land to be bid, WTP, WTA,
 * buyer, and seller. It may or may not become a transaction. All bid records
 * are kept through the program. They are also indexed by parcel ID.
 * 
 * @see LandMarketListing
 * @see LandListingRecord
 * 
 * @author Shipeng Sun
 */
public class LandBidRecord<T extends SimpleMarketLand, S extends SimpleMarketAgent & LandBuyingProcess<T>>
{
	public enum BidRecordStatus {open, accepted, rejected, pending, completed}

	/**
	 * A unique identifier for the bid record.
	 */
	protected int recordID;

	/**
	 * Maximum record ID used to generate a new record ID
	 */
	private static int maxRecordID = 0;

	/**
	 * buyer is a simple agent, but it must implement LandBuyingProcess Interface
	 */
	private S buyer;
	
	/**
	 * The listing record. It contains the information about seller and land.
	 */
	protected LandListingRecord<T,? extends SimpleMarketAgent> listingRecObj;

	/**
	 * The bid price offered by buyer 
	 */
	protected double bidPrice;
	
	/**
	 * The Repast tick count at the bidding moment.
	 */
	protected double bidTime;
	
	/**
	 * The WTP from buyer
	 */
	protected double willingToPay;
	
	/**
	 * The utility for the buyer at the bid moment
	 */
	protected double buyerUtility;
	
	/**
	 * The current status of this bid
	 */
	protected BidRecordStatus recordStatus;
	
	/**
	 * A generic object used to save some unstructured attributes associated with the land.
	 * The use and interpretation of this object is subject to users. This class will do nothing about it.
	 */
	protected Object genericAttributeContainer;
	
	protected ArrayList<Number> landBidAttributes;
	
	// TODO: add a time/status history for the record.
	
	/**
	 * 
	 */
	public LandBidRecord()
	{
		setRecordID(++LandBidRecord.maxRecordID);
		
		// Default status is open.
		this.recordStatus = BidRecordStatus.open;
	}

	/**
	 * @return the recordID
	 */
	public int getRecordID()
	{
		return recordID;
	}

	/**
	 * @param recordID
	 *            the recordID to set
	 */
	public void setRecordID(int recordID)
	{
		this.recordID = recordID;
	}

	/**
	 * @return the bidPrice
	 */
	public double getBidPrice()
	{
		return bidPrice;
	}

	/**
	 * @param bidPrice
	 *            the bidPrice to set
	 */
	public void setBidPrice(double bidPrice)
	{
		this.bidPrice = bidPrice;
	}

	/**
	 * @return the bidTime
	 */
	public double getBidTime()
	{
		return bidTime;
	}

	/**
	 * @param bidTime the bidTime to set
	 */
	public void setBidTime(double bidTime)
	{
		this.bidTime = bidTime;
	}

	/**
	 * @return the buyerUtility
	 */
	public double getBuyerUtility()
	{
		return buyerUtility;
	}

	/**
	 * @param buyerUtility the buyerUtility to set
	 */
	public void setBuyerUtility(double buyerUtility)
	{
		this.buyerUtility = buyerUtility;
	}

	/**
	 * @return the willingToPay
	 */
	public double getWillingToPay()
	{
		return willingToPay;
	}

	/**
	 * @param willingToPay
	 *            the willingToPay to set
	 */
	public void setWillingToPay(double willingToPay)
	{
		this.willingToPay = willingToPay;
	}

	/**
	 * @return the willingToAccept
	 */
	public double getWillingToAccept()
	{
		return listingRecObj.getWillingToAccept();
	}

	/**
	 * @return the listingRecObj
	 */
	public LandListingRecord<T,?> getListingRecord()
	{
		return listingRecObj;
	}

	/**
	 * @param listingRecObj
	 *            the listingRecObj to set
	 */
	public void setListingRecObj(LandListingRecord<T,?> listingRecObj)
	{
		this.listingRecObj = listingRecObj;
	}

	/**
	 * @return the buyer
	 */
	public S getBuyer()
	{
		return buyer;
	}
	
	/**
	 * @param <T>
	 * @param s
	 */
	public void setBuyer(S buyer)
	{
		this.buyer = buyer;
	}

	/**
	 * @return
	 */
	public int getLandUnitID()
	{
		return listingRecObj.getLandUnitID();
	}

	/**
	 * @return
	 */
	public T getLandUnitObj()
	{
		return this.listingRecObj.getLandUnitObj();
	}

	/**
	 * @return the recordStatus
	 */
	public BidRecordStatus getRecordStatus()
	{
		return recordStatus;
	}

	/**
	 * @param recordStatus the recordStatus to set
	 */
	public void setRecordStatus(BidRecordStatus recordStatus)
	{
		this.recordStatus = recordStatus;
	}
	
	/**
	 * @param val
	 */
	public void addLandBidAttribute(Number val)
	{
		if(landBidAttributes != null)
		{
			landBidAttributes.add(val);
		}
		else
		{
			landBidAttributes = new ArrayList<Number>();
			landBidAttributes.add(val);
		}
	}
	
	public Number getLandBidAttribute(int index)
	{
		if(this.landBidAttributes != null)
		{
			return landBidAttributes.get(index);
		}
		else
		{
			return null;
		}
	}
	
	/**
	 * @return the landProperties
	 */
	public ArrayList<Number> getLandBidAttributes() 
	{
		return landBidAttributes;
	}

	/**
	 * @return the genericAttributeContainer
	 */
	public Object getGenericAttributeContainer() 
	{
		return genericAttributeContainer;
	}

	/**
	 * @param genericAttributeContainer the genericAttributeContainer to set
	 */
	public void setGenericAttributeContainer(Object genericAttributeContainer) 
	{
		this.genericAttributeContainer = genericAttributeContainer;
	}

	/**
	 * @param landAttributes
	 */
	public void addLandBidAttribute(Collection<Number> landAttributes) 
	{
		if(landBidAttributes == null)
		{
			landBidAttributes = new ArrayList<Number>();
		}

		landBidAttributes.addAll(landAttributes);
	}

	/**
	 * All attributes of the bid record, with three parts.
	 * First, all attributes from the listed record, @see LandListingRecord.getAllLandAttributes()
	 * Second, internal attributes, bidTime, bidPrice, WTP, and buyerUtility
	 * Third, external attributes added through AddLandBidAttribute methods
	 * 
	 * @return A listed of attributes related to the land and the listing and bidding
	 */
	public ArrayList<Number> getAllLandAttributes() 
	{
		ArrayList<Number> rslt = new ArrayList<Number>();
		rslt.addAll(this.getListingRecord().getAllLandAttributes());
		
		rslt.add(this.bidTime);
		rslt.add(this.bidPrice);
		rslt.add(this.willingToPay);
		rslt.add(this.buyerUtility);
		
		if(landBidAttributes == null || landBidAttributes.size() == 0)
		{
			return rslt;
		}
		
		rslt.addAll(landBidAttributes);
		return rslt;
	}
}
