/**
 * SLUCE 2 Project http://www.cscs.umich.edu/research/projects/sluce/
 */
package sluceii.land.market;

import java.util.ArrayList;
import java.util.Collection;

import sluceii.land.market.exchange.LandSellingProcess;

/**
 * This is the record listed on the market. It contains the parcel, seller, ask
 * price, and willingness to accept information. Buyers search a list of these
 * records and evaluate some of them to decide which parcels to bid on.
 * 
 * @see LandMarketListing
 * @author Shipeng Sun
 */
public class LandListingRecord<T extends SimpleMarketLand, S extends SimpleMarketAgent & LandSellingProcess<T>>
{
	/**
	 * @author Shipeng Sun
	 * The status of the listed parcel/property/house
	 */
	public enum ListingRecordStatus {listed, bidded, done};
	
	/**
	 * The maximum ID of these records, used to generate the next ID.
	 */
	public static int maxListingRecordID;
	
	/**
	 * ID
	 */
	protected int listingRecordID;
	
	/**
	 * the status of the listing record
	 */
	protected ListingRecordStatus recordStaus;

	/**
	 * The land unit listed in this record
	 */
	protected T landUnitObj;

	/**
	 * The seller who is selling the land
	 */
	private S seller;

	/**
	 * The asking price from the seller. It may equal to or higher than WTA.
	 */
	protected double askPrice;

	/**
	 * WTA for the seller
	 */
	protected double willingToAccept;
	
	/**
	 * The maximum bids allowed on this parcel. The seller will not accept bids beyond this number.
	 */
	protected int maxBidsAllowed;
	
	/**
	 * The date/time when the property is listed.
	 */
	protected double listedTime;
	
	/**
	 * A generic object used to save some unstructured attributes associated with the land.
	 * The use and interpretation of this object is subject to users. This class will do nothing about it.
	 */
	protected Object genericAttributeContainer;

	/**
	 * Specific land attributes that are not market related, such as distance to downtown, open space amenity at the bidding time,
	 * and others as necessary. Data are added when seller put the land on the market.
	 * 
	 * The nature of these attributes is subject to agents' interpretation.
	 */
	protected ArrayList<Number> landAttributes;
	
	/**
	 * Default constructor
	 */
	public LandListingRecord()
	{
		this.listingRecordID = ++ maxListingRecordID;
		
		// The default status when the listing record is created is listed.
		this.recordStaus = ListingRecordStatus.listed;
	}

	/**
	 * Get the ID of the land for sale.
	 * @return ID of the land unit object
	 */
	public int getLandUnitID()
	{
		return landUnitObj.getLandUnitID();
	}

	/**
	 * Get the land object for sale.
	 * @return the parcelObj
	 */
	public T getLandUnitObj()
	{
		return this.landUnitObj;
	}

	/**
	 * Set the land to sell
	 * @param parcelObj
	 *            the parcelObj to set
	 */
	public void setLandUnitObj(T landUnitObj)
	{
		this.landUnitObj = landUnitObj;
	}

	/**
	 * @return the askPrice
	 */
	public double getAskPrice()
	{
		return askPrice;
	}

	/**
	 * @param askPrice
	 *            the askPrice to set
	 */
	public void setAskPrice(double askPrice)
	{
		this.askPrice = askPrice;
	}

	/**
	 * @return the willingToAccept
	 */
	public double getWillingToAccept()
	{
		return willingToAccept;
	}

	/**
	 * @param willingToAccept
	 *            the willingToAccept to set
	 */
	public void setWillingToAccept(double willingToAccept)
	{
		this.willingToAccept = willingToAccept;
	}

	/**
	 * @return the seller
	 */
	public S getSeller()
	{
		return seller;
	}

	/**
	 * @param <S>
	 * @param <S>
	 * @param seller the seller to set
	 */
	public void setSeller(S seller)
	{
		this.seller = seller;
	}

	/**
	 * @return the maxBidsAllowed
	 */
	public int getMaxBidsAllowed()
	{
		return maxBidsAllowed;
	}

	/**
	 * @param maxBidsAllowed the maxBidsAllowed to set
	 */
	public void setMaxBidsAllowed(int maxBidsAllowed)
	{
		this.maxBidsAllowed = maxBidsAllowed;
	}

	/**
	 * @return the recordStaus
	 */
	public ListingRecordStatus getRecordStaus()
	{
		return recordStaus;
	}

	/**
	 * @param recordStaus the recordStaus to set
	 */
	public void setRecordStaus(ListingRecordStatus recordStaus)
	{
		this.recordStaus = recordStaus;
	}

	/**
	 * @return the listedTime
	 */
	public double getListedTime()
	{
		return listedTime;
	}

	/**
	 * @param listedTime the listedTime to set
	 */
	public void setListedTime(double listedTime)
	{
		this.listedTime = listedTime;
	}

	/**
	 * @return the genericAttributeContainer
	 */
	public Object getGenericAttributeContainer() {
		return genericAttributeContainer;
	}

	/**
	 * @param genericAttributeContainer the genericAttributeContainer to set
	 */
	public void setGenericAttributeContainer(Object genericAttributeContainer) {
		this.genericAttributeContainer = genericAttributeContainer;
	}

	/**
	 * @return the landProperties
	 */
	public ArrayList<Number> getLandAttributes() 
	{
		return landAttributes;
	}

	/**
	 * Add an attribute value 
	 * @param val
	 */
	public boolean addLandAttribute(Number val)
	{
		if(landAttributes != null)
		{
			return landAttributes.add(val);
		}
		else
		{
			landAttributes = new ArrayList<Number>();
			return landAttributes.add(val);
		}
	}

	/**
	 * Add a collection of attribute values
	 * @param vals
	 */
	public boolean addLandAttribute(Collection<Number> vals)
	{
		if(landAttributes != null)
		{
			return landAttributes.addAll(vals);
		}
		else
		{
			landAttributes = new ArrayList<Number>();
			return landAttributes.addAll(vals);
		}
	}
	
	/**
	 * @param index
	 * @return
	 */
	public Number getLandAttribute(int index)
	{
		if(this.landAttributes != null)
		{
			return landAttributes.get(index);
		}
		else
		{
			return null;
		}
	}
	
	/**
	 * This method should be used to retrieve all attributes, including two parts.
	 * First are those internally added by the market institution, including
	 * recordID
	 * land type (farm/subdivision, parcel)
	 * maxBidsAllowed
	 * askPrice
	 * WTA
	 * listedTime
	 * 
	 * Second are those explicitly added through addLandAttribute method
	 * 
	 * @return A list of land attributes
	 */
	public ArrayList<Number> getAllLandAttributes() 
	{
		ArrayList<Number> rslt = new ArrayList<Number>();
		rslt.add(this.listingRecordID);
		rslt.add(landUnitObj.getLandUnitTypeCode());
		rslt.add(this.maxBidsAllowed);
		rslt.add(this.askPrice);
		rslt.add(this.willingToAccept);
		rslt.add(this.listedTime);
		
		if(landAttributes != null && landAttributes.size() > 0) rslt.addAll(landAttributes);
		
		return rslt;
	}
}
