/**
 * SLUCE 2 Project
 * http://www.cscs.umich.edu/research/projects/sluce/
 * @version 1.0
 */
package sluceii.land.market;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;

import repast.simphony.random.RandomHelper;
import sluceii.land.market.LandBidRecord.BidRecordStatus;
import sluceii.land.market.LandListingRecord.ListingRecordStatus;
import sluceii.land.market.exchange.LandSellingProcess;

/**
 * This class represents the land market listing. All sellers put their land-to-sell on this listing.
 * Potential buyers bid on this list.
 * 
 * @author Shipeng
 *
 */
public class LandMarketListing<T extends SimpleMarketLand>
{
	/**
	 * This ID is used to serialize the class to external hard drive.
	 */
	 private static final long serialVersionUID = 8228686603144586836L;
	
	/**
	 * Vacant (available, on-the-market) parcel/house list
	 */
	protected Hashtable<Integer, LandListingRecord<T,? extends SimpleMarketAgent>> toSellLandUnitList;

	/**
	 * The market listing records organized by seller ID (agent ID)
	 * No record will be deleted from this hash table.
	 * NOTES: the toSellParcelList will remove records whose land transaction has been completed.
	 */
	protected Hashtable<Integer, ArrayList<LandListingRecord<T, ? extends SimpleMarketAgent>>> sellerIndexedListingRecords;
	
	/**
	 * This hash table organizes bid records using parcel ID.
	 * Key: parcel ID
	 * Value: all bid records on this parcel 
	 */
	protected Hashtable<Integer, ArrayList<LandBidRecord<T, ? extends SimpleMarketAgent>>> landUnitIndexedBidRecords;
	
	/**
	 * All parcels indexed by their ID.
	 */
	protected Hashtable<Integer, T> landUnitList;
	
	/**
	 * Land broker is an integral part of land markets.
	 */
	protected Hashtable<Integer,SimpleMarketAgent> landBrokerList;
	
	protected LandMarketTransactionDataDepot tranactionDataObj;
	
	/**
	 * Default constructor
	 */
	public LandMarketListing()
	{
		this.toSellLandUnitList = new Hashtable<Integer,LandListingRecord<T,? extends SimpleMarketAgent>>();
		this.landUnitIndexedBidRecords = new Hashtable<Integer, ArrayList<LandBidRecord<T, ? extends SimpleMarketAgent>>>();
		this.sellerIndexedListingRecords = new Hashtable<Integer, ArrayList<LandListingRecord<T, ? extends SimpleMarketAgent>>>();
		this.landUnitList = null;

		landBrokerList = new Hashtable<Integer,SimpleMarketAgent>();
		tranactionDataObj = new LandMarketTransactionDataDepot();
	}

	/**
	 * Remove all elements in the market
	 */
	public void clear()
	{
		this.landUnitIndexedBidRecords.clear();
		this.sellerIndexedListingRecords.clear();
		this.toSellLandUnitList.clear();
		if(landBrokerList != null) landBrokerList.clear();
		tranactionDataObj.clear();
	}	
	
	/**
	 * @param parcelObj
	 */
	public void addLandUnitOnMarket(LandListingRecord<T,? extends SimpleMarketAgent> listObj)
	{
		if(listObj == null) return;
		
		this.toSellLandUnitList.put(listObj.getLandUnitID(),listObj);
		
		if(this.sellerIndexedListingRecords.containsKey(listObj.getSeller().getAgentID()))
		{
			ArrayList<LandListingRecord<T,? extends SimpleMarketAgent>> recordList = this.sellerIndexedListingRecords.get(listObj.getSeller().getAgentID());
			recordList.add(listObj);
		}
		else
		{
			ArrayList<LandListingRecord<T,? extends SimpleMarketAgent>> recordList = new ArrayList<LandListingRecord<T,? extends SimpleMarketAgent>>();
			recordList.add(listObj);
			this.sellerIndexedListingRecords.put(listObj.getSeller().getAgentID(), recordList);
		}
	}

	/**
	 * @param parcelObj
	 */
	public <S extends SimpleMarketAgent & LandSellingProcess<T>> void addParcelOnMarket(T landUnitObj, S seller, double askPrice, double wta, int maxBidsAllowed, double listedTime, ArrayList<Number> extraAttributes)
	{
		if(landUnitObj == null) return;
		
		//this.toSellParcelList.add(parcelObj);
		// TODO: Remove the old one. The seller might update their ask price and WTA.
		//removeListing(parcelObj.getParcelID());
		
		LandListingRecord<T, S> listObj = new LandListingRecord<T, S>();
		listObj.landUnitObj = landUnitObj;
		listObj.setSeller(seller);
		listObj.askPrice = askPrice;
		listObj.willingToAccept = wta;
		listObj.maxBidsAllowed = maxBidsAllowed;
		
		listObj.recordStaus = ListingRecordStatus.listed;
		listObj.setListedTime(listedTime);
		
		if(extraAttributes != null) listObj.addLandAttribute(extraAttributes);
		
		this.addLandUnitOnMarket(listObj);
	}

	/**
	 * Remove this parcel from the listing
	 * 
	 * @param parcelID
	 */
	public boolean removeListing(int landUnitID)
	{
		if(toSellLandUnitList.remove(landUnitID) == null) return false;
		return true;
	}
	
	/**
	 * Remove this parcel from the listing
	 * 
	 * @param parcelID
	 */
	public boolean removeListing(LandListingRecord<T, ? extends SimpleMarketAgent> listingRecObj)
	{
		if(this.toSellLandUnitList.contains(listingRecObj))
		{
			if(toSellLandUnitList.remove(listingRecObj.landUnitObj.getLandUnitID()) == listingRecObj)
			{
				return true;
			}
			return false;
		}
		else
		{
			return false;
		}
	}

	/**
	 * 
	 * @param bidRecord
	 */
	public void addBidRecord(LandBidRecord<T, ? extends SimpleMarketAgent> bidRecord)
	{
		try
		{
			if(landUnitIndexedBidRecords.containsKey(bidRecord.getLandUnitID()))
			{
				ArrayList<LandBidRecord<T, ? extends SimpleMarketAgent>> list = landUnitIndexedBidRecords.get(bidRecord.getLandUnitID());
				list.add(bidRecord);
			}
			else
			{
				ArrayList<LandBidRecord<T, ? extends SimpleMarketAgent>> list = new ArrayList<LandBidRecord<T, ? extends SimpleMarketAgent>>();
				list.add(bidRecord);
				landUnitIndexedBidRecords.put(bidRecord.getLandUnitID(),list);
			}
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
	}
	
	/**
	 * Get the 
	 * @param parcelID
	 * @return
	 */
	public int getCurrentBidsNumber(int landUnitID)
	{
		if(landUnitIndexedBidRecords.containsKey(landUnitID))
		{
			ArrayList<LandBidRecord<T, ? extends SimpleMarketAgent>> parcelBids = landUnitIndexedBidRecords.get(landUnitID);
			return parcelBids.size();
		}
		else
		{
			return 0;
		}
	}
	
	/**
	 * The number of open bid on this record.
	 * @param parcelID
	 * @return
	 */
	public int getOpenBidNumber(int landUnitID)
	{
		if(landUnitIndexedBidRecords.containsKey(landUnitID))
		{
			int rslt = 0;
			ArrayList<LandBidRecord<T,? extends SimpleMarketAgent>> parcelBids = landUnitIndexedBidRecords.get(landUnitID);
			
			for(LandBidRecord<T,? extends SimpleMarketAgent> lbRec : parcelBids)
			{
				if(lbRec.getRecordStatus() == BidRecordStatus.open)
				{
					rslt ++;
				}
			}
			
			return rslt;
		}
		else
		{
			return 0;
		}
		
	}

	/**
	 * @return the toSellParcelList
	 */
	public Hashtable<Integer,LandListingRecord<T,? extends SimpleMarketAgent>> getToSellLandUnitList()
	{
		return toSellLandUnitList;
	}
	
	/**
	 * @param luID
	 * @return
	 */
	public LandListingRecord<T,? extends SimpleMarketAgent> getListingRecordByLandUnitID(int luID)
	{
		return this.toSellLandUnitList.get(luID);
	}

	/**
	 * Return all bid records on the parcel with this parcelID.
	 * @param parcelID
	 * @return
	 */
	public ArrayList<LandBidRecord<T,? extends SimpleMarketAgent>> getBidRecordsByLandUnitID(int landUnitID)
	{
		if(landUnitID < 0)
		{ 
			return null;
		}
		
		return landUnitIndexedBidRecords.get(landUnitID);
	}

	/**
	 * @return the parcelList
	 */
	public Hashtable<Integer, T> getLandUnitList()
	{
		return this.landUnitList;
	}

	/**
	 * @param parcelList the parcelList to set
	 */
	public void setLandUnitList(Hashtable<Integer, T> landUnitList)
	{
		this.landUnitList = landUnitList;
	}

	/**
	 * A tabular data object.
	 * Each line: seller info, buyer info, land transaction info.
	 * @return the transactionRecordValues
	 */
	public ArrayList<ArrayList<Number>> getTransactionRecordValues() 
	{
		return tranactionDataObj.transactionRecordValues;
	}

	/**
	 * NOTE: This does NOT return field names related to agents.
	 * It only has fields related to the land listing/bidding record.
	 * However, the getTransactionRecordValues method does return values related to agents, including buyers and sellers.
	 * Those values are specified in classes that implement SimpleMarketAgent::getAgentAttributeValueList method.
	 * 
	 * @return names of the fields that are related to land listed and bid.
	 */
	public ArrayList<String> getTransactionRecordLandFields() 
	{
		return tranactionDataObj.getAllLandAttributeNames();
	}

	/**
	 * @param landBrokerObj
	 */
	public void registerLandBorker(SimpleMarketAgent landBrokerObj)
	{
		this.landBrokerList.put(landBrokerObj.getAgentID(), landBrokerObj);
	}
	
	/**
	 * @param landBrokerObj
	 * @param agentID
	 * @return
	 */
	public SimpleMarketAgent unregisterLandBroker(SimpleMarketAgent landBrokerObj, Integer agentID)
	{
		return landBrokerList.remove(landBrokerObj.getAgentID());
	}

	/**
	 * @param agentID
	 * @return
	 */
	public SimpleMarketAgent unregisterLandBroker(Integer agentID)
	{
		return landBrokerList.remove(agentID);
	}

	/**
	 * Randomly choose a land broker from the list.
	 * @return
	 */
	public SimpleMarketAgent getRandomBroker()
	{
		Object[] brokers = this.landBrokerList.values().toArray();
		int rndIndex = (int)(RandomHelper.nextDouble()*(brokers.length -1));
		
		return (rndIndex >= 0 && rndIndex < brokers.length)?((SimpleMarketAgent)brokers[rndIndex]):null;
	}
	
	/**
	 * @param agentID
	 * @return
	 */
	public SimpleMarketAgent getBrokerByID(Integer agentID)
	{
		return landBrokerList.get(agentID);
	}

	/**
	 * @return the isToStoreTransactionRecords
	 */
	public boolean isToStoreTransactionRecords() 
	{
		return tranactionDataObj.isToStoreTransactionRecords;
	}

	/**
	 * @param isToStoreTransactionRecords the isToStoreTransactionRecords to set
	 */
	public void setToStoreTransactionRecords(boolean isToStoreTransactionRecords) 
	{
		tranactionDataObj.isToStoreTransactionRecords = isToStoreTransactionRecords;
	}

	/**
	 * @param transactionAttriList
	 */
	public void addTransactionDataRecord(ArrayList<Number> transactionAttriList) 
	{
		tranactionDataObj.transactionRecordValues.add(transactionAttriList);
	}
	
	/**
	 * @param pName
	 * @return
	 */
	public boolean addLandListingAttributeName(String pName)
	{
		return tranactionDataObj.addLandListingAttributeName(pName);
	}
	
	/**
	 * @param pNames
	 * @return
	 */
	public boolean addLandListingAttributeName(Collection<String> pNames)
	{
		return tranactionDataObj.addLandListingAttributeName(pNames);
	}

	/**
	 * @param pName
	 * @return
	 */
	public boolean addBidAttributeName(String pName)
	{
		return tranactionDataObj.addBidAttributeName(pName);
	}
	
	/**
	 * @param pNames
	 * @return
	 */
	public boolean addBidAttributeName(Collection<String> pNames)
	{
		return tranactionDataObj.addBidAttributeName(pNames);	
	}
	
	/**
	 * @param pw
	 * @param sellerInfoFields
	 * @param buyerInfoFields
	 */
	public void exportTransactionData(PrintWriter pw, ArrayList<String> sellerInfoFields, ArrayList<String> buyerInfoFields)
	{
		tranactionDataObj.exportTransactionFieldNames(pw, sellerInfoFields, buyerInfoFields);
		tranactionDataObj.exportTransactionRecords(pw);
	}
	

}
