/**
 * SLUCE 2 Project
 * http://www.cscs.umich.edu/research/projects/sluce/
 */
package sluceii.landscape;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;

import repast.simphony.engine.environment.RunEnvironment;
import sluceii.agent.SimpleAgent;
import sluceii.land.market.utility.ConsumptionBundle;
import sluceii.landscape.LandUnit.LandAggregateMethod;
import sluceii.params.AppParameters;
import sluceii.params.ModelParameters;
import umn.geog.helia.gis.GISConstants;
import umn.geog.helia.gis.HeliaDataLoader;
import umn.geog.helia.gis.HeliaGIS;
import umn.geog.helia.gis.sa.BasicOperator;
import umn.geog.helia.gis.sa.SimpleSAOperator;
import umn.geog.helia.landscape.LandscapeLayerGroup;
import umn.geog.helia.landscape.LandscapeLoader;
import umn.geog.helia.landscape.RasterLayer;

/**
 * This class provides land cover data services to all agents.
 * It has a root landscape group object, which store all raster data layers.
 * 
 * This class mainly utilizes HELIA GIS java components.
 * @see hegis.umn.edu
 * 
 * @author Shipeng Sun
 * @version 1.0
 *
 */
public class LandCoverDataHelper
{
	/**
	 * The root landscape layer group object. Shared among all agents.
	 */
	public static LandscapeLayerGroup rootLayerGroup = new LandscapeLayerGroup();
	
	/**
	 * A GIS class used to load data 
	 */
	public static HeliaDataLoader heliaObj = new HeliaDataLoader();
	
	/**
	 * HELIA GIS object with spatial analysis funcationalities
	 */
	private static HeliaGIS hegisObj = new HeliaGIS();
	
	/**
	 * A simple spatial analysis operator
	 */
	private static SimpleSAOperator saObj = new SimpleSAOperator();
	
	/**
	 * A hash table/dictionary of all parcel objects
	 */
	public static Hashtable<Integer, Parcel> parcelList = new Hashtable<Integer, Parcel>();
	
	/**
	 * A hash table/dictionary of all subdivision objects
	 */
	public static Hashtable<Integer, Subdivision> subdivisionList = new Hashtable<Integer, Subdivision>();

	/**
	 * 
	 */
	public LandCoverDataHelper()
	{
		// TODO Auto-generated constructor stub
	}
	
	public static void clear()
	{
		rootLayerGroup.deepClear();
		parcelList.clear();
		subdivisionList.clear();
	}
	
	/**
	 * Load the root landscape layer group information to the object.
	 * This would not load the actual raster data, it just load layer meta data.
	 * 
	 * @return success or failure
	 */
	public static boolean loadRootLandscapeLayerGroupInfo()
	{
		String rootXML = AppParameters.dataFolderName + AppParameters.landscapeGroupXMLFile;
		LandscapeLoader llObj = new LandscapeLoader(rootXML,rootLayerGroup);
		return llObj.loadLandscapeLayers();
	}
	
	/**
	 * @param rootXMLFile
	 * @return
	 */
	public static boolean loadRootLandscapeLayerGroupInfo(String rootXMLFile)
	{
		LandscapeLoader llObj = new LandscapeLoader(rootXMLFile,rootLayerGroup);
		return llObj.loadLandscapeLayers();
	}
	
	/**
	 * @param rootPath
	 * @param rootFile
	 * @return
	 */
	public static boolean loadRootLandscapeLayerGroupInfo(String rootPath, String rootFile)
	{
		LandscapeLoader llObj = new LandscapeLoader(rootPath+rootFile,rootLayerGroup);
		return llObj.loadLandscapeLayers();
	}
	
	/**
	 * Alias of loadRootLandscapeLayerGroupInfo
	 * @return success or failure
	 */
	public static boolean loadRootLandscapeLayerGroupMetadata()
	{
		return LandCoverDataHelper.loadRootLandscapeLayerGroupInfo();
	}
	
	/**
	 * @param rootXMLFile
	 * @return
	 */
	public static boolean loadRootLandscapeLayerGroupMetadata(String rootXMLFile)
	{
		return LandCoverDataHelper.loadRootLandscapeLayerGroupInfo(rootXMLFile);
	}
	
	/**
	 * @param rootPath
	 * @param rootFile
	 * @return
	 */
	public static boolean loadRootLandscapeLayerGroupMetadata(String rootPath, String rootFile)
	{
		return LandCoverDataHelper.loadRootLandscapeLayerGroupInfo(rootPath, rootFile);
	}
	
	/**
	 * Load the root landscape layer group data to the object.
	 * This would load the actual raster data. Make sure load the meta-data before calling this.
	 * 
	 * @return success or failure
	 */
	public static boolean loadRootLandscapeLayerGroupData()
	{
		return heliaObj.loadLandscapeGroupData(rootLayerGroup);
	}

	/**
	 * Load the root landscape layer group information and data to the object.
	 * This would load the actual raster data.
	 * 
	 * @return success or failure
	 */
	public static boolean loadRootLandscapeLayerGroup()
	{
		if(LandCoverDataHelper.loadRootLandscapeLayerGroupInfo())
		{
			return LandCoverDataHelper.loadRootLandscapeLayerGroupData();
		}
		else
		{
			return false;
		}
	}
	
	/**
	 * @param rootXMLFileName
	 * @return
	 */
	public static boolean loadRootLandscapeLayerGroup(String rootXMLFileName)
	{
		if(LandCoverDataHelper.loadRootLandscapeLayerGroupInfo(rootXMLFileName))
		{
			return LandCoverDataHelper.loadRootLandscapeLayerGroupData();
		}
		else
		{
			return false;
		}
	}
	
	/**
	 * @param rootPath
	 * @param xmlFileName
	 * @return
	 */
	public static boolean loadRootLandscapeLayerGroup(String rootPath, String xmlFileName)
	{
		if(LandCoverDataHelper.loadRootLandscapeLayerGroupInfo(rootPath, xmlFileName))
		{
			return LandCoverDataHelper.loadRootLandscapeLayerGroupData();
		}
		else
		{
			return false;
		}
	}	
	
	/**
	 * @return
	 */
	public static boolean saveAllLandscapeLayers()
	{
		return heliaObj.saveLandscapeGroupData(rootLayerGroup);
	}
	
	/**
	 * Save a raster layer. It can add a time stamp and/or tick count to the file name.
	 * @param dataLayer The raster layer object
	 * @param addTimeStamp Whether to add a time stamp to the file name
	 * @param addTickCount Whether to add a tick count to the file name
	 * @return success or failure. If the layer has not been changed, it will not be saved. 
	 */
	public static boolean saveRasterLayer(RasterLayer<?> dataLayer, boolean addTimeStamp, boolean addTickCount)
	{
		String fileName = dataLayer.getStoragePath();
		String pathName = fileName.substring(0,fileName.length()-4);
		String exntName = fileName.substring(fileName.length()-4,fileName.length());
		String newFileName = pathName;
		
		if(addTimeStamp)
		{
			DateFormat dateFormat = new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss_SSS");
            java.util.Date date = new java.util.Date();
            String datetime = dateFormat.format(date);

            newFileName = newFileName + "_" + datetime;
		}
		
		if(addTickCount)
		{
			int tickCount = (int)RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
			newFileName = newFileName + "_" + tickCount;
		}

		newFileName = newFileName + exntName;
		
		return LandCoverDataHelper.heliaObj.saveLandscapeLayerData(dataLayer, newFileName);

	}

	/**
	 * Save a raster layer. It can add a time stamp and/or tick count to the file name.
	 * @param dataLayer The raster layer object
	 * @param addTimeStamp Whether to add a time stamp to the file name
	 * @param addTickCount Whether to add a tick count to the file name
	 * @return success or failure. If the layer has not been changed, it will not be saved. 
	 */
	public static String saveRasterLayer2(RasterLayer<?> dataLayer, boolean addTimeStamp, boolean addTickCount, boolean shortFileName)
	{
		String fileName = dataLayer.getStoragePath();
		String pathName = fileName.substring(0,fileName.length()-4);
		String exntName = fileName.substring(fileName.length()-4,fileName.length());
		String newFileName = "";
		
		if(addTimeStamp)
		{
			DateFormat dateFormat = new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss_SSS");
            java.util.Date date = new java.util.Date();
            String datetime = dateFormat.format(date);

            newFileName = newFileName + "_" + datetime;
		}
		
		if(addTickCount)
		{
			int tickCount = (int)RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
			newFileName = newFileName + "_" + tickCount;
		}

		newFileName = pathName + newFileName + exntName;
		
		LandCoverDataHelper.heliaObj.saveLandscapeLayerData(dataLayer, newFileName);
		
		if(shortFileName)
		{
			int index = Math.max(newFileName.lastIndexOf("/"), newFileName.lastIndexOf("\\"));
			return newFileName.substring(index + 1);
		}
		else
		{
			return newFileName;
		}
	}

	/**
	 * Save a raster layer. It can add a time stamp and/or tick count to the file name.
	 * @param dataLayer The raster layer object
	 * @param addTimeStamp Whether to add a time stamp to the file name
	 * @param addTickCount Whether to add a tick count to the file name
	 * @return success or failure. If the layer has not been changed, it will not be saved. 
	 */
	public static String saveRasterLayer3(RasterLayer<?> dataLayer, boolean addTimeStamp, boolean addTickCount, String postFixName, boolean shortFileName)
	{
		String fileName = dataLayer.getStoragePath();
		String pathName = fileName.substring(0,fileName.length()-4);
		String exntName = fileName.substring(fileName.length()-4,fileName.length());
		String newFileName = "";
		
		if(addTimeStamp)
		{
			DateFormat dateFormat = new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss_SSS");
            java.util.Date date = new java.util.Date();
            String datetime = dateFormat.format(date);

            newFileName = newFileName + "_" + datetime;
		}
		
		if(addTickCount)
		{
			int tickCount = (int)RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
			newFileName = newFileName + "_" + tickCount;
		}
		
		if(postFixName.isEmpty() || postFixName.equalsIgnoreCase(""))
		{
			
		}
		else
		{
			newFileName = newFileName + "_" + postFixName;
		}

		newFileName = pathName + newFileName + exntName;
		
		LandCoverDataHelper.heliaObj.saveLandscapeLayerData(dataLayer, newFileName);
		
		if(shortFileName)
		{
			int index = Math.max(newFileName.lastIndexOf("/"), newFileName.lastIndexOf("\\"));
			return newFileName.substring(index + 1);
		}
		else
		{
			return newFileName;
		}
	}
	
	
	/**
	 * 
	 * @param land the land object, could be a cell, parcel or subdivision,
	 * @param agent the agent who checks out the consumption bundle
	 * @return
	 */
	public static ConsumptionBundle getConsumptionBundle(LandUnit land, SimpleAgent agent)
	{
		// TODO: implement this using real data
		ConsumptionBundle cbObj = new ConsumptionBundle();
		RasterLayer<?> envLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.landUseTypeLayerIndex);
		RasterLayer<?> disLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.distanceLayerIndex);
		RasterLayer<?> luTypeLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AppParameters.landUseTypeLayerIndex);
		
		double P = 1- land.getValue(disLayer, LandAggregateMethod.min).doubleValue()/(1+ AppParameters.maxDistanceTo);
		LandCoverDataHelper.saObj.setCurrentOperator(BasicOperator.OperatorType.neighborhoodAverage);
		double A = LandCoverDataHelper.hegisObj.neighborhood(envLayer, land.getXCenterIndex(), land.getYCenterIndex(), GISConstants.neighborhoodCircle, AppParameters.neighborhoodSize, LandCoverDataHelper.saObj).doubleValue();
		
		double lotSize = land.getNumOfCells();
		lotSize = Math.sqrt(lotSize/18); // Scale down the lot size
		
		if(AppParameters.excludeCenterInNeighborhood)
		{
			Number centerVal = envLayer.getValueAt(land.getXCenterIndex(), land.getYCenterIndex());
			if(centerVal.intValue() == 1)
			{
				A = A - 1.0/((2*AppParameters.neighborhoodSize+1)*(2*AppParameters.neighborhoodSize+1));
			}
		}
		
		A = (A<0)?0:A;
		A=(A>1)?1:A;
		
		A = 1-A/1.25;
		//A = 1.0; // exclude open space factor now.
		
		if(land instanceof Parcel)
		{
			cbObj.addValueToObservedList(P);
			cbObj.addValueToObservedList(A);
			cbObj.addValueToObservedList(lotSize);
			
			//cbObj.addValueToPreferenceWeightList(AppParameters.utilityBeta);
			//cbObj.addValueToPreferenceWeightList(1 - AppParameters.utilityBeta);
			cbObj.addValueToPreferenceWeightList(agent.getPreferenceForProximity());
			cbObj.addValueToPreferenceWeightList(1.0 - 0.2*agent.getPreferenceForProximity());
			cbObj.addValueToPreferenceWeightList(1.0 - 0.8*agent.getPreferenceForProximity());
		}
		else if(land instanceof Subdivision)
		{
			
		}
		else
		{
			return null;
		}
		
		return cbObj;
	}

	/**
	 * Add a parcel object to the list
	 * @param parcelObj
	 * @return
	 */
	public static boolean addToParcelList(Parcel parcelObj)
	{
		if(parcelObj == null || parcelObj.getParcelID() < 0)
		{
			return false;
		}
		else
		{
			LandCoverDataHelper.parcelList.put(parcelObj.getParcelID(),parcelObj);
			return true;
		}
	}
	
	/**
	 * Add a subdivision object to the list
	 * @param subdivisionObj
	 * @return
	 */
	public static boolean addToSubdivisionList(Subdivision subdivisionObj)
	{
		if(subdivisionObj == null || subdivisionObj.getSubdivisionID() < 0)
		{
			return false;
		}
		else
		{
			LandCoverDataHelper.subdivisionList.put(subdivisionObj.getSubdivisionID(),subdivisionObj);
			return true;
		}
	}
	
	/**
	 * Calculate the X, Y indicies of a given cell ID
	 * @param cellID
	 * @return
	 */
	public static ArrayList<Integer> cellID2XY(int cellID)
	{
		ArrayList<Integer> rslt = new ArrayList<Integer>();
		RasterLayer<?> refLayerObj = (RasterLayer<?>)rootLayerGroup.getLandscapeLayer(AppParameters.referenceLayerIndex);
		int y = (int)(cellID/refLayerObj.getCols());
		int x = cellID - y*refLayerObj.getCols();
		rslt.add(x);rslt.add(y);
		return rslt;
	}
	
	/**
	 * @param x
	 * @param y
	 * @return
	 */
	public static int XY2CellID(int x, int y)
	{
		RasterLayer<?> refLayerObj = (RasterLayer<?>)rootLayerGroup.getLandscapeLayer(AppParameters.referenceLayerIndex);
		return x + y*refLayerObj.getCols();
	}
	
	/**
	 * Find the rook neighbors of a parcel land unit
	 * @return
	 */
	public static HashSet<Integer> findNeighbors(Parcel pObj, RasterLayer<?> luIDLayer)
	{
		HashSet<Integer> rslt = new HashSet<Integer>();
		
		ArrayList<Integer> cellCoord;
		int x, y, i;
		for(Integer cellID : pObj.getCellIDList())
		{
			cellCoord = cellID2XY(cellID);
			x = cellCoord.get(0);
			y = cellCoord.get(1);
			
			// Check up
			i = 0;
			while((y + i++) <= luIDLayer.getRows())
			{
				if(luIDLayer.getValueAt(x, y + i) == cellID) break;
				rslt.add(luIDLayer.getValueAt(x, y + i).intValue());
				break;
			}
			
			// check down
			i = 0;
			while((y + i--) <= luIDLayer.getRows())
			{
				if(luIDLayer.getValueAt(x, y + i) == cellID) break;
				rslt.add(luIDLayer.getValueAt(x, y + i).intValue());
				break;
			}
			
			// Check right
			i = 0;
			while((x + i++) <= luIDLayer.getCols())
			{
				if(luIDLayer.getValueAt(x + i, y) == cellID) break;
				rslt.add(luIDLayer.getValueAt(x + i, y).intValue());
				break;
			}
			
			// check left
			i = 0;
			while((x + i--) <= luIDLayer.getCols())
			{
				if(luIDLayer.getValueAt(x + i, y) == cellID) break;
				rslt.add(luIDLayer.getValueAt(x + i, y).intValue());
				break;
			}			
		}
		
		return rslt;
	}
	
	/**
	 * This method use a parcel ID layer to update the parcel list.
	 * It mainly assigns cells to each parcel. 
	 * Before calling this method, add all parcels to the parcel list.
	 * @param parcelIDLayer
	 * @return
	 */
	public static boolean updateParcelList(RasterLayer<?> parcelIDLayer)
	{
		if(parcelIDLayer == null || LandCoverDataHelper.parcelList.size() == 0) return false;
		
		int parcelID = 0, cellID = 0;
		Parcel pObj = null;
		
		for(int i = 0 ; i < parcelIDLayer.getCols(); i++)
		{
			for(int j = 0; j < parcelIDLayer.getRows(); j++)
			{
				parcelID = parcelIDLayer.getValueAt(i, j).intValue();
				if(parcelID == ModelParameters.notALayerValue) continue;
				if(!LandCoverDataHelper.parcelList.containsKey(parcelID)) continue;
				
				pObj = LandCoverDataHelper.parcelList.get(parcelID);
				
				if(pObj != null)
				{
					// There are no explicit cell ID, using their index is fine.
					cellID = i + j*parcelIDLayer.getCols();
					pObj.addCellID(cellID);
					
					// TODO: This method uses the first cell of the parcel as the location.
					// Change this to arithmetic center.
					if(!pObj.hasPositionInfo())
					{
						pObj.setXCenterIndex(i);
						pObj.setYCenterIndex(j);
						pObj.setXCenterPos(parcelIDLayer.getCellCenterX(i));
						pObj.setYCenterPos(parcelIDLayer.getCellCenterY(j));
					}
				}
			}
		}
		
		return true;
	}
	
	/**
	 * Update coordinates of all parcels using underlying cells.
	 * @param refLayer A raster reference layer for spatial scale
	 * @return
	 */
	public static boolean updateParcelCoordinates(RasterLayer<?> refLayer)
	{
		if(LandCoverDataHelper.parcelList == null) return false;
		int cellNum, cellID, x, y, xC, yC;
		double dX, dY;
		for(Parcel curObj : LandCoverDataHelper.parcelList.values())
		{
			xC = yC = x = y = 0;
			dX = dY = 0;
			cellNum = curObj.cellIDList.size();
			
			for(int j = 0; j < cellNum; j++)
			{
				cellID = curObj.cellIDList.get(j);
				y = (int)Math.floor(cellID/refLayer.getCols());
				x = cellID - (y*refLayer.getCols());
				
				xC += x; 
				yC += y;

				dX += refLayer.getCellCenterX(x);
				dY += refLayer.getCellCenterY(y);
			}
			
			curObj.setXCenterIndex(xC/cellNum);
			curObj.setYCenterIndex(yC/cellNum);
			
			curObj.setXCenterPos(dX/cellNum);
			curObj.setYCenterPos(dY/cellNum);
		}
		return true;
	}

	/**
	 * Add parcel IDs to each subdivision object.
	 * Before calling this method, add all subdivision objects into the list.
	 * @param parcelIDLayer
	 * @param subdivisionIDLayer
	 * @return
	 */
	public static boolean updateSubdivisionList(RasterLayer<?> parcelIDLayer, RasterLayer<?> subdivisionIDLayer)
	{
		if(parcelIDLayer == null || parcelList.size() == 0 || subdivisionIDLayer == null || subdivisionList.size() == 0) return false;
		
		int parcelID = 0, subdivisionID = 0;
		Subdivision sObj = null;
		
		for(int i = 0 ; i < subdivisionIDLayer.getCols(); i++)
		{
			for(int j = 0; j < subdivisionIDLayer.getRows(); j++)
			{
				subdivisionID = subdivisionIDLayer.getValueAt(i, j).intValue();
				parcelID = parcelIDLayer.getValueAt(i, j).intValue();
				sObj = subdivisionList.get(subdivisionID);
				
				// Add this parcel to the subdivision
				if(sObj != null)
				{
					sObj.addParcelID(parcelID);
				}
			}
		}
		
		return true;
	}

	/**
	 * Update the coordinates of all subdivisions using lower-level parcels
	 * @return true or false.
	 */
	public static boolean updateSubdivisionCoordinates()
	{
		if(LandCoverDataHelper.subdivisionList == null) return false;
		int num, x, y, xC, yC;
		double dX, dY;
		Parcel pObj;
		
		for(Subdivision curObj : LandCoverDataHelper.subdivisionList.values())
		{
			xC = yC = x = y = 0;
			dX = dY = 0;
			num = curObj.lowerLeverlIDList.size();
			Hashtable<Integer, Parcel> parcelsObj = curObj.getLowerLevelUnits();
			
			for(int pID : curObj.lowerLeverlIDList)
			{
				pObj = parcelsObj.get(pID);
				
				if(pObj == null) return false;
				
				x = pObj.getXCenterIndex();
				y = pObj.getYCenterIndex();

				xC += x;
				yC += y;
				
				dX += pObj.getXCenterPos();
				dY += pObj.getYCenterPos();
			}
			
			curObj.setXCenterIndex(xC/num);
			curObj.setYCenterIndex(yC/num);
			
			curObj.setXCenterPos(dX/num);
			curObj.setYCenterPos(dY/num);
		}
		return true;
	}
	
	/**
	 * Save a {@code RasterLayer}
	 * @param dataLayer the {@code RasterLayer} to be saved
	 * @param addTimeStamp if to add time stamp ( yyyy_MM_dd_HH_mm_ss_SSS ) to file name
	 * @return success or failure
	 */
	public static boolean saveRasterLayer(RasterLayer<?> dataLayer, boolean addTimeStamp)
	{
		if(addTimeStamp)
		{
			String fileName = dataLayer.getStoragePath();
			DateFormat dateFormat = new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss_SSS");
            java.util.Date date = new java.util.Date();
            String datetime = dateFormat.format(date);

            String timeFileName = fileName.substring(0,fileName.length()-4) + "_" + datetime + fileName.substring(fileName.length()-4,fileName.length());
            
            return LandCoverDataHelper.heliaObj.saveLandscapeLayerData(dataLayer,timeFileName);
		}
		else
		{
			return LandCoverDataHelper.heliaObj.saveLandscapeLayerData(dataLayer);
		}
	}
	
}
