###############################################################################
# This is a collection of functions used to plot multi-Dimensional parameter spaces.
# 
# Author: Shipeng Sun
###############################################################################

##########################################################################
# Function: To Draw Boxes for the next level categories
# It is a recursive function.
# @param x the X of the left-top corner
# @param y the Y of the left-top corner
# @param totWidth total width of the top-level box
# @param height the height of each layer of boxes
# @param dataVector the data used to draw the box, boxes are created column by column. 
# @param plotValues whether to plot values
# @param cexT enlarge factor the value texts
# @param minTextWidth minimum text width. If a box is narrower than this, value text will not be plotted
# @return none
##########################################################################
plotNextLevelBox <- function(x, y, totWidth, height, dataVector, plotValues = FALSE, cexT = 0.8, minTextWidth = 20) 
{
	vals <- dataVector[,1]
	fieldName <- names(dataVector)[1]
	
	totNum <- length(vals)
	uVals <- unique(vals)
	uNums <- length(uVals)
	
	#print(paste(dim(dataVector),sep=","))
	
	curLeftX <- x
	
	for(i in 1:uNums)
	{
		uVal <- uVals[i]
		uNum <- length(vals[vals==uVal])
		uWidth <- 0
		
		if(uNums > 1)
		{
			uNum <- length(vals[vals==uVal])
			uWidth <- totWidth*uNum/totNum
			rect(curLeftX,y-height,curLeftX+uWidth,y)
			
			if(plotValues && (uWidth >= minTextWidth))
			{
				text(curLeftX+uWidth/2.0, y - height/2, paste(uVal,"",sep=""),cex=cexT)
			}	
			
			#print(paste(curLeftX,y-height,curLeftX+uWidth,y,sep=","))
		}
		
		if(dim(dataVector)[2] >= 2)
		{
			plotNextLevelBox(curLeftX,y-height,uWidth,height,as.data.frame(dataVector[,2:dim(dataVector)[2]]),plotValues,cexT,minTextWidth)
		}
		
		curLeftX <- curLeftX + uWidth
	}
}

######################################################################################################################
# Standardize a data vector. 
# It also comes with values and coordinates used to plot the new axis labeled with the original values.
# This function is mainly used to plot multiple data vectors with varying data ranges.
# After standarlization, the X and Y ranges can be fixed so that some plotting works like legend, text are easier to do.
# @param dataVector a data vector to be standardized and transformed
# @param targetLBound the targeted lower bound of the standardization
# @param targetUBound the targeted upper bound
# @param sourceLBound the lower bound value of the dataVector (sometimes we might need to change this number to move up/right the whole plot)
# @param sourceLBound the upper bound value of the dataVector
# @param labelNum the number of labels. The actual number depends on 
# @return a list with $standardizedData, $actualLabelNum, $labelValues, $labelPositions,and $labelStrings
######################################################################################################################
standardize.dataVector <- function(dataVector, targetLBound = 0.0, targetUBound = 1.0, sourceLBound = NULL, sourceUBound = NULL, labelNum = 5)
{
	if(is.null(sourceLBound))
	{
		sourceLBound <- min(dataVector)
	}
	
	if(is.null(sourceUBound))
	{
		sourceUBound <- max(dataVector)
	}
	
	labelVals <- pretty(sourceLBound:sourceUBound, labelNum)
	labelPoss <- targetLBound + (labelVals - sourceLBound)*(targetUBound - targetLBound)/(sourceUBound - sourceLBound)
	
	labelVals <- labelVals[labelPoss >= targetLBound && labelPoss <= targetUBound]
	labelPoss <- labelVals[labelPoss >= targetLBound && labelPoss <= targetUBound]
	labelStrs <- paste(labelVals,"",sep="")
	
	rslt <- list()
	rslt$standardizedData <- (targetLBound + (dataVector - sourceLBound)*(targetUBound - targetLBound)/(sourceUBound - sourceLBound))
	rslt$actualLabelNum <- length(labelVals)
	rslt$labelValues <- labelVals
	rslt$labelPositions <- labelPoss
	rslt$labelStrings <- labelStrs
	
	return(rslt)
}
