/**
 * SLUCE 2 Project http://www.cscs.umich.edu/research/projects/sluce/
 */
package sluceii.agent;

import java.util.ArrayList;

import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.valueLayer.GridValueLayer;
import sluceii.agent.decisionmaking.HouseholdDecisionToBuyHelper;
import sluceii.agent.decisionmaking.HouseholdHousingSearchHelper;
import sluceii.land.exchange.LandBuyingProcess;
import sluceii.land.exchange.LandSellingProcess;
import sluceii.landmarket.LandBidRecord;
import sluceii.landmarket.LandListingRecord;
import sluceii.landmarket.LandMarketListing;
import sluceii.landmarket.LandBidRecord.BidRecordStatus;
import sluceii.landscape.LandCoverDataHelper;
import sluceii.landscape.Parcel;
import sluceii.params.Constants;
import sluceii.params.LogHelper;
import sluceii.params.LogHelperFactory;
import sluceii.app.aag.AAGAppParameters;
import sluceii.params.SimpleLogHelper;
import sluceii.params.LogHelper.LoggerType;
import sluceii.agent.utility.CobbDouglas;
import sluceii.agent.utility.UtilityStrategy;
import sluceii.app.aag.ContextCreator;
import umn.geog.helia.landscape.RasterLayer;

/**
 * Modified from Derek Robinson's Agent Class. This is the base class for all
 * household agents. Some of them are current homeowners. Some of them might
 * become homeowners later. Some of them may sell their house and buy another
 * house.
 * 
 * @author Shipeng Sun
 */
public class Household extends SimpleAgent implements LandBuyingProcess
{
	/**
	 * How many and which parcels/properties this agent has bid on.
	 * Currently, buyers can only bid on one single land/parcel/property.
	 */
	protected ArrayList<LandBidRecord> bidRecordList;
	
	/**
	 * A Static reference to the land market listing
	 */
	public static LandMarketListing landMarket;
	
	
	/**
	 * 
	 */
	private float pressureToBuy;
	
	/**
	 * The strategy used by this agent to calculate his utility
	 * 
	 */
	private UtilityStrategy utilStrategy;

	/**
	 * Export all transaction information to a file
	 * price, distance, preference, budget, etc.
	 */
	public static LogHelper transactionLogHelper;
	
	/**
	 * Default constructor
	 */
	public Household()
	{
		super();

		// TODO: load strategy according to parameters loaded from external files 
		this.utilStrategy = new CobbDouglas();
		this.pressureToBuy = 1.0f;
	}

	/**
	 * @return the utilStrategy
	 */
	public UtilityStrategy getUtilStrategy()
	{
		return utilStrategy;
	}

	/**
	 * @param utilityStrategy
	 * @uml.property name="UtilityStrategy"
	 */
	public void setUtilityStrategy(UtilityStrategy utilityStrategy)
	{
		this.utilStrategy = utilityStrategy;
	}

	/**
	 * @return the pressureToBuy
	 */
	public float getPressureToBuy()
	{
		return pressureToBuy;
	}

	/**
	 * @param pressureToBuy the pressureToBuy to set
	 */
	public void setPressureToBuy(float pressureToBuy)
	{
		this.pressureToBuy = pressureToBuy;
	}

	/* (non-Javadoc)
	 * @see sluceii.agent.SimpleAgent#step()
	 * 
	 * General procedure for households bid/buy land/house
	 * 1. Decision to buy/bid for new land/house 
	 * 		Inputs: socio-economic variables, utility strategy, neighborhood conditions, ...
	 * 		Output: yes/no, or a probability
	 * 2. Search market list and find candidates
	 * 		Inputs: market list (built by rural land owners who want to sell)
	 * 		Output: Parcel list with one or more items
	 * 3. Bid on Parcels
	 * 		Inputs: Parcel list after search
	 * 				How many? (one or more)
	 * 				How much for each of them
	 * 		Action: Update the bid record listing
	 * 
	 * ************************************************
	 * The following are actions triggered by sellers.
	 * 4. Win one or more bids.
	 * 		Inputs: Seller accept offer (Yes/No)
	 * 		Action: Take the best bid
	 * 5. Cancel other bids.
	 * 		Action: update the market listing
	 * 6. After bid
	 * 		Action: Relocate, update resources, and apply new management strategy
	 */
	@Override
	@ScheduledMethod(start = 1, interval = 1, shuffle = true, priority = 9)
	public void step()
	{
		try
		{
			// TODO Auto-generated method stub
			super.step();
			HouseholdDecisionToBuyHelper decisionHelper = new HouseholdDecisionToBuyHelper();
			decisionHelper.setLandBuyer(this);
			
			if(decisionHelper.isToBuy())
			{
				boolean bidSuccess = false;
				int numTries = 0;
				int numSuccess = 0;
				do
				{
					// This can also be implemented in search market listing. Do NOT return those parcels 
					// that have reached maximum allowed bids.
					ArrayList<LandListingRecord> parcelList = this.searchMarketListing(AAGAppParameters.numSearchParcel);
					if(parcelList.size()>0)
					{
						for(LandListingRecord candidateParcel : parcelList)
						{
							bidSuccess = bidOnParcel(candidateParcel);
							if(bidSuccess) numSuccess++; // Only bid for one single parcel for one time
							if(numSuccess >= AAGAppParameters.maxBidCapacity) break;
						}
					}
	
				} while (!bidSuccess && numTries++ < AAGAppParameters.maxBiddingTries);
			}
			else
			{
				// Increase the year that this household in current status
			}
		}
		catch(Exception e)
		{
			System.err.println("# Error in Household: step - " + e.toString());
			e.printStackTrace(System.err);
		}
	}

	/**
	 * This is the action taken by this household agent after his/her bid is accepted by the seller.
	 * ************************************************
	 * The following are actions triggered by sellers.
	 * 4. Win one or more bids.
	 * 		Inputs: Seller accept offer (Yes/No)
	 * 		Action: Take the best bid
	 * 5. Cancel other bids.
	 * 		Action: update the market listing
	 * 6. After bid
	 * 		Action: Relocate and Apply new management strategy
	 * 
	 * TODO: Is it possible that a buyer receive multiple offers from the sellers?
	 * 		In that case, the buyer must be able to save all offers and pick the 'best' one. 
	 */
	public void afterBidAction(LandBidRecord landBid)
	{
		// Change the current bid parcel from rural to urban use
		RasterLayer<?> luTypeLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AAGAppParameters.landUseTypeLayerIndex);
		Parcel pObj = landBid.getListingRecord().getParcelObj();
		pObj.putValue(luTypeLayer, 1);
		
		RasterLayer<?> tpcLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AAGAppParameters.finalPriceLayerIndex);
		pObj.putValue(tpcLayer, landBid.getBidPrice());
		
		GridValueLayer vlLUType = (GridValueLayer)SimpleAgent.context.getValueLayer(Constants.LANDUSETYPE);
		pObj.putValue(vlLUType, tpcLayer);
		
		// Save land price, agent's preference for proximity, distance, open space amenity, utility
		if(AAGAppParameters.exportTransactionRecords)
		{
			// Only for GUI version
//			if(Household.transactionLogHelper == null)
//			{
//				Household.transactionLogHelper = new SimpleLogHelper(AAGAppParameters.dataFolderName + "/Log/trans.csv",2,true);
//			}
			
			if(Household.transactionLogHelper == null && AAGAppParameters.exportTransactionRecords)
			{
				Household.transactionLogHelper = LogHelperFactory.getTransactionLogger(2);
			}
			
			// dist, open space amenity, tranCost, Y, U
			double dist  = landBid.getLandProperty(0).doubleValue();
			double osa   = landBid.getLandProperty(1).doubleValue();
			double tcost = landBid.getLandProperty(2).doubleValue();
			double tbudg = landBid.getLandProperty(3).doubleValue() + tcost;
			double util  = landBid.getLandProperty(4).doubleValue();
			
			double bidPrice = landBid.getBidPrice();
			
			double x = landBid.getParcelObj().getXCenterPos();
			double y = landBid.getParcelObj().getYCenterPos();
			
			
			if(transactionLogHelper.getLoggerType() == LoggerType.txtLogger)
			{
				// AgentID, x, y, bidPrice, preference for proximity, distance, open space amenity, tranCost, Y (total budget), U 
				((SimpleLogHelper)Household.transactionLogHelper).writeLogWithoutTime(getAgentID() + "," + RunEnvironment.getInstance().getCurrentSchedule().getTickCount() + "," + x + "," + y + "," + bidPrice + "," + getPreferenceForProximity() + "," + dist + "," + osa + "," + tcost + "," + tbudg + "," + util, 0);
			}
			else if(transactionLogHelper.getLoggerType() == LoggerType.mySQLLogger)
			{
				// AgentID, bidPrice, preference for proximity, budget, 
				Household.transactionLogHelper.logTransaction(0, RunEnvironment.getInstance().getCurrentSchedule().getTickCount(), getAgentID(), 
						bidPrice, getPreferenceForProximity(), dist, osa, tcost, tbudg, util);
			}
			
		}
		
		// Update the location of the agent
		agentGrid.moveTo(this, pObj.getXCenterIndex(),pObj.getYCenterIndex());
		this.setPosition(pObj.getXCenterIndex(), pObj.getYCenterIndex());
		
		this.pressureToBuy = 0.0f;
		ContextCreator.activity = true;
		
		ContextCreator.setTotalTransactionNum(++ContextCreator.totalTansactionNum);
	}
	
	/**
	 * This method implements the search process for a potential home buyer.
	 * 
	 * TODO: How to search the vacancy list. 
	 * Do we assume complete information or incomplete information?
	 * 
	 * @param number the number returned 
	 * @return A List of candidate parcels for bidding
	 */
	private ArrayList<LandListingRecord> searchMarketListing(int number)
	{
		// How many parcels/houses the buyer want to evaluate?
		HouseholdHousingSearchHelper searchHelper = new HouseholdHousingSearchHelper(this);
		ArrayList<LandListingRecord> parcelList = searchHelper.searchProperties(number);
		
		return parcelList;
	}
	
	/**
	 * Bid on the specified parcel object.
	 * Depending on the limit on the number of bids on a parcel, this may or may not succeed.
	 * 
	 * @param parcelObj
	 * @return if the operation is successful
	 */
	public boolean bidOnParcel(LandListingRecord listedParcelObj)
	{
		// The buyer cannot bid on this one.
		if(landMarket.getOpenBidNumber(listedParcelObj.getParcelID())>= listedParcelObj.getMaxBidsAllowed())
		{
			return false;
		}
		
		HouseholdDecisionToBuyHelper decisionHelper = new HouseholdDecisionToBuyHelper();
		decisionHelper.setLandBuyer(this);
		
		// dist, open space amenity, tranCost, Y, b, U, wtp, bidPrice
		ArrayList<Number> landProperties = decisionHelper.calcLandProperties(listedParcelObj.getParcelObj());
		float wtp = landProperties.get(6).floatValue();
		float bidPrice = landProperties.get(7).floatValue();
		
		//float wtp = (float)decisionHelper.calcWTPPrice(listedParcelObj.getParcelObj());
		//float bidPrice = (float)decisionHelper.calcBidPrice(listedParcelObj.getParcelObj());
		
		if(bidPrice < listedParcelObj.getWillingToAccept()) return false;
		
		LandBidRecord bidRecord = new LandBidRecord();
		bidRecord.setListingRecObj(listedParcelObj);
		
		//System.out.println(listedParcelObj.getParcelObj().getXCenterIndex() + "," + listedParcelObj.getParcelObj().getYCenterIndex());
		
		bidRecord.setBidPrice(bidPrice);
		bidRecord.setWillingToPay(wtp);
		bidRecord.setRecordStatus(BidRecordStatus.open);
		bidRecord.setBuyer(this);
		
		if (AAGAppParameters.exportTransactionRecords)
		{
			// dist, open space amenity, tranCost, Y (total budget), U
			bidRecord.addLandProperty(landProperties.get(0));
			bidRecord.addLandProperty(landProperties.get(1));
			bidRecord.addLandProperty(landProperties.get(2));
			bidRecord.addLandProperty(landProperties.get(3));	
			bidRecord.addLandProperty(landProperties.get(5));
		}
		
		landMarket.addBidRecord(bidRecord);
		
		if(AAGAppParameters.preInfoAvailable)
		{
			RasterLayer<?> luTypeLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AAGAppParameters.landUseTypeLayerIndex);
			Parcel pObj = listedParcelObj.getParcelObj();
			pObj.putValue(luTypeLayer, 1);
		}
		
		return true;
		
	}

	/**
	 * The seller has accepted this buyer/household's bid
	 */
	public void bidAccepted(LandBidRecord landBidRecord)
	{
		// Add this property/land parcel to this agent's property list
		this.propertyList.add(landBidRecord.getParcelObj());
		landBidRecord.setRecordStatus(BidRecordStatus.accepted);
		
		// Notify seller to clean up.
		// The seller must have implemented LandSellingProcess interface.
		LandSellingProcess sellerInterface = (LandSellingProcess)landBidRecord.getListingRecord().getSeller();
		sellerInterface.afterBidAction(landBidRecord);
		
		this.captialResource.subtractAsset(landBidRecord.getBidPrice());
		
		this.afterBidAction(landBidRecord);
		
		// Cancel other bids, necessary?
		if(bidRecordList != null)
		{
			bidRecordList.clear();
		}
	}

	@Override
	public void bidRejected(LandBidRecord landBidRecord)
	{
		// TODO Auto-generated method stub
		landBidRecord.setRecordStatus(BidRecordStatus.rejected);
		bidRecordList.remove(landBidRecord);
		
		if(bidRecordList.isEmpty())
		{
			//instantly bid on other parcels
			//this.step();
		}
	}
}
