package sluceii.agent;

import java.util.ArrayList;

import repast.simphony.annotate.AgentAnnot;
import repast.simphony.context.Context;
import repast.simphony.space.grid.Grid;
import sluceii.agent.resources.AssetResource;
import sluceii.agent.resources.AssetResource.AssetType;
import sluceii.landscape.LandCover;
import umn.geog.helia.landscape.RasterLayer;

/**
 * @author Derek Robinson
 * Modified by Shipeng Sun
 *
 */
@AgentAnnot(displayName = "SimpleAgent")
public class SimpleAgent
{
	/**
	 * A reference to the context
	 */
	public static Context<SimpleAgent> context;
	
	/**
	 * A reference to the agent grid, through which agents can be found.
	 */
	public static Grid<SimpleAgent> agentGrid;


	/**
	 * A unique identifier of the agent
	 */
	protected int agentID;
	
	/**
	 * the indices of X, Y in the grid
	 */
	protected int xIndex, yIndex;
	
	/**
	 * The actual coordinates of X, Y
	 */
	protected float xPos, yPos;
	
	/**
	 * The property/parcel list of this owner
	 */
	protected ArrayList<LandCover> propertyList;
	
	/**
	 * this is a static reference layer used to convert real coordinates to/from grid coordinates
	 */
	public static RasterLayer<?> referenceLayer;
	
	/**
	 * In this version, a simple capital asset is used to decide if an agent affords a parcel/property.
	 */
	protected AssetResource captialResource;
	
	/**
	 * The beta in C-D utility function
	 */
	protected double preferenceForProximity;
	
	/**
	 * Default constructor
	 */
	public SimpleAgent()
	{
		this.captialResource = new AssetResource(AssetType.capital);
		this.captialResource.setAssetQuantity(0);
		
		this.propertyList = new ArrayList<LandCover>();
	}
		
	/**
	 * @return the agentID
	 */
	public int getAgentID()
	{
		return agentID;
	}

	/**
	 * @return
	 */
	public int getX()
	{
		return this.xIndex;
	}
	
	/**
	 * @return
	 */
	public int getY()
	{
		return this.yIndex;
	}
	
	/**
	 * @param agentID the agentID to set
	 */
	public void setAgentID(int agentID)
	{
		this.agentID = agentID;
	}

	/**
	 * Set the position of this agent in the landscape/grid
	 * When possible, the row/column index will be updated accordingly.
	 * @param x
	 * @param y
	 */
	public void setPosition(float x, float y)
	{
		this.xPos = x;
		this.yPos = y;
		
		if(SimpleAgent.referenceLayer != null)
		{
			this.xIndex = referenceLayer.getCellCol(x);
			this.yIndex = referenceLayer.getCellRow(y);
		}
	}
	
	/**
	 * Set the index of row and column in the virtual grid.
	 * When possible, the real coordinates will be updated accordingly.
	 * @param xIndex
	 * @param yIndex
	 */
	public void setPosition(int xIndex, int yIndex)
	{
		this.xIndex = xIndex;
		this.yIndex = yIndex;
		
		if(SimpleAgent.referenceLayer != null)
		{
			this.xPos = (float)SimpleAgent.referenceLayer.getCellCenterX(xIndex);
			this.yPos = (float)SimpleAgent.referenceLayer.getCellCenterY(yIndex);
			
		}
	}
	
	/**
	 * @return the propertyList
	 */
	public ArrayList<LandCover> getPropertyList()
	{
		return propertyList;
	}

	/**
	 * @param propertyList the propertyList to set
	 */
	public void setPropertyList(ArrayList<LandCover> propertyList)
	{
		this.propertyList = propertyList;
	}

	/**
	 Schedule the step method for agents. The method is scheduled starting at 
	 tick one with an interval of 1 tick. Specifically, the step starts at 1,
	 and recurs at 2,3,4,...etc
	 
	 This shuffle property of this method is turned on, so agents can have random orders to 
	 execute this method. Otherwise, some agents always take actions earlier than others.
	 */
	public void step()
	{
		// Override by subclasses
	}

	public int isAgent()
	{
		return 0;
	}

	public int isCell()
	{
		return 0;
	}
	
	/**
	 * @return the captialResource
	 */
	public AssetResource getCaptialResource()
	{
		return captialResource;
	}

	/**
	 * @param captialResource the captialResource to set
	 */
	public void setCaptialResource(AssetResource captialResource)
	{
		this.captialResource = captialResource;
	}

	/**
	 * @return the preferenceForProximity
	 */
	public double getPreferenceForProximity()
	{
		return preferenceForProximity;
	}

	/**
	 * @param preferenceForProximity the preferenceForProximity to set
	 */
	public void setPreferenceForProximity(double preferenceForProximity)
	{
		this.preferenceForProximity = preferenceForProximity;
	}

	/**
	 * Update the agent's resources after buying or selling a parcel/house/property
	 */
	protected void updateResources()
	{
		
	}
	
	/**
	 * @param landProperty
	 * @return
	 */
	public boolean addProperty(LandCover landProperty)
	{
		return propertyList.add(landProperty);
	}
	
	/**
	 * @param landProperty
	 * @return
	 */
	public boolean removeProperty(LandCover landProperty)
	{
		return this.propertyList.remove(landProperty);
	}
}