/**
 * SLUCE 2 Project
 * http://www.cscs.umich.edu/research/projects/sluce/
 */
package sluceii.agent.decisionmaking;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.TreeMap;

import cern.jet.random.Uniform;

import repast.simphony.random.RandomHelper;
import sluceii.agent.Household;
import sluceii.agent.decisionmaking.LinkedComparator.LinkedComparatorType;
import sluceii.agent.utility.ConsumptionBundle;
import sluceii.agent.utility.UtilityStrategy;
import sluceii.app.aag.AAGAppParameters;
import sluceii.app.aag.AAGMiscFunctions;
import sluceii.landmarket.LandListingRecord;
import sluceii.landscape.LandCover;
import sluceii.landscape.LandCoverDataHelper;
import sluceii.landscape.LandCover.LandAggregateMethod;
import umn.geog.helia.landscape.RasterLayer;

/**
 * @author Shipeng Sun
 *
 */
public class HouseholdHousingSearchHelper implements HousingSearch
{
	public enum SpatialSearchStrategy {random, neighborhood, exponential};
	/**
	 * the agent searching for vacancies
	 */
	protected Household housingSearcher;
	
	/**
	 * Default constructor
	 */
	public HouseholdHousingSearchHelper()
	{
		// TODO Auto-generated constructor stub
	}
	
	public HouseholdHousingSearchHelper(Household searcher)
	{
		this.housingSearcher = searcher;
	}
	

	/**
	 * @return the housingSearcher
	 */
	public Household getHousingSearcher()
	{
		return housingSearcher;
	}

	/**
	 * @param housingSearcher the housingSearcher to set
	 */
	public void setHousingSearcher(Household housingSearcher)
	{
		this.housingSearcher = housingSearcher;
	}

	/* (non-Javadoc)
	 * @see sluceii.agent.decisionmaking.HousingSearch#searchProperties(int)
	 */
	@Override
	public ArrayList<LandListingRecord> searchProperties(int num)
	{
		// Search the listing records stored in landMarket
		ArrayList<LandListingRecord> allList = Household.landMarket.getToSellParcelList();
		allList.trimToSize();
		
		// Search using spatial rules to get a sub-sample
		ArrayList<LandListingRecord> spatialList = spatialSearch(allList,AAGAppParameters.numSearchParcel, SpatialSearchStrategy.random);
		spatialList.trimToSize();
		
		// Calculate utilities
		UtilityStrategy utilStrategy = this.housingSearcher.getUtilStrategy();
		ArrayList<Double> utilList = calcUtilities(spatialList,utilStrategy);
		utilList.trimToSize();
		
		ArrayList<LandListingRecord> highUtilityList = this.sortArraysWithDictionary(utilList, spatialList, true);
		highUtilityList.trimToSize();
		
		ArrayList<LandListingRecord> rsltList = new ArrayList<LandListingRecord>();
		LandListingRecord llRecord = null;
		Iterator<LandListingRecord> itr = highUtilityList.iterator();
		int cnt = 0;
		
		// Don't include those parcels that have already had maximum bid numbers.
		while(itr.hasNext() && cnt < num)
		{
			llRecord = itr.next();
			if(Household.landMarket.getOpenBidNumber(llRecord.getParcelID()) < llRecord.getMaxBidsAllowed())
			{
				rsltList.add(llRecord);
				cnt++;
			}
		}
			
		return rsltList;		
	}
	
	/**
	 * Calculate the utilities of a list of parcel/properties/houses
	 * @param spatialList
	 * @param utilStrategy
	 * @return A list of utilities
	 */
	private ArrayList<Double> calcUtilities(ArrayList<LandListingRecord> reclList, UtilityStrategy utilStrategy)
	{
		// TODO Auto-generated method stub
		Iterator<LandListingRecord> itRec = reclList.iterator();
		ArrayList<Double> rsltUtilities = new ArrayList<Double>();
		LandListingRecord curRecord = null;
		LandCover landObj = null;
		ConsumptionBundle cBundle = null;
		
		while(itRec.hasNext())
		{
			curRecord = itRec.next();
			if(curRecord != null)
			{
				landObj = curRecord.getParcelObj();
				//cBundle = LandCoverDataHelper.getConsumptionBundle(landObj, this.housingSearcher);
				cBundle = AAGMiscFunctions.getConsumptionBundle(landObj, this.housingSearcher);
				
				rsltUtilities.add(utilStrategy.getUtility(cBundle));
			}
		}
		
		return rsltUtilities;
	}
	
	/**
	 * @param land
	 * @param utilStrategy
	 * @return
	 */
	public double calcUtility(LandListingRecord land, UtilityStrategy utilStrategy)
	{
		LandCover landObj = null;
		ConsumptionBundle cBundle = null;

		landObj = land.getParcelObj();
		//cBundle = LandCoverDataHelper.getConsumptionBundle(landObj, this.housingSearcher);
		cBundle = AAGMiscFunctions.getConsumptionBundle(landObj, this.housingSearcher);
		
		double rslt = utilStrategy.getUtility(cBundle);
		return rslt;
	}

	/**
	 * Right now, only random strategy is implemented.
	 * Parcels that have been bid over the limit and are not affordable for this buyer will be 
	 * bypassed.
	 * 
	 * TODO: 
	 * 
	 * @param recList
	 * @param num
	 * @param searchStrategy
	 * @return
	 */
	protected ArrayList<LandListingRecord> spatialSearch(ArrayList<LandListingRecord> recList, int num, SpatialSearchStrategy searchStrategy)
	{
		// All recorded will be sorted according to a probability associated with them.
		// Generate probabilities and associate them with listing records
		double rndNum = 0.0, dist = 0.0, transportCost = 0.0;
		LandListingRecord curRecord = null;
		RasterLayer<?> distLayer = (RasterLayer<?>)LandCoverDataHelper.rootLayerGroup.getLandscapeLayer(AAGAppParameters.distanceLayerIndex);
		ArrayList<Double> probArray = new ArrayList<Double>(recList.size());
		probArray.clear();
		
		switch (searchStrategy)
		{
			case random:
			{
				try
				{
					Iterator<LandListingRecord> itRec = recList.iterator();
					Uniform ufmRnd = RandomHelper.createUniform();
					
					while(itRec.hasNext())
					{
						// As the TreeMap sort elements from lowest to highest. It is easier to interpret when we
						// use 1- random, although the effects are the same.
						rndNum = 1- ufmRnd.nextDoubleFromTo(0, 1);
						curRecord = itRec.next();
						
						if(Household.landMarket.getCurrentBidsNumber(curRecord.getParcelID())>= curRecord.getMaxBidsAllowed())
						{
							// The parcel that has been bid will have little chance being selected.
							rndNum += 1.0;
						}

						dist = curRecord.getParcelObj().getValue(distLayer, LandAggregateMethod.average).doubleValue();
						transportCost = dist * AAGAppParameters.unitTransportCost;
						
						if(curRecord.getAskPrice()+ transportCost >= this.housingSearcher.getCaptialResource().getAssetQuantity())
						{
							rndNum += 1.0;
						}
						
						probArray.add(rndNum);
					}		
				}
				catch(Exception e)
				{
					e.printStackTrace();
				}
				break;
			}
			case neighborhood:
			{
				break;
			}
			case exponential:
			{
				break;
			}
			default:
			{
				
			}
		}
		
		// get the elements with highest probabilities and add it the result list
		ArrayList<LandListingRecord> rsltList = this.sortArraysWithDictionary(probArray, recList, false, Math.min(num,recList.size()));
		return rsltList;
	}
	
	private <T> ArrayList<T> sortArraysWithDictionary(ArrayList<Double> criteriaArray, ArrayList<T> targetArray, boolean isDecendingOrder)
	{
		// TODO Auto-generated method stub
		return sortArraysWithDictionary(criteriaArray, targetArray, isDecendingOrder,criteriaArray.size());
	}
	
	/**
	 * @param <T>
	 * @param criteriaArray
	 * @param targetArray
	 * @param isDecendingOrder
	 * @return
	 */
	private <T> ArrayList<T> sortArraysWithDictionary(ArrayList<Double> criteriaArray, ArrayList<T> targetArray, boolean isDecendingOrder, int num)
	{
		if(criteriaArray == null || targetArray == null || targetArray.size() != criteriaArray.size() || num > targetArray.size())
		{
			return null;
		}
		
		TreeMap<Double, ArrayList<T>> sTree = new TreeMap<Double, ArrayList<T>>();
		double key;
		ArrayList<T> val;
		
		for(int i = 0; i <  criteriaArray.size(); i++)
		{
			key = criteriaArray.get(i);
			
			if(isDecendingOrder) key *= -1.0;
			if(sTree.containsKey(key))
			{
				val = sTree.get(key);
				val.add(targetArray.get(i));
			}
			else
			{
				val = new ArrayList<T>();
				val.add(targetArray.get(i));
				sTree.put(key, val);
			}
		}
		
		ArrayList<T> rslt = new ArrayList<T>(criteriaArray.size());
		Collection<ArrayList<T>> vals = sTree.values();
		
		int cnt = 0;
		for(ArrayList<T> valObj : vals)
		{
			rslt.addAll(valObj);
			cnt += valObj.size();
			
			if(cnt >= num) break;
		}
		
		while(rslt.size() > num)
		{
			rslt.remove(rslt.size()-1);
		}
		
		rslt.trimToSize();
		return rslt;
	}
	
	/**
	 * @param <T>
	 * @param criteriaArray
	 * @param targetArray
	 * @param isDecendingOrder
	 * @param num
	 * @return
	 */
	private <T> ArrayList<T> sortArraysWithLinkedSorting(ArrayList<Double> criteriaArray, ArrayList<T> targetArray, boolean isDecendingOrder, int num)
	{
		if(criteriaArray == null || targetArray == null || targetArray.size() != criteriaArray.size() || num > targetArray.size())
		{
			return null;
		}
		
		Integer indexArray[] = new Integer[targetArray.size()];
		for(int i = 0; i < indexArray.length; i++)
		{
			indexArray[i] = i;
		}
		
		// This is always in ascending order
		Arrays.sort(indexArray, new LinkedComparator(LinkedComparatorType.index, criteriaArray));
		
		ArrayList<T> rslt = new ArrayList<T>(num);
		
		for(int i = 0; i < num; i++)
		{
			if(isDecendingOrder)
			{
				rslt.add(targetArray.get(indexArray[indexArray.length - 1- i]));
			}
			else
			{
				rslt.add(targetArray.get(indexArray[i]));
			}
		}
		
		return rslt;
	}
	
	@SuppressWarnings("unused")
	private <T> ArrayList<T> sortArraysWithLinkedSorting(ArrayList<Double> criteriaArray, ArrayList<T> targetArray, boolean isDecendingOrder)
	{
		return sortArraysWithLinkedSorting(criteriaArray, targetArray, isDecendingOrder, targetArray.size());
	}
}
