/**
 * SLUCE 2 Project http://www.cscs.umich.edu/research/projects/sluce/
 */
package sluceii.landmarket;

import java.util.ArrayList;

import sluceii.agent.SimpleAgent;
import sluceii.land.exchange.LandBuyingProcess;
import sluceii.landscape.Parcel;

/**
 * This is the land bid record. It has the parcel/land to be bid, WTP, WTA,
 * buyer, and seller. It may or may not become a transaction. All bid records
 * are kept through the program. They are also indexed by parcel ID.
 * 
 * @see LandMarketListing
 * @see LandListingRecord
 * 
 * @author Shipeng Sun
 */
public class LandBidRecord
{
	public enum BidRecordStatus {open, accepted, rejected, pending, completed};
	
	/**
	 * A unique identifier for the bid record.
	 */
	protected int recordID;

	private static int maxRecordID = 0;

	/**
	 * buyer is a simple agent, but it must implement LandBuyingProcess Interface
	 */
	private SimpleAgent buyer;
	
	/**
	 * The listing record. It contains the information about seller and landcover.
	 */
	protected LandListingRecord listingRecObj;

	/**
	 * The bid price offered by buyer 
	 */
	protected float bidPrice;
	
	/**
	 * The WTP from buyer
	 */
	protected float willingToPay;
	
	/**
	 * Specific land properties, such as distance to downtown, open space amenity at the bidding time,
	 * and others as necessary.
	 * 
	 * The nature of these properties is subject to agents' interpretation.
	 */
	protected ArrayList<Number> landProperties;

	/**
	 * The current status of this bid
	 */
	protected BidRecordStatus recordStatus;
	
	// TODO: add a time/status history for the record.
	
	/**1
	 * 
	 */
	public LandBidRecord()
	{
		// TODO Auto-generated constructor stub
		setRecordID(++LandBidRecord.maxRecordID);
		
		// Default status is open.
		recordStatus = BidRecordStatus.open;
		landProperties = null; // Create this on need. Otherwise, it takes too much space.
	}

	/**
	 * @return the recordID
	 */
	public int getRecordID()
	{
		return recordID;
	}

	/**
	 * @param recordID
	 *            the recordID to set
	 */
	public void setRecordID(int recordID)
	{
		this.recordID = recordID;
	}

	/**
	 * @return the bidPrice
	 */
	public float getBidPrice()
	{
		return bidPrice;
	}

	/**
	 * @param bidPrice
	 *            the bidPrice to set
	 */
	public void setBidPrice(float bidPrice)
	{
		this.bidPrice = bidPrice;
	}

	/**
	 * @return the willingToPay
	 */
	public float getWillingToPay()
	{
		return willingToPay;
	}

	/**
	 * @param willingToPay
	 *            the willingToPay to set
	 */
	public void setWillingToPay(float willingToPay)
	{
		this.willingToPay = willingToPay;
	}

	/**
	 * @return the willingToAccept
	 */
	public float getWillingToAccept()
	{
		return this.listingRecObj.getWillingToAccept();
	}

	/**
	 * @return the listingRecObj
	 */
	public LandListingRecord getListingRecord()
	{
		return listingRecObj;
	}

	/**
	 * @param listingRecObj
	 *            the listingRecObj to set
	 */
	public void setListingRecObj(LandListingRecord listingRecObj)
	{
		this.listingRecObj = listingRecObj;
	}

	/**
	 * @return the buyer
	 */
	public SimpleAgent getBuyer()
	{
		return buyer;
	}
	
	/**
	 * @param <T>
	 * @param buyer
	 */
	public <T extends SimpleAgent & LandBuyingProcess> void setBuyer(T buyer)
	{
		this.buyer = buyer;
	}

	/**
	 * @return
	 */
	public int getParcelID()
	{
		return listingRecObj.getParcelID();
	}

	/**
	 * @return
	 */
	public Parcel getParcelObj()
	{
		// TODO Auto-generated method stub
		return this.listingRecObj.getParcelObj();
	}

	/**
	 * @return the recordStatus
	 */
	public BidRecordStatus getRecordStatus()
	{
		return recordStatus;
	}

	/**
	 * @param recordStatus the recordStatus to set
	 */
	public void setRecordStatus(BidRecordStatus recordStatus)
	{
		this.recordStatus = recordStatus;
	}
	
	/**
	 * @param val
	 */
	public void addLandProperty(Number val)
	{
		if(landProperties != null)
		{
			landProperties.add(val);
		}
		else
		{
			landProperties = new ArrayList<Number>();
			landProperties.add(val);
		}
	}
	
	public Number getLandProperty(int index)
	{
		if(this.landProperties != null)
		{
			return landProperties.get(index);
		}
		else
		{
			return null;
		}
	}
}
