/**
 * SLUCE 2 Project
 * http://www.cscs.umich.edu/research/projects/sluce/
 * @version 1.0
 */
package sluceii.landmarket;

import java.util.ArrayList;
import java.util.Hashtable;

import sluceii.agent.SimpleAgent;
import sluceii.land.exchange.LandSellingProcess;
import sluceii.landmarket.LandBidRecord.BidRecordStatus;
import sluceii.landmarket.LandListingRecord.ListingRecordStatus;
import sluceii.landscape.Parcel;

/**
 * This class represents the land market listing. All sellers put their land-to-sell on this listing.
 * Potential buyers bid on this list.
 * 
 * @author Shipeng
 *
 */
public class LandMarketListing
{
	/**
	 * This ID is used to serialize the class to external hard drive.
	 */
	 private static final long serialVersionUID = 8228686603144586836L;
	
	/**
	 * Vacant (available, on-the-market) parcel/house list
	 */
	protected ArrayList<LandListingRecord> toSellParcelList;
	
	/**
	 * The market listing records organized by seller ID (agent ID)
	 * No record will be deleted from this hash table.
	 * NOTES: the toSellParcelList will remove records whose land transaction has been completed.
	 */
	protected Hashtable<Integer, ArrayList<LandListingRecord>> sellerIndexedListingRecords;
	
	/**
	 * This hash table organizes bid records using parcel ID.
	 * Key: parcel ID
	 * Value: all bid records on this parcel 
	 */
	protected Hashtable<Integer, ArrayList<LandBidRecord>> parcelIndexedBidRecords;
	
	/**
	 * All parcels indexed by their ID.
	 */
	protected Hashtable<Integer, Parcel> parcelList;
	
	/**
	 * Default constructor
	 */
	public LandMarketListing()
	{
		// TODO Auto-generated constructor stub
		this.toSellParcelList = new ArrayList<LandListingRecord>();
		this.parcelIndexedBidRecords = new Hashtable<Integer, ArrayList<LandBidRecord>>();
		this.sellerIndexedListingRecords = new Hashtable<Integer, ArrayList<LandListingRecord>>();
		this.parcelList = null;
	}

	/**
	 * @param parcelObj
	 */
	public void addParcelOnMarket(LandListingRecord listObj)
	{
		// TODO Auto-generated method stub
		if(listObj == null) return;
		
		this.toSellParcelList.add(listObj);
		
		if(this.sellerIndexedListingRecords.containsKey(listObj.getSeller().getAgentID()))
		{
			ArrayList<LandListingRecord> recordList = this.sellerIndexedListingRecords.get(listObj.getSeller().getAgentID());
			recordList.add(listObj);
		}
		else
		{
			ArrayList<LandListingRecord> recordList = new ArrayList<LandListingRecord>();
			recordList.add(listObj);
			this.sellerIndexedListingRecords.put(listObj.getSeller().getAgentID(), recordList);
		}
	}

	/**
	 * @param parcelObj
	 */
	public <T extends SimpleAgent & LandSellingProcess> void addParcelOnMarket(Parcel parcelObj, T seller, float askPrice, float wta, int maxBidsAllowed)
	{
		if(parcelObj == null) return;
		
		//this.toSellParcelList.add(parcelObj);
		// TODO: Remove the old one. The seller might update their ask price and WTA.
		//removeListing(parcelObj.getParcelID());
		
		LandListingRecord listObj = new LandListingRecord();
		listObj.parcelObj = parcelObj;
		listObj.setSeller(seller);
		listObj.askPrice = askPrice;
		listObj.willingToAccept = wta;
		listObj.maxBidsAllowed = maxBidsAllowed;
		
		listObj.recordStaus = ListingRecordStatus.listed;
		
		this.addParcelOnMarket(listObj);
	}

	/**
	 * Remove this parcel from the listing
	 * 
	 * @param parcelID
	 */
	public boolean removeListing(int parcelID)
	{
		// TODO Auto-generated method stub
		for(LandListingRecord lrObj : toSellParcelList)
		{
			if(lrObj.getParcelID() == parcelID)
			{
				toSellParcelList.remove(lrObj);
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Remove this parcel from the listing
	 * 
	 * @param parcelID
	 */
	public boolean removeListing(LandListingRecord listingRecObj)
	{
		// TODO Auto-generated method stub
		if(toSellParcelList.contains(listingRecObj))
		{
			toSellParcelList.remove(listingRecObj);
			return true;
		}
		else
		{
			return false;
		}
	}

	/**
	 * 
	 * @param bidRecord
	 */
	public void addBidRecord(LandBidRecord bidRecord)
	{
		try
		{
			if(parcelIndexedBidRecords.containsKey(bidRecord.getParcelID()))
			{
				ArrayList<LandBidRecord> list = parcelIndexedBidRecords.get(bidRecord.getParcelID());
				list.add(bidRecord);
			}
			else
			{
				ArrayList<LandBidRecord> list = new ArrayList<LandBidRecord>();
				list.add(bidRecord);
				parcelIndexedBidRecords.put(bidRecord.getParcelID(),list);
			}
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
	}
	
	/**
	 * Get the 
	 * @param parcelID
	 * @return
	 */
	public int getCurrentBidsNumber(int parcelID)
	{
		if(parcelIndexedBidRecords.containsKey(parcelID))
		{
			ArrayList<LandBidRecord> parcelBids = parcelIndexedBidRecords.get(parcelID);
			return parcelBids.size();
		}
		else
		{
			return 0;
		}
	}
	
	/**
	 * The number of open bid on this record.
	 * @param parcelID
	 * @return
	 */
	public int getOpenBidNumber(int parcelID)
	{
		if(parcelIndexedBidRecords.containsKey(parcelID))
		{
			int rslt = 0;
			ArrayList<LandBidRecord> parcelBids = parcelIndexedBidRecords.get(parcelID);
			
			for(LandBidRecord lbRec : parcelBids)
			{
				if(lbRec.getRecordStatus() == BidRecordStatus.open)
				{
					rslt ++;
				}
			}
			
			return rslt;
		}
		else
		{
			return 0;
		}
		
	}

	/**
	 * @return the toSellParcelList
	 */
	public ArrayList<LandListingRecord> getToSellParcelList()
	{
		return toSellParcelList;
	}

	/**
	 * Return all bid records on the parcel with this parcelID.
	 * @param parcelID
	 * @return
	 */
	public ArrayList<LandBidRecord> getBidRecordsByParcelID(int parcelID)
	{
		if(parcelID < 0)
		{ 
			return null;
		}
		
		if(parcelIndexedBidRecords.containsKey(parcelID))
		{
			return parcelIndexedBidRecords.get(parcelID);
		}
		else
		{
			return null;
		}
	}

	/**
	 * @return the parcelList
	 */
	public Hashtable<Integer, Parcel> getParcelList()
	{
		return parcelList;
	}

	/**
	 * @param parcelList the parcelList to set
	 */
	public void setParcelList(Hashtable<Integer, Parcel> parcelList)
	{
		this.parcelList = parcelList;
	}

	/**
	 * Remove all elements in the market
	 */
	public void clear()
	{
		this.parcelIndexedBidRecords.clear();
		this.sellerIndexedListingRecords.clear();
		this.toSellParcelList.clear();
	}
}
