/**
 * SLUCE 2 Project
 * http://www.cscs.umich.edu/research/projects/sluce/
 */
package sluceii.landscape;

import umn.geog.helia.landscape.CompoundCell;
import umn.geog.helia.landscape.LandscapeLayerGroup;
import umn.geog.helia.landscape.RasterLayer;

/**
 * proportion of land cover and biome types the model will have to run once for
 * each of these land-cover / biome types, so we will have to check if prct
 * > 0 and if so then run it for that land-cover / biome type
 * 
 * @author Derek Robinson
 */
public abstract class LandCover
{
	/**
	 * this is the method used to calculate aggregate/statistics for parcel/subdivision from
	 * cells/parcels.
	 * @author Shipeng
	 */
	public enum LandAggregateMethod {average, min, max, percentile_ten, percentile_fifty};
	
	/**
	 * the ID of the land cover unit
	 */
	protected int landCoverID;
	
	/**
	 * the x and y of the center of the land cover
	 * NOTE: not necessarily the centroid, but must be within the land cover
	 */
	protected int xCenterIndex, yCenterIndex;
	
	
	/**
	 * 
	 */
	protected double xCenterPos, yCenterPos;
	
	/**
	 * The biophysical feature of a land cover
	 * @author Shipeng Sun
	 *
	 */
	public class LandBiophysicalFeature
	{
		/**
		 * % of the area defined as impervious
		 */
		@SuppressWarnings("unused")
		private double prctImpervious; 
		
		/**
		 * % of the area defined as water
		 */
		protected double prctWater; 
		
		/**
		 * % of the area defined as evergreen needle leaf
		 */
		protected double prctENF;
		
		/**
		 * % of the area defined as deciduous needle leaf
		 */
		protected double prctDNF; 
		
		/**
		 * % of the area defined as deciduous broadleaf
		 */
		protected double prctDBF;
		
		/**
		 * % of the area defined as c3 grasses
		 */
		protected double prctC3grass; 
		
		/**
		 * % of the area defined as c4 grasses
		 */
		protected double prctC4grass;
		
		/**
		 *  % of the area defined as shrub
		 */
		protected double prctShrub; 
		
		/**
		 * % of the area defined as evergreen broadleaf
		 */
		protected double prctEBF;
	};
	
	public LandBiophysicalFeature biophysicalValues;

	/**
	 * 
	 */
	public LandCover()
	{
		this.xCenterIndex = -1;
		this.yCenterIndex = -1;
		this.xCenterPos = Double.NaN;
		this.yCenterPos = Double.NaN;
	}
	
	/**
	 * @return the xCenterIndex
	 */
	public int getXCenterIndex()
	{
		return xCenterIndex;
	}

	/**
	 * @param centerIndex the xCenterIndex to set
	 */
	public void setXCenterIndex(int centerIndex)
	{
		xCenterIndex = centerIndex;
	}

	/**
	 * @return the yCenterIndex
	 */
	public int getYCenterIndex()
	{
		return yCenterIndex;
	}

	/**
	 * @param centerIndex the yCenterIndex to set
	 */
	public void setYCenterIndex(int centerIndex)
	{
		yCenterIndex = centerIndex;
	}

	/**
	 * @return the xCenterPos
	 */
	public double getXCenterPos()
	{
		return xCenterPos;
	}

	/**
	 * @param centerPos the xCenterPos to set
	 */
	public void setXCenterPos(double centerPos)
	{
		xCenterPos = centerPos;
	}

	/**
	 * @return the yCenterPos
	 */
	public double getYCenterPos()
	{
		return yCenterPos;
	}

	/**
	 * @param centerPos the yCenterPos to set
	 */
	public void setYCenterPos(double centerPos)
	{
		yCenterPos = centerPos;
	} 
	
	/**
	 * @return the landCoverID
	 */
	public int getLandCoverID()
	{
		return landCoverID;
	}

	/**
	 * @param landCoverID the landCoverID to set
	 */
	public void setLandCoverID(int landCoverID)
	{
		this.landCoverID = landCoverID;
	}

	/**
	 * @return
	 */
	public boolean hasPositionInfo()
	{
		return !(getXCenterIndex() < 0 || getYCenterIndex() < 0 || Double.isNaN(getXCenterPos()) || Double.isNaN(getYCenterPos()));
	}
	
	/**
	 * An abstract method used to obtain values from a LandCover object.
	 * @param dataLayer
	 * @param aggregateMethod
	 * @return The value of the LandCover aggregated using the given method
	 */
	public abstract Number getValue(RasterLayer<?> dataLayer, LandAggregateMethod aggregateMethod);
	
	/**
	 * This value will overwrite all sub level units.
	 * @param dataLayer
	 * @param val
	 * @return
	 */
	public abstract boolean setValue(RasterLayer<?> dataLayer, Number val);
	
	/**
	 * @param layerGroup
	 * @param aggregateMethod
	 * @return
	 */
	public abstract CompoundCell getValue(LandscapeLayerGroup layerGroup, LandAggregateMethod aggregateMethod);
}
