/**
 * SLUCE 2 Project
 * http://www.cscs.umich.edu/research/projects/sluce/
 */
package sluceii.landscape;

import java.io.File;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Hashtable;

import repast.simphony.engine.environment.RunEnvironment;
import sluceii.agent.SimpleAgent;
import sluceii.landscape.LandCover.LandAggregateMethod;
import sluceii.app.aag.AAGAppParameters;
import sluceii.agent.utility.ConsumptionBundle;
import umn.geog.helia.gis.HeliaDataLoader;
import umn.geog.helia.landscape.LandscapeLayerGroup;
import umn.geog.helia.landscape.LandscapeLoader;
import umn.geog.helia.landscape.RasterLayer;

/**
 * This class provides land cover data services to all agents.
 * It has a root landscape group object, which store all raster data layers.
 * 
 * This class mainly utilizes HELIA GIS java components.
 * @see hegis.umn.edu
 * 
 * @author Shipeng Sun
 * @version 1.0
 *
 */
public class LandCoverDataHelper
{
	/**
	 * The root landscape layer group object. Shared among all agents.
	 */
	public static LandscapeLayerGroup rootLayerGroup = new LandscapeLayerGroup();
	
	/**
	 * A GIS class used to load data 
	 */
	public static HeliaDataLoader heliaObj = new HeliaDataLoader();
	
	/**
	 * A hash table/dictionary of all parcel objects
	 */
	public static Hashtable<Integer, Parcel> parcelList = new Hashtable<Integer, Parcel>();
	
	/**
	 * A hash table/dictionary of all subdivision objects
	 */
	public static Hashtable<Integer, Subdivision> subdivisionList = new Hashtable<Integer, Subdivision>();

	/**
	 * 
	 */
	public LandCoverDataHelper()
	{
		// TODO Auto-generated constructor stub
	}
	
	public static void clear()
	{
		rootLayerGroup.deepClear();
		parcelList.clear();
		subdivisionList.clear();
	}
	
	/**
	 * Load the root landscape layer group information to the object.
	 * This would not load the actual raster data, it just load layer meta data.
	 * 
	 * @return success or failure
	 */
	public static boolean loadRootLandscapeLayerGroupInfo()
	{
		String rootXML = AAGAppParameters.dataFolderName + AAGAppParameters.landscapeGroupXMLFile;
		LandscapeLoader llObj = new LandscapeLoader(rootXML,rootLayerGroup);
		return llObj.loadLandscapeLayers();
	}
	
	/**
	 * @param rootXMLFile
	 * @return
	 */
	public static boolean loadRootLandscapeLayerGroupInfo(String rootXMLFile)
	{
		LandscapeLoader llObj = new LandscapeLoader(rootXMLFile,rootLayerGroup);
		return llObj.loadLandscapeLayers();
	}
	
	/**
	 * @param rootPath
	 * @param rootFile
	 * @return
	 */
	public static boolean loadRootLandscapeLayerGroupInfo(String rootPath, String rootFile)
	{
		LandscapeLoader llObj = new LandscapeLoader(rootPath+rootFile,rootLayerGroup);
		return llObj.loadLandscapeLayers();
	}
	
	/**
	 * Alias of loadRootLandscapeLayerGroupInfo
	 * @return success or failure
	 */
	public static boolean loadRootLandscapeLayerGroupMetadata()
	{
		return LandCoverDataHelper.loadRootLandscapeLayerGroupInfo();
	}
	
	/**
	 * @param rootXMLFile
	 * @return
	 */
	public static boolean loadRootLandscapeLayerGroupMetadata(String rootXMLFile)
	{
		return LandCoverDataHelper.loadRootLandscapeLayerGroupInfo(rootXMLFile);
	}
	
	/**
	 * @param rootPath
	 * @param rootFile
	 * @return
	 */
	public static boolean loadRootLandscapeLayerGroupMetadata(String rootPath, String rootFile)
	{
		return LandCoverDataHelper.loadRootLandscapeLayerGroupInfo(rootPath, rootFile);
	}
	
	/**
	 * Load the root landscape layer group data to the object.
	 * This would load the actual raster data. Make sure load the meta-data before calling this.
	 * 
	 * @return success or failure
	 */
	public static boolean loadRootLandscapeLayerGroupData()
	{
		return heliaObj.loadLandscapeGroupData(rootLayerGroup);
	}

	/**
	 * Load the root landscape layer group information and data to the object.
	 * This would load the actual raster data.
	 * 
	 * @return success or failure
	 */
	public static boolean loadRootLandscapeLayerGroup()
	{
		if(LandCoverDataHelper.loadRootLandscapeLayerGroupInfo())
		{
			return LandCoverDataHelper.loadRootLandscapeLayerGroupData();
		}
		else
		{
			return false;
		}
	}
	
	/**
	 * @param rootXMLFileName
	 * @return
	 */
	public static boolean loadRootLandscapeLayerGroup(String rootXMLFileName)
	{
		if(LandCoverDataHelper.loadRootLandscapeLayerGroupInfo(rootXMLFileName))
		{
			return LandCoverDataHelper.loadRootLandscapeLayerGroupData();
		}
		else
		{
			return false;
		}
	}
	
	/**
	 * @param rootPath
	 * @param xmlFileName
	 * @return
	 */
	public static boolean loadRootLandscapeLayerGroup(String rootPath, String xmlFileName)
	{
		if(LandCoverDataHelper.loadRootLandscapeLayerGroupInfo(rootPath, xmlFileName))
		{
			return LandCoverDataHelper.loadRootLandscapeLayerGroupData();
		}
		else
		{
			return false;
		}
	}	
	
	/**
	 * @return
	 */
	public static boolean saveAllLandscapeLayers()
	{
		return heliaObj.saveLandscapeGroupData(rootLayerGroup);
	}
	
	/**
	 * Save a raster layer. It can add a time stamp and/or tick count to the file name.
	 * @param dataLayer The raster layer object
	 * @param addTimeStamp Whether to add a time stamp to the file name
	 * @param addTickCount Whether to add a tick count to the file name
	 * @return success or failure. If the layer has not been changed, it will not be saved. 
	 */
	public static boolean saveRasterLayer(RasterLayer<?> dataLayer, boolean addTimeStamp, boolean addTickCount)
	{
		String fileName = dataLayer.getStoragePath();
		createFolderForFile(fileName);
		String pathName = fileName.substring(0,fileName.length()-4);
		String exntName = fileName.substring(fileName.length()-4,fileName.length());
		String newFileName = pathName;
		
		if(addTimeStamp)
		{
			DateFormat dateFormat = new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss_SSS");
            java.util.Date date = new java.util.Date();
            String datetime = dateFormat.format(date);

            newFileName = newFileName + "_" + datetime;
		}
		
		if(addTickCount)
		{
			int tickCount = (int)RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
			newFileName = newFileName + "_" + tickCount;
		}

		newFileName = newFileName + exntName;
		
		return LandCoverDataHelper.heliaObj.saveLandscapeLayerData(dataLayer, newFileName);

	}

	/**
	 * Save a raster layer. It can add a time stamp and/or tick count to the file name.
	 * @param dataLayer The raster layer object
	 * @param addTimeStamp Whether to add a time stamp to the file name
	 * @param addTickCount Whether to add a tick count to the file name
	 * @return success or failure. If the layer has not been changed, it will not be saved. 
	 */
	public static String saveRasterLayer2(RasterLayer<?> dataLayer, boolean addTimeStamp, boolean addTickCount, boolean shortFileName)
	{
		String fileName = dataLayer.getStoragePath();
		createFolderForFile(fileName);
		String pathName = fileName.substring(0,fileName.length()-4);
		String exntName = fileName.substring(fileName.length()-4,fileName.length());
		String newFileName = "";
		
		if(addTimeStamp)
		{
			DateFormat dateFormat = new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss_SSS");
            java.util.Date date = new java.util.Date();
            String datetime = dateFormat.format(date);

            newFileName = newFileName + "_" + datetime;
		}
		
		if(addTickCount)
		{
			int tickCount = (int)RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
			newFileName = newFileName + "_" + tickCount;
		}

		newFileName = pathName + newFileName + exntName;
		
		LandCoverDataHelper.heliaObj.saveLandscapeLayerData(dataLayer, newFileName);
		
		if(shortFileName)
		{
			int index = Math.max(newFileName.lastIndexOf("/"), newFileName.lastIndexOf("\\"));
			return newFileName.substring(index + 1);
		}
		else
		{
			return newFileName;
		}
	}

	/**
	 * Save a raster layer. It can add a time stamp and/or tick count to the file name.
	 * @param dataLayer The raster layer object
	 * @param addTimeStamp Whether to add a time stamp to the file name
	 * @param addTickCount Whether to add a tick count to the file name
	 * @return success or failure. If the layer has not been changed, it will not be saved. 
	 */
	public static String saveRasterLayer3(RasterLayer<?> dataLayer, boolean addTimeStamp, boolean addTickCount, String postFixName, boolean shortFileName)
	{
		String fileName = dataLayer.getStoragePath();
		createFolderForFile(fileName);
		String pathName = fileName.substring(0,fileName.length()-4);
		String exntName = fileName.substring(fileName.length()-4,fileName.length());
		String newFileName = "";
		
		if(addTimeStamp)
		{
			DateFormat dateFormat = new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss_SSS");
            java.util.Date date = new java.util.Date();
            String datetime = dateFormat.format(date);

            newFileName = newFileName + "_" + datetime;
		}
		
		if(addTickCount)
		{
			int tickCount = (int)RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
			newFileName = newFileName + "_" + tickCount;
		}
		
		if(postFixName.isEmpty() || postFixName.equalsIgnoreCase(""))
		{
			
		}
		else
		{
			newFileName = newFileName + "_" + postFixName;
		}

		newFileName = pathName + newFileName + exntName;
		
		LandCoverDataHelper.heliaObj.saveLandscapeLayerData(dataLayer, newFileName);
		
		if(shortFileName)
		{
			int index = Math.max(newFileName.lastIndexOf("/"), newFileName.lastIndexOf("\\"));
			return newFileName.substring(index + 1);
		}
		else
		{
			return newFileName;
		}
	}
	
	/**
	 * Save a raster layer. It can add a time stamp and/or tick count to the file name.
	 * @param dataLayer The raster layer object
	 * @param dirName An externally given path, it it is empty, "NO" or "NO OUTPUT". No files will be written.
	 * @param addTimeStamp Whether to add a time stamp to the file name
	 * @param addTickCount Whether to add a tick count to the file name
	 * @return success or failure. If the layer has not been changed, it will not be saved. 
	 */
	public static String saveRasterLayer4(RasterLayer<?> dataLayer, String dirName, boolean addTimeStamp, boolean addTickCount, String postFixName, boolean shortFileName)
	{
		if(dirName == "" || dirName.equalsIgnoreCase("No") || dirName.equalsIgnoreCase("No output"))
		{
			return "";
		}
		
		String fileName = dataLayer.getStoragePath();
		createFolderForFile(fileName);
		
		int index = Math.max(fileName.lastIndexOf("/"), fileName.lastIndexOf("\\"));
		String relativeFileName = fileName.substring(index + 1);
		String newFileName = relativeFileName.substring(0,relativeFileName.length()-4);
		String exntName = fileName.substring(fileName.length()-4,fileName.length());
 	
		if(addTimeStamp)
		{
			DateFormat dateFormat = new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss_SSS");
            java.util.Date date = new java.util.Date();
            String datetime = dateFormat.format(date);

            newFileName = newFileName + "_" + datetime;
		}
		
		if(addTickCount)
		{
			int tickCount = (int)RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
			newFileName = newFileName + "_" + tickCount;
		}
		
		if(postFixName.isEmpty() || postFixName.equalsIgnoreCase(""))
		{
			
		}
		else
		{
			newFileName = newFileName + "_" + postFixName;
		}

		newFileName = dirName + newFileName + exntName;
		
		LandCoverDataHelper.heliaObj.saveLandscapeLayerData(dataLayer, newFileName);
		
		if(shortFileName)
		{
			index = Math.max(newFileName.lastIndexOf("/"), newFileName.lastIndexOf("\\"));
			return newFileName.substring(index + 1);
		}
		else
		{
			return newFileName;
		}
	}	   		
	
	/**
	 * If the folder contained in the filaName doesn't exist, create the folder.
	 * @param fileName Full file name with path
	 * @return true if the folder exists now; false if it is not correctly created.
	 */
	public static boolean createFolderForFile(String fileName)
	{
		// Create the folder if it does not exist.
		// This must be explicitly called on Linux/Unix/Mac OS.
		int index = Math.max(fileName.lastIndexOf("/"), fileName.lastIndexOf("\\"));
		String folder = fileName.substring(0,index);
		File folderObj = new File(folder);
		if(!folderObj.exists())
		{
			return folderObj.mkdir();
		}	
		
		return true;
	}
	
	/**
	 * 
	 * @param land the land object, could be a cell, parcel or subdivision,
	 * @param agent the agent who checks out the consumption bundle
	 * @return
	 */
	public static ConsumptionBundle getConsumptionBundle(LandCover land, SimpleAgent agent)
	{
		// TODO: implement this using real data
		ConsumptionBundle cbObj = new ConsumptionBundle();
		RasterLayer<?> envLayer = (RasterLayer<?>)rootLayerGroup.getLandscapeLayer(111);
		RasterLayer<?> disLayer = (RasterLayer<?>)rootLayerGroup.getLandscapeLayer(122);
		
		if(land instanceof Parcel)
		{
			cbObj.addValueToObservedList( land.getValue(envLayer, LandAggregateMethod.average).doubleValue());
			cbObj.addValueToObservedList( land.getValue(disLayer, LandAggregateMethod.min).doubleValue());
			
			cbObj.addValueToPreferenceWeightList(0.5);
			cbObj.addValueToPreferenceWeightList(0.5);
			
		}
		else if(land instanceof Subdivision)
		{
			
		}
		else
		{
			return null;
		}
		
		return cbObj;
	}
	
	/**
	 * Add a parcel object to the list
	 * @param parcelObj
	 * @return
	 */
	public static boolean addToParcelList(Parcel parcelObj)
	{
		if(parcelObj == null || parcelObj.getParcelID() < 0)
		{
			return false;
		}
		else
		{
			LandCoverDataHelper.parcelList.put(parcelObj.getParcelID(),parcelObj);
			return true;
		}
	}
	
	/**
	 * Add a subdivision object to the list
	 * @param subdivisionObj
	 * @return
	 */
	public static boolean addToSubdivisionList(Subdivision subdivisionObj)
	{
		if(subdivisionObj == null || subdivisionObj.getSubdivisionID() < 0)
		{
			return false;
		}
		else
		{
			LandCoverDataHelper.subdivisionList.put(subdivisionObj.getSubdivisionID(),subdivisionObj);
			return true;
		}
	}
	
	/**
	 * This method use a parcel ID layer to update the parcel list.
	 * It mainly assigns cells to each parcel. 
	 * Before calling this method, add all parcels to the parcel list.
	 * @param parcelIDLayer
	 * @return
	 */
	public static boolean updateParcelList(RasterLayer<?> parcelIDLayer)
	{
		if(parcelIDLayer == null || LandCoverDataHelper.parcelList.size() == 0) return false;
		
		int parcelID = 0, cellID = 0;
		Parcel pObj = null;
		
		for(int i = 0 ; i < parcelIDLayer.getCols(); i++)
		{
			for(int j = 0; j < parcelIDLayer.getRows(); j++)
			{
				parcelID = parcelIDLayer.getValueAt(i, j).intValue();
				if(parcelID == AAGAppParameters.notALayerValue) continue;
				if(!LandCoverDataHelper.parcelList.containsKey(parcelID)) continue;
				
				pObj = LandCoverDataHelper.parcelList.get(parcelID);
				
				if(pObj != null)
				{
					// There are no explicit cell ID, using their index is fine.
					cellID = i + j*parcelIDLayer.getCols();
					pObj.addCellID(cellID);
					
					// TODO: This method uses the first cell of the parcel as the location.
					// Change this to arithmetic center.
					if(!pObj.hasPositionInfo())
					{
						pObj.setXCenterIndex(i);
						pObj.setYCenterIndex(j);
						pObj.setXCenterPos(parcelIDLayer.getCellCenterX(i));
						pObj.setYCenterPos(parcelIDLayer.getCellCenterY(j));
					}
				}
			}
		}
		
		return true;
	}

	/**
	 * Add parcel IDs to each subdivision object.
	 * Before calling this method, add all subdivision objects into the list.
	 * @param parcelIDLayer
	 * @param subdivisionIDLayer
	 * @return
	 */
	public static boolean updateSubdivisionList(RasterLayer<?> parcelIDLayer, RasterLayer<?> subdivisionIDLayer)
	{
		if(parcelIDLayer == null || parcelList.size() == 0 || subdivisionIDLayer == null || subdivisionList.size() == 0) return false;
		
		int parcelID = 0, subdivisionID = 0;
		Subdivision sObj = null;
		
		for(int i = 0 ; i < subdivisionIDLayer.getCols(); i++)
		{
			for(int j = 0; j < subdivisionIDLayer.getRows(); j++)
			{
				subdivisionID = subdivisionIDLayer.getValueAt(i, j).intValue();
				parcelID = parcelIDLayer.getValueAt(i, j).intValue();
				sObj = subdivisionList.get(subdivisionID);
				
				// Add this parcel to the subdivision
				if(sObj != null)
				{
					sObj.addParcelID(parcelID);
				}
			}
		}
		
		return true;
	}
	
	/**
	 * @param dataLayer
	 * @param addTimeStamp
	 * @return
	 */
	public static boolean saveRasterLayer(RasterLayer<?> dataLayer, boolean addTimeStamp)
	{
		if(addTimeStamp)
		{
			String fileName = dataLayer.getStoragePath();
			DateFormat dateFormat = new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss_SSS");
            java.util.Date date = new java.util.Date();
            String datetime = dateFormat.format(date);

            String timeFileName = fileName.substring(0,fileName.length()-4) + "_" + datetime + fileName.substring(fileName.length()-4,fileName.length());
            
            return LandCoverDataHelper.heliaObj.saveLandscapeLayerData(dataLayer,timeFileName);
		}
		else
		{
			return LandCoverDataHelper.heliaObj.saveLandscapeLayerData(dataLayer);
		}
	}
	
}
