/**
 * SLUCE 2 Project
 * http://www.cscs.umich.edu/research/projects/sluce/
 */
package sluceii.landscape;

import java.util.ArrayList;
import java.util.Hashtable;

import umn.geog.helia.landscape.CompoundCell;
import umn.geog.helia.landscape.LandscapeLayerGroup;
import umn.geog.helia.landscape.RasterLayer;

/**
 * This is an abstract class used to implement the hierarchical structure of land unit.
 * A typical example of such hierarchy is cell, parcel, subdivision, or neighborhood, 
 * city, county, state.
 * 
 * A higher level unit has a list of the IDs of the lower level. It also maintains a 
 * table of the lower level units.
 * 
 * @author Shipeng Sun
 *
 */
public abstract class LandCoverUnit<T extends LandCover> extends LandCover
{
	/**
	 * THe ID list of the lower level land use units
	 */
	protected ArrayList<Integer> lowerLeverlIDList;
	
	/**
	 * 
	 */
	public LandCoverUnit()
	{
		// TODO Auto-generated constructor stub
	}

	/**
	 * Get the hash table/dictionary of the lower lever land use units.
	 * Its keys are the identifiers of those land use units.
	 * @return A hash table/dictionary of the lower lever land use units. 
	 */
	public abstract Hashtable<Integer, T> getLowerLevelUnits();
	
	/**
	 * @param llUnits The hash table/dictionary of the lower lever land use units
	 */
	public abstract void setLowerLevelUnits(Hashtable<Integer, T> llUnits);
	
	/**
	 * The Unit ID
	 * @return
	 */
	public int getUnitID()
	{
		return this.getLandCoverID();
	}

	/**
	 * @param unitID the unitID to set
	 */
	public void setUnitID(int unitID)
	{
		this.setLandCoverID(unitID);
	}
	
	/* (non-Javadoc)
	 * @see sluceii.landscape.LandCover#setValue(umn.geog.helia.landscape.RasterLayer, java.lang.Number)
	 */
	public boolean setValue(RasterLayer<?> dataLayer, Number val)
	{
		int unitID = 0;
		T lowerLevelUnitObj = null;
	
		Hashtable<Integer, T> llUnits = this.getLowerLevelUnits();
		
		for(int i = 0; i < this.lowerLeverlIDList.size(); i++)
		{
			unitID = this.lowerLeverlIDList.get(i);
			
			if(llUnits.containsKey(unitID))
			{
				lowerLevelUnitObj = llUnits.get(unitID);
				
				if(lowerLevelUnitObj == null)
				{
					return false;
				}
				else
				{
					lowerLevelUnitObj.setValue(dataLayer, val);
				}
			}
		}
		return true;
	}

	/* (non-Javadoc)
	 * @see sluceii.landscape.LandCover#getValue(umn.geog.helia.landscape.RasterLayer, sluceii.landscape.LandCover.LandAggregateMethod)
	 */
	public Number getValue(RasterLayer<?> dataLayer, LandAggregateMethod aggregateMethod )
	{
		if(dataLayer == null)
		{
			return Float.NaN;
		}
		
		T lowerLevelUnitObj = null;
	
		Hashtable<Integer, T> llUnits = this.getLowerLevelUnits();

		int unitID = 0;
		if(aggregateMethod == LandAggregateMethod.average)
		{
			Number val = 0.0;
			for(int i = 0; i < this.lowerLeverlIDList.size(); i++)
			{
				unitID = this.lowerLeverlIDList.get(i);
				
				if(llUnits.containsKey(unitID))
				{
					lowerLevelUnitObj = llUnits.get(unitID);
					if(lowerLevelUnitObj != null)
					{
						val = val.doubleValue() + lowerLevelUnitObj.getValue(dataLayer, aggregateMethod).doubleValue();
					}
				}
			}
			
			return val.doubleValue()/lowerLeverlIDList.size();
		}
		else if(aggregateMethod == LandAggregateMethod.max)
		{
			double maxVal = Double.NEGATIVE_INFINITY;
			for(int i = 0; i < this.lowerLeverlIDList.size(); i++)
			{
				unitID = this.lowerLeverlIDList.get(i);
				
				if(llUnits.containsKey(unitID))
				{
					lowerLevelUnitObj = llUnits.get(unitID);
					if(lowerLevelUnitObj != null)
					{
						maxVal = Math.max(maxVal,lowerLevelUnitObj.getValue(dataLayer, aggregateMethod).doubleValue());
					}
				}
			}
			
			return maxVal;
		}
		else if(aggregateMethod == LandAggregateMethod.min)
		{
			double minVal = Double.POSITIVE_INFINITY;
			
			for(int i = 0; i < this.lowerLeverlIDList.size(); i++)
			{
				unitID = this.lowerLeverlIDList.get(i);
				
				if(llUnits.containsKey(unitID))
				{
					lowerLevelUnitObj = llUnits.get(unitID);
					if(lowerLevelUnitObj != null)
					{
						minVal = Math.max(minVal,lowerLevelUnitObj.getValue(dataLayer, aggregateMethod).doubleValue());
					}
				}
			}
			
			return minVal;
		}
	
		return Float.NaN;
	}	
	
	/* (non-Javadoc)
	 * @see sluceii.landscape.LandCover#getValue(umn.geog.helia.landscape.LandscapeLayerGroup, sluceii.landscape.LandCover.LandAggregateMethod)
	 */
	public CompoundCell getValue(LandscapeLayerGroup layerGroup, LandAggregateMethod aggregateMethod)
	{
		layerGroup.buildFastLayersArray();
		int totalLayersNum = layerGroup.getTotalLayersNum();
		
		CompoundCell cell = new CompoundCell((short)totalLayersNum);
		
		for(int i = 0; i < totalLayersNum; i++)
		{
			RasterLayer<?> layerObj = (RasterLayer<?>)layerGroup.getLandscapeLayerByIndex(i);
			if(layerObj != null)
			{
				cell.put(i, this.getValue(layerObj, aggregateMethod));
			}
		}
		
        return cell;
	}

	/**
	 * 
	 */
	public void addUnitID(int unitID)
	{
		// TODO Auto-generated method stub
		if(!lowerLeverlIDList.contains(unitID))
		{
			this.lowerLeverlIDList.add(unitID);
		}
	}
}
