/**
 * SLUCE 2 Project
 * http://www.cscs.umich.edu/research/projects/sluce/
 */
package sluceii.landscape;

import java.util.ArrayList;

import repast.simphony.valueLayer.GridValueLayer;
import umn.geog.helia.landscape.CompoundCell;
import umn.geog.helia.landscape.LandscapeLayerGroup;
import umn.geog.helia.landscape.RasterLayer;

/**
 * @author Shipeng Sun
 *
 */
public class Parcel extends LandCover
{
	/**
	 * The market value of this parcel. This could be estimated or actual sale value.
	 * NOTE: this is different from the WTA or WTP.
	 */
	protected float marketValue;
	
	/**
	 * A parcel has multiple cells.
	 * TODO: see if IDs or objects are stored in the object.
	 * An alternative method is the parcel ID layer.
	 */
	ArrayList<Integer> cellIDList;
	
	
	/**
	 * 
	 */
	public Parcel()
	{
		super();
		// TODO Auto-generated constructor stub
		cellIDList = new ArrayList<Integer>();
	}


	/**
	 * @return the parcelID
	 */
	public int getParcelID()
	{
		return this.getLandCoverID();
	}


	/**
	 * @param parcelID the parcelID to set
	 */
	public void setParcelID(int parcelID)
	{
		this.landCoverID = parcelID;
	}


	/**
	 * @return the marketValue
	 */
	public float getMarketValue()
	{
		return marketValue;
	}


	/**
	 * @param marketValue the marketValue to set
	 */
	public void setMarketValue(float marketValue)
	{
		this.marketValue = marketValue;
	}


	/**
	 * Add a cell ID into the cell ID list
	 * @param cellID
	 */
	public boolean addCellID(int cellID)
	{
		if(this.cellIDList == null) return false;
		
		// No duplicate cell IDs
		if(this.cellIDList.contains(cellID))
		{
			return false;
		}
		else
		{
			return this.cellIDList.add(cellID);
		}
	}
	
	/**
	 * @param cellID
	 * @return
	 */
	public boolean removeCellID(int cellID)
	{
		if(cellIDList == null) return false;
		
		// No duplicate cell IDs
		if(cellIDList.contains(cellID))
		{
			cellIDList.remove(cellIDList.indexOf(cellID));
			return true;
		}
		else
		{
			return false;
		}

	}

	/**
	 * Get the value of the parcel from a raster layer using the given aggregation/statistics method.
	 * @param dataLayer
	 * @param aggregateMethod
	 * @return
	 */
	public Number getValue(RasterLayer<?> dataLayer, LandAggregateMethod aggregateMethod)
	{
		if(dataLayer == null)
		{
			return 0.0f;
		}

		int cellID = 0, x = 0, y = 0;
		if(aggregateMethod == LandAggregateMethod.average)
		{
			double val = 0.0;
			for(int i = 0; i < this.cellIDList.size(); i++)
			{
				cellID = this.cellIDList.get(i);
				y = (int)Math.floor(cellID/dataLayer.getCols());
				x = cellID - (y*dataLayer.getCols());
				
				val += dataLayer.getValueAt(x, y).doubleValue();
			}
			
			return val/cellIDList.size();
		}
		else if(aggregateMethod == LandAggregateMethod.max)
		{
			double maxVal = Double.NEGATIVE_INFINITY;
			for(int i = 0; i < this.cellIDList.size(); i++)
			{
				cellID = this.cellIDList.get(i);
				y = (int)Math.floor(cellID/dataLayer.getCols());
				x = cellID - (y*dataLayer.getCols());
				
				maxVal = Math.max(maxVal, dataLayer.getValueAt(x, y).doubleValue());
			}
			
			return maxVal;
		}
		else if(aggregateMethod == LandAggregateMethod.min)
		{
			double minVal = Double.POSITIVE_INFINITY;
			for(int i = 0; i < this.cellIDList.size(); i++)
			{
				cellID = this.cellIDList.get(i);
				y = (int)Math.floor(cellID/dataLayer.getCols());
				x = cellID - (y*dataLayer.getCols());
				
				minVal = Math.min(minVal, dataLayer.getValueAt(x, y).doubleValue());
			}
			
			return minVal;
		}
	
		return 0.0f;
	}
	
	/**
	 * @param layerGroup
	 * @param aggregateMethod
	 * @return
	 */
	public CompoundCell getValue(LandscapeLayerGroup layerGroup, LandAggregateMethod aggregateMethod)
	{
		layerGroup.buildFastLayersArray();
		int totalLayersNum = layerGroup.getTotalLayersNum();
		
		CompoundCell cell = new CompoundCell((short)totalLayersNum);
		
		for(int i = 0; i < totalLayersNum; i++)
		{
			RasterLayer<?> layerObj = (RasterLayer<?>)layerGroup.getLandscapeLayerByIndex(i);
			if(layerObj != null)
			{
				cell.put(i, this.getValue(layerObj, aggregateMethod));
			}
		}
		
        return cell;
	}
	
	/**
	 * @param layer The layer to be updated
	 * @param val The value to be set
	 * @return
	 */
	public boolean putValue(RasterLayer<?> dataLayer, Number val)
	{
		if(dataLayer == null || Float.isNaN(val.floatValue()))
		{
			return false;
		}

		int cellID = 0, x = 0, y = 0;

		for(int i = 0; i < this.cellIDList.size(); i++)
		{
			cellID = this.cellIDList.get(i);
			y = (int)Math.floor(cellID/dataLayer.getCols());
			x = cellID - (y*dataLayer.getCols());
			
			dataLayer.putValueAt(x, y, val);
		}
			
		return true;
	}
	
	/**
	 * Set the val to the GridValueLayer according to the reference layer and this parcel object
	 * @param vlObj
	 * @param refLayer
	 * @param val
	 * @return
	 */
	public boolean putValue(GridValueLayer vlObj, RasterLayer<?> refLayer, Number val)
	{
		if(vlObj == null || refLayer == null || Float.isNaN(val.floatValue()))
		{
			return false;
		}

		int cellID = 0, x = 0, y = 0;

		for(int i = 0; i < this.cellIDList.size(); i++)
		{
			cellID = this.cellIDList.get(i);
			y = (int)Math.floor(cellID/refLayer.getCols());
			x = cellID - (y*refLayer.getCols());
			
			vlObj.set(val.doubleValue(), x, y);
		}
			
		return true;
	}
	
	/**
	 * Use the value layer to update the GridValueLayer at the locations of the parcel
	 * @param vlObj
	 * @param valLayer
	 * @return
	 */
	public boolean putValue(GridValueLayer vlObj, RasterLayer<?> valLayer)
	{
		if(vlObj == null || valLayer == null)
		{
			return false;
		}

		int cellID = 0, x = 0, y = 0;

		for(int i = 0; i < this.cellIDList.size(); i++)
		{
			cellID = this.cellIDList.get(i);
			y = (int)Math.floor(cellID/valLayer.getCols());
			x = cellID - (y*valLayer.getCols());
			
			vlObj.set(valLayer.getValueAt(x, y).doubleValue(), x, y);
		}
			
		return true;
	}
	
	/**
	 * Alias of putValue method
	 * @param dataLayer
	 * @param val
	 * @return
	 */
	public boolean setValue(RasterLayer<?> dataLayer, Number val)
	{
		return this.putValue(dataLayer, val);
	}
}
